/* -*- C -*-
 * Spamtest connection library
 * File: addrlist.c
 *      (C)  Alex Tutubalin <lexa@lexa.ru>
 *	Apr-May 2003
 *
 * Created: Wed Apr 23 22:08:08 2003
 */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "mess822.h"
#include "msgstore.h"
#include "spamtest.h"

int rfc822_unquote_one(char *dest,unsigned int size, char *from, char *defdomain)
{
    stralloc addrlist={0};
    int i,j;
    if(!dest || !from || !defdomain) return -1;
    if(size<1) return -1;
    if(strlen(from)<1) 
	return 0; // GIGO
    if(!mess822_addrlist(&addrlist,from))
	{
	    CLEAN_SA(addrlist);
	    return -1;
	}
    for(i=j=0; j< addrlist.len;j++)
	if(!addrlist.s[j])
	    {
		if( addrlist.s[i]=='+')
		    {
			strncpy(dest,addrlist.s+i+1,size);
			break;
		    }
		i=j+1;
	    }
    i = strlen(dest);
    if(i==1 && dest[0] == '@')
	{
	    dest[0]='\0'; // emtpy address
	}
    else if(i>1 && dest[i-1]=='@')
	{
	    strncpy(dest+i,defdomain,size-i);
	}
    if(*dest)
	dest[size-1]='\0';
    CLEAN_SA(addrlist);
    return 0;
}


int rfc822_quote_one(char *dest,unsigned int size, char *from, char *defdomain)
{
    stralloc quoted={0},tempaddr={0};
    int ret;
    if(!dest || !from || !defdomain) return -1;
    if(size<1) return 0; //GIGO
    if(strlen(from)<1) 
	return 0; // GIGO

    if((ret=rfc822_unquote_one(dest,size,from,defdomain)))
	{
	    CLEAN_SA(quoted);
	    CLEAN_SA(tempaddr);
	    return ret;
	}
    stralloc_copys(&tempaddr,dest);
    if(!mess822_quotelist(&quoted,&tempaddr))
	{
	    CLEAN_SA(quoted);
	    CLEAN_SA(tempaddr);
	    return -1;
	}
    strncpy(dest,quoted.s,quoted.len>size?size:quoted.len);
    dest[size-1]='\0';
    CLEAN_SA(quoted);
    CLEAN_SA(tempaddr);
    return 0;
}


recipient_list_t *rfc822_unquote_list(recipient_list_t *from,char *defdomain)
{
    recipient_list_t *ret;
    char *tempbuf=NULL;
    int  allocsize=0,n;
    int  dlen,l;
    
    if(!from || !defdomain) return NULL;
    dlen = strlen(defdomain);
    if(!(ret = rcptlist_init(from->storeptr)))
	return NULL;
    // process each addr, gotting MANY replies
    for(n=0;n<rcptlist_count(from);n++)
	{
	    // we need some space
	    l = strlen(rcptlist_getn(from,n));
	    if(allocsize < l+dlen+16)
		{
		    tempbuf = heap_realloc(from->storeptr,tempbuf,allocsize,l+dlen+16);
		    allocsize = l+dlen+16;
		}
	    if(!tempbuf)
		break;
	    if(rfc822_unquote_one(tempbuf,allocsize,rcptlist_getn(from,n),defdomain))
		break;
	    rcptlist_add(ret,tempbuf);
	}
    return ret;
}

recipient_list_t *rfc822_quote_list(recipient_list_t *from,char *defdomain)
{
    recipient_list_t *ret;
    char *tempbuf=NULL;
    int  allocsize=0,n;
    int  dlen,l;
    
    if(!from || !defdomain) return NULL;
    dlen = strlen(defdomain);
    if(!(ret = rcptlist_init(from->storeptr)))
	return NULL;
    // process each addr, gotting MANY replies
    for(n=0;n<rcptlist_count(from);n++)
	{
	    // we need some space
	    l = strlen(rcptlist_getn(from,n));
	    if(allocsize < l+dlen+16)
		{
		    tempbuf = heap_realloc(from->storeptr,tempbuf,allocsize,l+dlen+16);
		    allocsize = l+dlen+16;
		}
	    if(!tempbuf)
		break;
	    if(rfc822_quote_one(tempbuf,allocsize,rcptlist_getn(from,n),defdomain))
		break;
	    rcptlist_add(ret,tempbuf);
	}
    return ret;
}


#ifdef TEST

void recipient_writer(const char *rcpt, size_t unused)
{
    printf("RCPT TO: %s\n",rcpt);
}


int main()
{
  char abuf[100],bbuf[100];
  recipient_list_t *rcpt,*unq,*q;
  msg_heap_t *heap;
  int i;
  
  for(i=0;i<1;i++)
      {
	  heap = heap_init(10240);
	  
	  printf("Single Interface:\n");
	  if(rfc822_unquote_one(abuf,100,"(Joe User) joe@user.com","sun.com"))
	      exit(1);
	  printf("Unquoted: {%s}\n",abuf);
	  if(rfc822_quote_one(bbuf,100,abuf,"sun.com"))
	      exit(1);
	  printf("Final: {%s}\n",bbuf);
	  
	  if(rfc822_quote_one(bbuf,100," user ","sun.com"))
	      exit(1);
	  printf("Final2: {%s}\n",bbuf);
	  
	  printf("\n===\nMulti Interface:\nInput:\n");
	  if(!(rcpt = rcptlist_init(heap)))
	      exit(1);
	  rcptlist_add(rcpt,"");
	  rcptlist_add(rcpt,"vasya");
	  rcptlist_add(rcpt,"<petya>");
	  rcptlist_add(rcpt,"\"Kolya\"<petya>");
	  rcptlist_add(rcpt,"(AAA )\"Kolya\" petya");
	  rcptlist_add(rcpt,"vasya@joe.com");
	  rcptlist_add(rcpt,"Comment <vasya@joe.com>");
	  rcptlist_iterate(rcpt,recipient_writer);
	  
	  unq = rfc822_unquote_list(rcpt,"pupkin.ru");
	  if(!unq) exit(1);
	  printf("Unquoted:\n",bbuf);
	  rcptlist_iterate(unq,recipient_writer);
	  
	  q = rfc822_quote_list(unq,"pupkin.ru");
	  if(!q) exit(1);
	  printf("Quoted:\n",bbuf);
	  rcptlist_iterate(q,recipient_writer);
	  heap_free(heap);
      }
  return 0;
}
#endif
