/* -*- C -*-
 * File: ap-sample.c
 *
 * Created: Thu Apr 24 15:10:23 2003
 */
#include <sys/types.h>
#include <sys/time.h>
#include <sys/uio.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <syslog.h>
#include "msgstore.h"
#include "spamtest.h"

int process_messages(FILE *f,long long *bytes, int *messages,char *daemon);
int process_single_message(FILE *f,char *relay, long long *bytes,int *messages, char *daemon);
char *notspace(char *str);
char *get_tok(char *str, char** endp);
int send_body_and_process_result(FILE *f, spamtest_session_t *s, long long *bytes);

int process_accept_multi(spamtest_status_t *st,message_t *msg);
int process_accept_single(spamtest_status_t *st,message_t *msg);
int simple_reply(spamtest_status_t *st,char *reply);

void recipient_writer(const char *rcpt, size_t len);
void body_writer(const char *rcpt, size_t len);
void header_writer(char *hdr,char *val);

int logger_log_level=LOG_ERR;


#define MYDOMAIN "lexa.ru"
#define TIMEOUT_MS  30000 /* we'll use it per sts_* call */
#define MAX_MSGBUF  1024000

static int multiple_messages_allowed=0;
static char *tempdir = NULL;
static size_t max_message_rss = 0;

int process_messages(FILE *f,long long *bytes, int *messages,char *daemon)
{
    char linebuf[2048],*p;

    while(fgets(linebuf,2048,f))
	{
	    if(!strncmp(linebuf,"HELO:",5))
		if((p = get_tok(linebuf+5,NULL)))
		    {
			if(process_single_message(f,p,bytes,messages,daemon)<0)
			    return -1;
		    }
	}
    return 0;
}


int main(int ac, char *av[])
{
    int repetitions = 1;
    int i;
    struct timeval tv1,tv2;
    long long bytes=0;
    int messages=0;
    char *filename=NULL;
    char *daemon_address=NULL;
    char ch;
    FILE *f;
    int use_file=0;


    signal(SIGPIPE,SIG_IGN);
    openlog("ap-sample",LOG_PID,LOG_LOCAL1);

    while((ch = getopt(ac, av, "r:t:vT:m:R:")) != -1)
	{
	    switch(ch)
		{
		case 'r':
		    repetitions = atoi(optarg);
		    break;
		case 'm':
		    multiple_messages_allowed=1;
		    break;
		case 'T':
		    tempdir = strdup(optarg);
		    break;
		case 'R':
		    max_message_rss = atoi(optarg)*1024;
		    break;
		case 'v':
		    if(logger_log_level<LOG_DEBUG)
			logger_log_level=LOG_DEBUG;
		    else
			logger_log_level++;
		    break;
		    
		}
	    

	}

    ac-=optind;
    av+=optind;
    daemon_address = av[0]?strdup(av[0]):NULL;
    filename = av[1]?strdup(av[1]):NULL;
    if(!daemon_address)
	{

	    fprintf(stderr,"Usage: %s [params] spamtest-server inputfile\n",av[0]);
	    fprintf(stderr,"Param:\n"
		    "\t-r N - number of repetitions (for benchmarking)\n"
		    "\t-m  - multiplications of messages allowed\n"
		    "\t-T path - temp dir for internal message store (used with -R)\n"
		    "\t-R kilobytes - max in-memory size of message store\n"
		    "\t-v - increase verbosity\n"
		);
	    exit(1);
	}

    if(filename)
	{
	    if(!(f = fopen(filename,"r")))
		{
		    perror(filename);
		    exit(1);
		}
	    else
		use_file = 1;
	}
    else
	f = stdin;

    gettimeofday(&tv1,NULL);
    for(i=0;i<repetitions;i++)
	{
	    if(feof(f))
		fseek(f,0,SEEK_SET);
	    if(-1==process_messages(f,&bytes,&messages,daemon_address))
		{
		    fprintf(stderr,"Cannot process message, break\n");
		    break;
		}
	}
    if(use_file) fclose(f);
    gettimeofday(&tv2,NULL);
    if(messages>0)
	{
	    double elapsed = ((tv2.tv_sec-tv1.tv_sec)*1000000.0
			      +(tv2.tv_usec-tv1.tv_usec))/1000000.0;
	    fprintf(stderr,"%d messages sent in %7.5f seconds, %5.3f msg/sec\n",messages,
		    elapsed,messages/elapsed);
	    fprintf(stderr,"%lld bytes sent, %5.3f kbytes/sec\n",bytes,
		   bytes/elapsed/1024.0);
	}
    exit(0);
}

/*this function called as s->logp and can work with any user defined function */
int sample_logger(spamtest_session_t *s, int loglevel, const char *fmt, ...)
{
    va_list ap;
    if(loglevel>logger_log_level) return 0;
    va_start(ap,  fmt);
    vfprintf(stderr,fmt, ap);
    va_end(ap);
    fprintf(stderr,"\n");
    return 0;
}




int process_single_message(FILE *f,char *relay, long long *bytes,int *messages, char *daemon)
{
    char msgline[2048],*p;
    spamtest_session_t s;
    int r=0;

    bzero(&s,sizeof(s));
    s.access_address = daemon;
    s.mail_domain = MYDOMAIN;
    s.rw_timeout= TIMEOUT_MS;
    s.connect_timeout= TIMEOUT_MS;
    s.logp = sample_logger;
    s.flags = STS_USE_POLL;
    if((r=sts_init(&s))<0)
	{
	    fprintf(stderr,"sts_init: %s\n",sts_strerror(r));
	    return -1;
	}

    if(STS_ERR_OUT_OF_MEMORY==sts_prepare_envelope(&s,NULL,NULL,relay,NULL))
	return -1;
    
    /* begin processing message */
    while(fgets(msgline,2048,f))
	{
	    if(!strncmp(msgline,"MAIL FROM:",10))
		if((p = get_tok(msgline+10,NULL)))
		    {
			if(STS_ERR_OUT_OF_MEMORY==sts_prepare_envelope(&s,p,NULL,NULL,NULL)<0)
			    return -1;
		    }
	    if(!strncmp(msgline,"HELONAME:",9))
		if((p = get_tok(msgline+9,NULL)))
		    {
			if(STS_ERR_OUT_OF_MEMORY==sts_add_attribute(&s,"helo_name",p)<0)
			    return -1;
		    }
	    if(!strncmp(msgline,"RCPT TO:",8))
		if((p = get_tok(msgline+8,NULL)))
		    {
			if(STS_ERR_OUT_OF_MEMORY==sts_prepare_envelope(&s,NULL,p,NULL,&r)<0)
			    return -1;
		    }
	    if(!strncmp(msgline,"DATA",4))
		{
		    r=send_body_and_process_result(f,&s,bytes);
		    if(r>=0)
			(*messages)++;
		    break;
		}
	}
    sts_close(&s);
    return r;
}


int send_body_and_process_result(FILE *f, spamtest_session_t *s, long long *bytes)
{
    message_t *message_store = NULL;
    char linebuf[2048];
    int l,i;
    int message_read = 0;
    spamtest_status_t *st=NULL;
    int r;
    

    if(s->should_store_message)
	if(!(message_store=message_init_with_tempfile(tempdir,max_message_rss)))
	    goto ERROR;

    /* message processing loop */
    while(fgets(linebuf,sizeof(linebuf)-1,f))
	{
	    if(!strcmp(linebuf,".\n")) // single dot - end of DATA
				       // command
		{
		    message_read =1;
		    break;
		}

	    l = strlen(linebuf);
	    *bytes += l;

	    /* if we've local message store - send data to it */
	    if(message_store )
		if(-1==message_put_body(message_store,linebuf,l))
		    goto ERROR;


	    /* simply send linebuf to filter */
#if 1
	    if(!(st = sts_send_body(s,linebuf,l,0)))
		goto ERROR;
#else
	    for(i=0;i<l;i++)
	    if(!(st = sts_send_body(s,linebuf+i,1,STS_REINIT_TIMEOUT)))
		goto ERROR;
#endif
	    /* analyze result code */
	    switch(st->status)
		{
		default:
		    fprintf(stderr,"Unknown result code %d\n",st->status);
		    /* no break here */
		case STS_SENDERROR:
		    goto ERROR;
		case STS_CONTINUE:
		    /* do nothing */
		    break;
		case STS_ACCEPT:
		case STS_FILTER_ERROR:
		case STS_REJECT:
		case STS_BLACKHOLE:
		case STS_SMTP_ACCEPT:
		    goto ENDREAD1;
		}
	}
  ENDREAD1:

    if(!message_read && st && message_store)
	{
	    /* eat up remaining of message */
	    while(fgets(linebuf,sizeof(linebuf)-1,f))
		{
		    if(!strcmp(linebuf,".\n")) // single dot - end of DATA
			{
			    message_read =1;
			    break;
			}
		    /* in Accept case we'll need remaining message to
		       send to receiver */
		    if(st->status == STS_ACCEPT && message_store)
			{
			    l = strlen(linebuf);
			    if(-1==message_put_body(message_store,linebuf,l))
				goto ERROR;
			}
		}
	}
    else if(st->status == STS_CONTINUE)
	{
	    /* message read, we should get final status */
	    if(!(st = sts_body_end(s,0)))
		goto ERROR;
	}
    /* now we have message status in st, message body in message
     * store, let's go*/
    switch(st->status)
	{
	case STS_ACCEPT:
	    r = simple_reply(st,"ACCEPT: 220 OK MESSAGE ACCEPTED, reply follows");
	    if(multiple_messages_allowed)
		r = process_accept_multi(st,message_store);
	    else
		r = process_accept_single(st,message_store);
	    break;
	case STS_SMTP_ACCEPT:
	    r = simple_reply(st,"ACCEPT: 220 OK (Accepted by SMTP-mode filter)");
	    break;
	case STS_BLACKHOLE:
	    r = simple_reply(st,"ACCEPT: 220 OK (Blackholed)");
	    break;
	case STS_FILTER_ERROR:
	    r = simple_reply(st,"ERROR: 451 ERROR (Filter error)");
	    break;
	case STS_SENDERROR:
	    r = simple_reply(st,"SENDERROR: message not sent");
	    break;
	case STS_REJECT:
	    r = simple_reply(st,"REJECT: 550 ERROR (Rejected)");
	    break;
	default:
	    r = simple_reply(st,"ERROR: 451 Unexpected response code");
	    break;
	}
    
    if(r<0) 
	goto ERROR;
    if(message_store) message_free(message_store);
    return 0;
    

  ERROR:
    if(st->status == STS_SENDERROR)
	fprintf(stderr,"Filtering error: %s",sts_strerror(st->errorcode));
    if(message_store) message_free(message_store);
    return -1;

}

int process_accept_multi(spamtest_status_t *st,message_t *msg)
{
    /* go through all generated messages and output them to stdout */
    int i,j;
    message_status_t *ms;
    for(i=0;i<st->count;i++)
	{
	    ms = &st->ms_array[i];
	    if(ms->action == STS_ACTION_BOUNCE)
		{
		    /* generate new message */
		    printf("HELO: 127.0.0.1\n");
		    printf("MAIL FROM: %s\n",ms->mailfrom?ms->mailfrom:"unknown");
		    rcptlist_iterate(ms->rcpts,recipient_writer);
		    printf("DATA\n");
		    if(ms->bouncemsg)
			{
			    for(j=0;j<msgtext_count_chunks(ms->bouncemsg);j++)
				fwrite(msgtext_get_chunk(ms->bouncemsg,j),1,
				       msgtext_chunk_len(ms->bouncemsg,j),
				       stdout);

			}
		    else
			printf("Subject: autogenerated\n\nAutogenerated bounce body\n");
		    printf(".\n");
		}
	    else if(ms->action == STS_ACTION_CHANGE)
		{
		    headers_list_t *hdrs = make_changed_headers(ms,msg);
		    if(hdrs)
			{
			    printf("HELO: 127.0.0.1\n");
			    printf("MAIL FROM: %s\n",ms->mailfrom?ms->mailfrom:"unknown");
			    rcptlist_iterate(ms->rcpts,recipient_writer);
			    printf("DATA\n");
			    headers_iterate(hdrs,header_writer);
			    if(msg)
				message_body_callback(msg,body_writer);
			    printf(".\n");
			}
		}
	}
    return 0;
}

int process_accept_single(spamtest_status_t *st,message_t *msg)
{
    spamtest_status_t *glued = glue_actions(st);
    int r=0;
    if(glued)
	{
	    r = process_accept_multi(glued,msg);
	    return r;
	}
    return -1;
}

int simple_reply(spamtest_status_t *st,char *reply)
{
    printf("%s filter message: %s\n",reply,st->reason?st->reason:" ");
    return 0;
}


void header_writer(char *hdr,char *val)
{
    printf("%s: %s",hdr,val);
}

void recipient_writer(const char *rcpt,size_t unused)
{
    printf("RCPT TO: %s\n",rcpt);
}
void body_writer(const char *data,size_t len)
{
    fwrite(data,1,len,stdout);
}

 char *notspace(char *str)
{
    while(*str && isspace(*str))str++;
    return str;
}
char *get_tok(char *str, char** endp)
{
    char *p = notspace(str),*q;
    if(!p) return NULL;
    q = p;
    while(*q && !isspace(*q)) q++;
    if(q)
	{
	    *q='\0';
	    q = notspace(q+1);
	}
    if(endp)
	*endp = q?q:NULL;
    return p;
}
