/* -*- C -*-
 * Spamtest connection library
 * File: bufio.h
 *      (C)  Alex Tutubalin <lexa@lexa.ru>
 *	Apr-May 2003
 *
 * Created: Wed Apr 23 22:08:08 2003
 */

#ifndef __BUFIO_H
#define __BUFIO_H

#include "heap.h"
#include "spamtest.h"

typedef struct 
{
    unsigned char *buf;
    unsigned int  firstbyte;
    unsigned int  lastbyte;
    unsigned int  alloc;
    int	pending_error;
}buffer_t;

typedef struct __bufio_t
{
    int fd;
    int flags;
    buffer_t rbuf;
    buffer_t wbuf;
} bufio_t;

#define DEFAULT_BUFSIZE (64*1024)
/* select timeout in milliseconds */
#define BUFIO_READ_TIMEOUT  10000
#define BUFIO_WRITE_TIMEOUT  10000

#define BUFIO_OK	0
#define BUFIO_ERR	(1<<0)
#define BUFIO_EOF	(1<<1)
#define BUFIO_WAITDATA	(1<<2)
#define BUFIO_CANNOTWRT	(1<<3)
#define BUFIO_TIMEOUT	(1<<4)

// flags
#define BUFIO_USE_POLL  1

/* creates bufio_t data, rsize/wsize - size of internal buffers, if 0
 * specified default will be used. O_NONBLOCK will set on file descriptor */
bufio_t *bufio_init(msg_heap_t *heap, int file,int rsize,int wsize);
void bufio_setflags(bufio_t *,unsigned int);
unsigned int bufio_getflags(bufio_t *);

/* shortcut for reading one byte */
int bufio_getbyte(bufio_t *buf, unsigned char *byte, spamtest_session_t *s);
/* same semantics as read(fd,data,size) */
int bufio_read(bufio_t *buf, void *data,  int maxsize, spamtest_session_t *s);
/* flushes _entire_ write buffer, return BUFIO_OK on success, other
   value on error */
int bufio_flush(bufio_t *buf, spamtest_session_t *s);


/* same semantics as write(2), BUFIO_* error codes returned in
   buf->wptr.pending_error */
int bufio_write(bufio_t *buf, const void *data, int size, spamtest_session_t *s);
/* shortcut for write one byte, returns BUFIO_* error codes */
int bufio_putbyte(bufio_t *buf,unsigned char byte,spamtest_session_t *s);

#endif
