/* -*- C -*-
 * File: ap-sample.c
 *
 * Created: Thu Apr 24 15:10:23 2003
 */
#include <sys/types.h>
#include <sys/time.h>
#include <sys/uio.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <syslog.h>
#include "msgstore.h"
#include "spamtest.h"

#define DEF_FROM "user@domain.com"
#define DEF_TO   "user@domain.com"
#define DEF_RELAY "192.168.10.10"
#define CONNECT_TO 10000
#define RW_TO      10000

void usage(char *p)
{
    fprintf(stderr,
	    "Usage: %s [-f SMTP-from] [-t SMTP-to] [-i Relay-IP]\n"
            "          [-a attr1=value1 -a attr2=value2 .. ] tcp:spamtest-host:port message-file\n\n"
	    "  -f user@domain  - sets SMTP from (default is %s)\n"
	    "  -t user@domain  - sets SMTP to (default is %s)\n"
	    "  -i aa.bb.cc.dd  - sets Relay-IP address (default is %s)\n"
            "  -a attr=value   - sets attribute\n"
	    "  -m              - minimal output (only value of X-SpamTest-Status)\n"
	    "  tcp:spamtest-host:port\n"
	    "     or\n"
	    "  unix:/path/to/spamtest.socket - ap-mailfilter address\n"
	    "  message-file    - file in unix mailbox format (single message) or \"-\" for stdin\n",
	    p,DEF_FROM,DEF_TO,DEF_RELAY);
    exit(1);
}

void min_dump_status(spamtest_status_t *st)
{
    int i,j;
    message_status_t *ms;
    hdraction_t *hdr;
    if(!st) return;
    for(i=0;i<st->count;i++)
	{
	    ms = &st->ms_array[i];
	    if(ms->action == STS_ACTION_CHANGE)
		{
		    for(j=0;j<ms->action_count;j++)
			{
			    hdr = &ms->act_array[j];
			    if(!hdr) continue;
			    if(strcasecmp(hdr->header,"X-SpamTest-Status"))
				continue;
			    if(hdr->type == STS_HEADER_ADD || hdr->type == STS_HEADER_CHG
				|| hdr->type == STS_HEADER_PREPEND || hdr->type == STS_HEADER_NEW)
				{
				    printf("%s\n",hdr->value);
				    return;
				}
			}
		}
	}
    printf("NoStatus\n");
}
void dump_status(spamtest_status_t *st)
{
    int i,j;
    message_status_t *ms;
    hdraction_t *hdr;
    if(!st) return;
    for(i=0;i<st->count;i++)
	{
	    ms = &st->ms_array[i];
	    if(ms->action == STS_ACTION_BOUNCE)
		{
		    printf("  %d: Bounce action, body not printed\n",i+1);
		}
	    else if(ms->action == STS_ACTION_CHANGE)
		{
		    printf("  %d: Change headers:\n",i+1);
		    for(j=0;j<ms->action_count;j++)
			{
			    hdr = &ms->act_array[j];
			    if(!hdr) continue;
			    if(hdr->type == STS_HEADER_DEL)
				printf("DELETE: %s\n",hdr->header?hdr->header:"header field missing");
			    else if(hdr->type == STS_HEADER_CHG)
				printf("CHANGE: %s|%s\n",hdr->header?hdr->header:"header field missing",
				       hdr->value?hdr->value:"header value missing");
			    else if(hdr->type == STS_HEADER_PREPEND)
				printf("PREPEND: %s|%s\n",hdr->header?hdr->header:"header field missing",
				       hdr->value?hdr->value:"header value missing");
			    else if(hdr->type == STS_HEADER_ADD)
				printf("ADD: %s|%s\n",hdr->header?hdr->header:"header field missing",
				       hdr->value?hdr->value:"header value missing");
			    else if(hdr->type == STS_HEADER_NEW)
				printf("NEW: %s|%s\n",hdr->header?hdr->header:"header field missing",
				       hdr->value?hdr->value:"header value missing");
			    else
				printf("Unknown HDR change code %d\n",hdr->type);
			}
		}
	    else
		{
		    printf("  %d: Unknown action type\n",i+1);
		}
	}
}

void extract_attribute(char* attr_string, char** attr_name, char** attr_value)
{
    char *p = strstr(attr_string, "=");
    if(!p) return;
    p[0] = '\0';

    *attr_name = strdup(attr_string);
    *attr_value = strdup(p+1);
}

#define BUFSZ 20480
int main(int ac, char *av[])
{
    char *from = DEF_FROM;
    char *to = DEF_TO;
    char *relay = DEF_RELAY;
    char *attr_string = NULL;
    char *attr_name = NULL;
    char *attr_value = NULL;
    char *prog = av[0];
    char  buf[BUFSZ],ch;
    int r,stat;
    FILE *in;
    spamtest_session_t s;
    spamtest_status_t *st=NULL;
    int minimorum=0;
    if(ac<3) usage(prog);

    // start Spamtest communications
    {
        bzero(&s,sizeof(s));

        s.access_address = av[ac-2];
        s.mail_domain = NULL;
        s.rw_timeout= RW_TO;
        s.connect_timeout= CONNECT_TO;
        if((r=sts_init(&s))<0){
	    fprintf(stderr,"sts_init: %s\n",sts_strerror(r));
	    exit(2);
	}
    }


    while((ch = getopt(ac, av, "a:f:t:i:hm")) != -1)
	{
	    switch(ch)
		{
                case 'a':
                        attr_string = strdup(optarg);
                        attr_name = NULL; attr_value = NULL;
                        extract_attribute(attr_string, &attr_name, &attr_value);
                        if(!attr_name){
                                sts_close(&s);
                                usage(prog);
                        }
                        if((r=sts_add_attribute(&s, attr_name, attr_value))<0){
                                fprintf(stderr,"sts_add_atribute: %s\n",sts_strerror(r));
                                sts_close(&s);
                                exit(2);
                        }
                        break;
		case 'f': from = strdup(optarg); break;
		case 't': to = strdup(optarg); break;
		case 'i': relay = strdup(optarg); break;
		case 'm': minimorum++; break;
		case 'h':
		case '?':
		default:
                    sts_close(&s);
		    usage(prog);
		}
	}
		    
    ac-=optind;
    av+=optind;
    if(ac < 2 || !av[0] || !av[1]){
        sts_close(&s);
	usage(prog);
    }

    if(av[1][0] == '-')
	in = stdin;
    else
	in = fopen(av[1],"r");
    if(!in)
	{
	    perror("fopen"); sts_close(&s); exit(1);
	}

    if((r=sts_prepare_envelope(&s,from,to,relay,&stat))<0)
	{
	    fprintf(stderr,"sts_prepare_envelope: %s\n",sts_strerror(r));
            sts_close(&s);
	    exit(2);
	}
    // read and send
    while((r=fread(buf,1,BUFSZ,in))>0)
	{
	    if(!(st=sts_send_body(&s,buf,r,STS_REINIT_TIMEOUT)))
		{
		    fprintf(stderr,"sts_send_body: got NULL\n");
                    sts_close(&s);
		    exit(2);
		}
	    if(st->status!=STS_CONTINUE)
		break;
	}
    if(st->status == STS_CONTINUE)
	if(!(st = sts_body_end(&s,STS_REINIT_TIMEOUT)))
		{
		    fprintf(stderr,"sts_body_end: got NULL\n");
                    sts_close(&s);
		    exit(2);
		}
    if(minimorum)
	{
	    if(st->status == STS_ACCEPT)
		{
		    min_dump_status(st);
                    sts_close(&s);
		    exit(0);
		}
	    else
		{
		    printf("UnexpectedStatus\n");
                    sts_close(&s);
		    exit(3);
		}
	}
    else
	{
	    switch(st->status)
		{
		default:
		    fprintf(stderr,"Unknown result code %d\n",st->status);
		    break;
		case STS_SENDERROR:
		    printf("Filter I/O ERROR\n"); 
		    break;
		case STS_CONTINUE:
		    fprintf(stderr,"Unexpected STS_CONTINUE\n");
		    break;
		case STS_FILTER_ERROR:
		    printf("Filter ERROR\n"); 
		    break;
		case STS_REJECT:
		    printf ("Message Rejected\n");
		    break;
		case STS_BLACKHOLE:
		    printf ("Message Blackholed\n");
		    break;
		case STS_SMTP_ACCEPT:
		    printf ("Message Accepted by SMTP-mode filter\n");
		    break;
		case STS_ACCEPT:
		    printf ("Message Accepted\n");
		    dump_status(st);
		    break;
		}
	}
    sts_close(&s);
    exit(0);
}

