/* -*- C -*-
 * Spamtest connection library
 * File: heap.c
 *      (C)  Alex Tutubalin <lexa@lexa.ru>
 *	Apr-May 2003
 *
 * Created: Wed Apr 23 22:08:08 2003
 */
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include "heap.h"

#ifndef DEFAULT_ALLOC
#define DEFAULT_ALLOC 1024000
#endif
#ifndef HEAP_ALLOC_ALIGN
#define HEAP_ALLOC_ALIGN 4
#endif

#if ( defined (_THREAD_SAFE) || defined (_REENTRANT) ) && defined __USE_LOCKING
#include <pthread.h>
static pthread_mutex_t heap_mutex;

void heap_library_init(void)
{
    pthread_mutex_init(&heap_mutex,NULL);
}

static void __heap_lock(void)
{
    pthread_mutex_lock(&heap_mutex);
}
static void __heap_unlock(void)
{
    pthread_mutex_unlock(&heap_mutex);
}
#else
static void __heap_unlock(void){}
static void __heap_lock(void){}
void heap_library_init(void){}

#endif

char *heap_pstrcat(msg_heap_t* heap,...)
{
    int len=0;
    char *cp,*argp,*res;
    va_list adummy;
    __heap_lock();
    va_start(adummy, heap);
    while ((cp = va_arg(adummy, char *)) != NULL)
	len += strlen(cp);
    __heap_unlock();
    res = heap_alloc(heap,len+1);
    if(!res) 
	{
	    return NULL;
	}
    cp = res;
    *cp ='\0';
    __heap_lock();
    va_start(adummy, heap);
    while ((argp = va_arg(adummy, char *)) != NULL) {
        strcpy(cp, argp);
        cp += strlen(argp);
    }
    va_end(adummy);
    __heap_unlock();
    return res;
}

void heap_free(msg_heap_t *heap)
{
    msg_heap_t *hp = heap;
    if(!heap) return;
    __heap_lock();
    while(heap)
	{
	    hp = heap->next;
	    if(heap->data)
		free(heap->data);
	    free(heap);
	    heap = hp;
	}
    __heap_unlock();
}

msg_heap_t *heap_init(size_t initsize)
{
    msg_heap_t *ret;
    __heap_lock();
    ret = malloc(sizeof(msg_heap_t));
    __heap_unlock();
    if(!ret) 
	{
	    return NULL;
	}
    bzero(ret,sizeof(*ret));
    __heap_lock();
    ret->data = malloc(ret->_alloc = initsize?initsize:DEFAULT_ALLOC);
    __heap_unlock();
    if(!ret->data)
	    return NULL;
//    bzero(ret->data,ret->_alloc);
    return ret;
}

void *heap_allocz(msg_heap_t *heap,unsigned int n)
{
    char *ret = heap_alloc(heap,n);
    if(ret)
	bzero(ret,n);
    return ret;
}

void *heap_alloc(msg_heap_t *heap,unsigned int n)
{
    // find place to alloc
    char *ret;
    int newalloc = DEFAULT_ALLOC;
    msg_heap_t *h = heap;
    n=((n+HEAP_ALLOC_ALIGN-1)/HEAP_ALLOC_ALIGN)*HEAP_ALLOC_ALIGN;
    while(h)
	{
	    if(h->_used+n<h->_alloc)
		{
		    ret = h->data+h->_used;
		    h->_used+=n;
//		    bzero(ret,n);
		    return (void*)ret;
		}
	    if(h->next)
		{
		    h=h->next;
		    if(newalloc < DEFAULT_ALLOC*3)
			newalloc*=2;
		}
	    else
		{
		    // allocate new
		    h->next = heap_init(n>newalloc?n*2:newalloc);
		    if(!h->next) return NULL;
		    return heap_alloc(h->next,n);
		}
	}
    /* can happen if heap not initialized */
    return NULL;
}

void *heap_reallocz(msg_heap_t *heap,const void *oldp, unsigned int oldsz, unsigned int newsz)
{
    char *ret = heap_realloc(heap,oldp,oldsz,newsz);
    if(ret && (newsz>oldsz))
    if(ret)
	bzero(ret+oldsz,newsz-oldsz);
    return ret;
}


void *heap_realloc(msg_heap_t *heap,const void *oldp, unsigned int oldsz, unsigned int newsz)
{
    char *ret;
    if(!newsz || !heap || oldsz<0) return NULL;
    if(newsz<=oldsz)
	return (void*)oldp;
    ret = heap_alloc(heap,newsz);
    if(ret && oldp && oldsz)
	{
	    __heap_lock();
	    memmove(ret,oldp,oldsz);
	    __heap_unlock();
	}
    return ret;
}


void *heap_strdupn(msg_heap_t *heap, const char *str,unsigned int len)
{
    char *ret = heap_alloc(heap,len+1);
    if(ret)
	{
	    __heap_lock();
	    memmove(ret,str,len+1);
	    __heap_unlock();
	}
    return ret;

}

void *heap_strdup(msg_heap_t *heap, const char *str)
{
    return heap_strdupn(heap,str,strlen(str));
}

