/* -*- C -*-
 * File: kas_client_sample.c
 *
 * Created: Sat Oct  2 23:23:56 2004
 */

#include <stdio.h>
#include <stdlib.h>
#include "kas_defs.h"
#include "kas_client.h"

char *connect_addr = "tcp:localhost:2255";

//#define TEST_MANY_RCPTS
#undef TEST_MANY_RCPTS
#define NRCPT 1000

void my_progress_cb(kas_session_t* _session, kas_progress_info_t *pp,  void* param)
{
    switch(pp->state)
	{
	case KAS_INFO_STARTSEND:
	    printf("Progress: sending message to engine (pid=%d)\n",pp->pid);
	    break;
	case KAS_INFO_WAITSTATUS:
	    printf("Progress: waiting for message status (pid=%d)\n",pp->pid);
	    break;
	case KAS_INFO_DONE:
	    printf("===== MESSAGE SCANNED =====\n");
	    switch(pp->message_status)
		{
		case KAS_EINFO_REJECT:
		    printf("Progress: Message rejected (pid=%d)\n",pp->pid); break;
		case KAS_EINFO_BLACKHOLE:
		    printf("Progress: Message blackholed (pid=%d)\n",pp->pid); break;
		case KAS_EINFO_BOUNCE:
		    printf("Progress: Message bounced by recipient %s (pid=%d)\n",
			   pp->rcpt?pp->rcpt:"not specified",
			   pp->pid); 
		    break;
		case KAS_EINFO_ACCEPT_TRUSTED:
		case KAS_EINFO_ACCEPT_CLEAN:
		    printf("Progress: Message clean for recipient %s (pid=%d)\n",
			   pp->rcpt?pp->rcpt:"not specified",
			   pp->pid); 
		    break;
		case KAS_EINFO_ACCEPT_PROBABLESPAM:
		case KAS_EINFO_ACCEPT_SPAM:
		    printf("Progress: Message is %s, recipient=%s (pid=%d)\n",
			   pp->message_status==KAS_EINFO_ACCEPT_SPAM?"\'SPAM\'":"\'Probable Spam\'",
			   pp->rcpt?pp->rcpt:"not specified",
			   pp->pid); 
		    printf("Progress: Methods: %s\n",pp->detection_methods?pp->detection_methods:"not specified");
		    printf("Progress: Categories: %s\n",pp->content_categories?pp->content_categories:"not specified");
		    printf("Progress: ExtdInfo: %s\n",pp->extended_info?pp->extended_info:"not specified");
		    break;
		    
		default:
		    printf("Progress: Message scanned, status=%d (pid=%d)\n",pp->message_status,pp->pid); break;
		}
	    break;
	case KAS_INFO_ERROR:
	    printf("Message scan error (pid=%d)\n",pp->pid);
	    break;
	default:
	    printf("Unknown progress status: %d (pid=%d)\n",pp->state,pp->pid);
	    break;
	}
}


void my_status_cb(kas_session_t* _session, kas_status_t *old_status,
		  kas_status_t *new_status,
		  void* param)
{
    int *count = (int*)param;
    // only some states are processed!
    switch(new_status->status)
	{
	case KAS_STATUS_WANTRESTART:
	    if(*count>10)
		{
		    printf("STATUS: Restarting resource-consuming engine for pid %d after %d scans\n",
			   new_status->scaner_pid,  *count);
		    *count=0;
		    KasSDisconnect(_session);
		    KasSConnect(_session,connect_addr,10000);
		    KasSWaitIdle(_session,4000);
		}
	    break;
	case KAS_STATUS_IOERROR:
	case KAS_STATUS_CONNECTERROR:
	    printf("STATUS: Network error for pid %d\n",new_status->scaner_pid);
	    break;
	case KAS_STATUS_SCAN_END:
	    printf("STATUS: Message scaned by pid %d\n",new_status->scaner_pid);
	    (*count)++;
	    break;
	case KAS_STATUS_IDLE:
	    printf("STATUS: PID %d ready\n",new_status->scaner_pid);
	    break;
	default:
	    printf("Event %d\n",new_status->status);
	    break;
	}
}




int main(int ac, char *av[])
{

    int should_reconnect;
#ifndef TEST_MANY_RCPTS
    char *rcptlist[]=	{"rcpt1@example.com","rcpt2@example.com","rcpt3@example.com"};
    kas_attribute_t attrs[]= {{"attr1","value1"},{"client_address","11.12.12.15"},{"attr3","value3"}};
    kas_metadata_t md={
	"192.168.0.1", /* relay IP */
	"example.com", /* HELO */
	"sender@example.com", /* sender*/
	3, /* rcpt count */
	rcptlist,
	3, /* attr count */
	attrs
    };
#else
    char **rcptlist;
    int i;
    kas_metadata_t md={
	"192.168.0.1", /* relay IP */
	"example.com", /* HELO */
	"sender@example.com", /* sender*/
	0, /* rcpt count */
	NULL
    };
#endif
    int  scan_count=0;
    int repeats = 1;
    kas_session_t *s;
    int rc;


#ifdef TEST_MANY_RCPTS
    rcptlist=malloc(NRCPT*sizeof(char*));
    for(i=0;i<NRCPT;i++)
	{
	    rcptlist[i]=malloc(100);
	    snprintf(rcptlist[i],100,"rcpd%d@example.com",i);
	}
    
    md.rcptlist = rcptlist;
    md.nrcpts=NRCPT;
#endif
    if(ac<2)
	{
	    printf("Usage: %s message-file [number-of-repetitions]\n",av[0]);
	    exit(1);
	}
    if(!(s= KasSCreateSession(NULL, /* data will be allocated */
			      81920, /* buffer size */
			      5000, /* read timeout in milliseconds */
			      5000, /* write timeout in milliseconds */
			      500*1024, /* larger mail will be  filtered as 'not spam' because of size */
			      26*1024, /* filter engine size limit (kilobytes) */
			      my_status_cb, /* status callback */
			      &scan_count) /* cb's data */
	   ))
	{
	    printf("Cannot create session");
	    exit(1);
	}
    KasSConnect(s,connect_addr,5000);
    if(KASerrOK!=KasSWaitIdle(s,10000))
	{
	    printf("Cannot connect\n");
	    exit(1);
	}
    KasSPing(s);
    if(KASerrOK!=KasSWaitIdle(s,10000))
	{
	    printf("Cannot ping\n");
	    exit(1);
	}
    if(ac>2)
	repeats = atoi(av[2]);
    
    while(repeats--)
	{
	     should_reconnect = s->standard_proto;
	    printf("Start File\n");
	    rc = KasSScanFile(s,&md,av[1],my_progress_cb,NULL);
	    if (rc!=KASerrOK && rc!=KASerrIN_PROGRESS)
		{
		    printf("Cannot scan file %s\n",av[1]);
		    exit(1);
		}
	    if(rc == KASerrIN_PROGRESS)
		{
		    if(KASerrOK!=(rc=KasSWaitIdle(s,6000)))
			{
			    printf("Cannot scan file within 6 seconds(%d)\n",rc);
			    exit(1);
			}
		}
	    if(should_reconnect)
		{
		    if(repeats)
			{
			    KasSConnect(s,connect_addr,5000);
			    if(KASerrOK!=KasSWaitIdle(s,10000))
				{
				    printf("Cannot connect\n");
				    exit(1);
				}
			}
		}
	}
    if(should_reconnect)
	KasSDisconnect(s);
    KasSSessionDestroy(s);
    return 0;
}

