/* -*- C -*-
 * File: kas_client_sample.c
 *
 * Created: Sat Oct  2 23:23:56 2004
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include "kas_defs.h"
#include "kas_mclient.h"

char *connect_addr;

//#define TEST_MANY_RCPTS
#undef TEST_MANY_RCPTS
#define NRCPT 1000


void my_progress_cb(kas_multisession_t* _session, unsigned int index,
		    kas_progress_info_t *pp,  void* param)
{
    char **filenames=(char**)param;
    char *file;

    if(filenames && filenames[index])
	file = filenames[index];
    else
	file="Unknown";

    switch(pp->state)
	{
	case KAS_INFO_STARTSEND:
	    printf("Progress: sending file %s to engine (pid=%d, index=%d)\n",file,pp->pid,index);
	    break;
	case KAS_INFO_WAITSTATUS:
	    printf("Progress: waiting for file %s status (pid=%d, index=%d)\n",file,pp->pid,index);
	    break;
	case KAS_INFO_DONE:
	    printf("===== FILE %s SCANNED by pid=%d (index=%d), status=%d =====\n",
		   file,pp->pid,index,pp->message_status);
	    switch(pp->message_status)
		{
		case KAS_EINFO_REJECT:
		    printf("Progress: Message %s rejected",file); break;
		case KAS_EINFO_BLACKHOLE:
		    printf("Progress: Message %s blackholed",file); break;
		case KAS_EINFO_BOUNCE:
		    printf("Progress: Message %s bounced by recipient %s\n",file,
			   pp->rcpt?pp->rcpt:"not specified");
		    break;
		case KAS_EINFO_ACCEPT_TRUSTED:
		case KAS_EINFO_ACCEPT_CLEAN:
		    printf("Progress: Message %s clean for recipient %s\n",file,
			   pp->rcpt?pp->rcpt:"not specified");
		    break;
		case KAS_EINFO_ACCEPT_PROBABLESPAM:
		case KAS_EINFO_ACCEPT_SPAM:
		    printf("Progress: Message %s is %s, recipient=%s\n",file,
			   pp->message_status==KAS_EINFO_ACCEPT_SPAM?"\'SPAM\'":"\'Probable Spam\'",
			   pp->rcpt?pp->rcpt:"not specified");
		    printf("Progress: Methods: %s\n",pp->detection_methods?pp->detection_methods:"not specified");
		    printf("Progress: Categories: %s\n",pp->content_categories?pp->content_categories:"not specified");
		    printf("Progress: ExtdInfo: %s\n",pp->extended_info?pp->extended_info:"not specified");
		    break;
		    
		default:
		    printf("Progress: File %s scanned, status=%d (pid=%d, index=%d)\n",
			   file,pp->message_status,pp->pid,index); break;
		}
	    break;
	case KAS_INFO_ERROR:
	    printf("File %s scan error (pid=%d, index=%d)\n",file,pp->pid,index);
	    break;
	default:
	    printf("Unknown progress status: %d (pid=%d, index=%d)\n",pp->state,pp->pid,index);
	    break;
	}
}


void my_status_cb(kas_multisession_t* _session, unsigned int index, kas_status_t *old_status,
		  kas_status_t *new_status,
		  void* param)
{
    int *count = (int*)param;
    // only some states are processed!
    switch(new_status->status)
	{
	case KAS_STATUS_WANTRESTART:
	    if(count[index]>10)
		{
		    printf("STATUS: Restarting resource-consuming engine for pid %d after %d scans\n",
			   new_status->scaner_pid,  *count);
		    count[index]=0;
		    KasMDisconnect(_session,index);
		    KasMConnect(_session,index,connect_addr,10000);
		    KasMWaitIdle(_session,1,4000);
		}
	    break;
	case KAS_STATUS_IOERROR:
	case KAS_STATUS_CONNECTERROR:
	    printf("STATUS: Network error for pid %d (index=%d)\n",new_status->scaner_pid,index);
	    break;
	case KAS_STATUS_SCAN_END:
	    printf("STATUS: Message scaned by pid %d (index=%d), %d.%d elapsed\n",new_status->scaner_pid,index,
		   (int)new_status->elapsed_time.tv_sec,(int)new_status->elapsed_time.tv_usec);
	    (*count)++;
	    break;
	case KAS_STATUS_IDLE:
	    printf("STATUS: PID %d ready (index=%d)\n",new_status->scaner_pid,index);
	    break;
	default:
//	    printf("Event %d\n",new_status->status);
	    break;
	}
}



int main(int ac, char *av[])
{

    char *rcptlist[]=	{"rcpt1@example.com","rcpt2@example.com","rcpt3@example.com"};
    kas_metadata_t md={
	"192.168.0.1", /* relay IP */
	"example.com", /* HELO */
	"sender@example.com", /* sender*/
	3, /* rcpt count */
	rcptlist
    };
    int messages = 1;
    char *progname = av[0];
    int sessions = 1;
    kas_multisession_t *s;
    int rc, index,i;
    int *scan_counts;
    char **filenames;

    while((rc = getopt(ac, av, "s:r:"))!=-1)
	{
	    switch(rc)
		{
		case 's': sessions=atoi(optarg); break;
		case 'r': messages=atoi(optarg); break;
		}
	}
    ac-=optind;
    av+=optind;
    if(ac<1 || sessions<0 || messages<0)
	{
	    printf("Usage: %s [-r repetitons] [-s parallel sessions]  msgfile1 msgfile2.... \n",progname);
	    exit(1);
	}
    scan_counts = calloc(sessions,sizeof(int));
    filenames = calloc(sessions,sizeof(char*));
    connect_addr= "tcp:localhost:2255";

    if(!(s= KasMCreateSession(sessions, /* session count */
			      81920, /* buffer size */
			      15000, /* read timeout in milliseconds */
			      5000, /* write timeout in milliseconds */
			      500*1024, /* larger mail will be  filtered as 'not spam' because of size */
			      26*1024, /* filter engine size limit (kilobytes) */
			      my_status_cb, /* status callback */
			      scan_counts) /* cb's data */
	   ))
	{
	    fprintf(stderr,"Cannot create multisession");
	    exit(1);
	}

    for(i=0;i<sessions;i++)
	{
	    rc=KasMConnect(s,i,connect_addr,10000);
	    if(rc!=KASerrOK && rc!=KASerrIN_PROGRESS)
		{
		    fprintf(stderr,"Cannot connect session %d\n",i);
		    exit(1);
		}
	}
    // wait for one third of all scanners to become IDLE (ready)
    if(KASerrOK!=KasMWaitIdle(s,
			      sessions>3?sessions/3:1,
			      10000))
	{
	    printf("Cannot connect\n");
	    exit(1);
	}
    printf("At some scanners started\n");

    while(messages)
	{
	    char *filename;
	    filename = av[messages%ac];
	    printf("KasMCountScanners: %d ready\n",KasMCountScanners(s,KAS_STATUS_IDLE));
	    index = KasMPeekScanner(s,KAS_STATUS_IDLE);
	    if(index >= 0)
		{
		    filenames[index]=filename;
		    if(messages % 2) 
			rc = KasMScanFile(s,index,&md,filename,my_progress_cb,filenames);
		    else
			{
			    struct stat st;
			    int file_fd;
			    if(stat(filename,&st))
				perror("Cannot stat");
			    else if((file_fd=open(filename,O_RDONLY))<0)
				perror("Cannot open");
			    else
				{
				    rc = KasMScanOpenFile(s,index,&md,file_fd,st.st_size,my_progress_cb,filenames);
				    
				    close(file_fd); //ScanOpenFile
						    //will dup()
						    //passed fd
				}
			}
		    if (rc!=KASerrOK && rc!=KASerrIN_PROGRESS)
			{
			    printf("Cannot scan file %s\n",filename);
			}
		    messages--; // message submitted
		}
	    /* do network events for 300 msec */
	    KasMProcessEvents(s,300);

	    /* == do all other Network I/O here == */

	    // try to restart disconnected scanners if any
	    index = KasMPeekScanner(s,KAS_STATUS_DISCONNECTED);
	    if(index>=0)
		{
		    printf("Scaner %d disconnected, restarting\n",index);
		    KasMConnect(s,index,connect_addr,10000);
		}
	}

    /* waiting for completion of all scans */
    KasMWaitIdle(s,sessions,10000);
    
    /* disconnect them */
    for(i=0;i<sessions;i++)
    	KasMDisconnect(s,i);
    KasMSessionDestroy(s);
    return 0;
}

