/* -*- C -*-
 * Spamtest connection library
 * File: mess822.c
 *      (C)  Alex Tutubalin <lexa@lexa.ru>
 *	Apr-May 2003
 *
 * Created: Wed Apr 23 22:08:08 2003
 */
#include "mess822.h"

#include <string.h>
#include <stdlib.h>

extern unsigned int byte_chr();
extern unsigned int byte_rchr();
extern void byte_copy();
extern void byte_copyr();
extern int byte_diff();
extern void byte_zero();

#define byte_equal(s,n,t) (!byte_diff((s),(n),(t)))

extern unsigned int str_rchr();
#define str_equal(s,t) (!strcmp((s),(t)))
#define str_len(s) strlen(s)

#define alloc(n) ((char*)malloc(n))
#define alloc_free(p) free(p);

//extern /*@null@*//*@out@*/char *alloc();
//extern void alloc_free();
extern int alloc_re(char **xx, int old, int new);


#define stralloc_0(sa) stralloc_append(sa,"")

static int docomment(stralloc *out,stralloc *comment)
{
  int i;
  int j;
  char ch;

  for (j = i = 0;j < comment->len;++j) {
    ch = comment->s[j];
    if (ch == ' ') if (!i || (comment->s[i - 1] == ' ')) continue;
    comment->s[i++] = ch;
  }
  while (i && (comment->s[i - 1] == ' ')) --i;
  comment->len = 0;

  if (i) {
    if (!stralloc_0(out)) return 0;
    if (!stralloc_catb(out,comment->s,i)) return 0;
    if (!stralloc_append(out,"(")) return 0;
  }

  return 1;
}

static int doit(stralloc *out,int *state,stralloc *addr,stralloc *comment)
{
  if (!*state) return 1;

  if (!stralloc_0(out)) return 0;
  if (*state == 1)
    if (!stralloc_append(out,"@")) return 0;
  if (!stralloc_catb(out,addr->s,addr->len)) return 0;
  if (!stralloc_append(out,"+")) return 0;

  if (!docomment(out,comment)) return 0;

  *state = 0;
  addr->len = 0;
  return 1;
}

static int addcomment(char *tok,stralloc *comment)
{
  int i;

  if (*tok == ',') return 1;
  if (*tok == ':') return 1;
  if (*tok == ';') return 1;
  if (*tok == '<') return 1;
  if (*tok == '>') return 1;

  if ((*tok == '=') || (*tok == '(')) ++tok;

  i = str_len(tok);
  while (i--)
    if (!stralloc_append(comment,tok + i)) return 0;

  return 1;
}

static int addaddr(char *tok,int *state,stralloc *addr,stralloc *comment)
{
  int i;

  if ((*tok != '=') && (*tok != '.') && (*tok != '@'))
    return addcomment(tok,comment);

  if (!*state) *state = 1;
  if (*tok == '@') *state = 2;

  if (*tok == '=') ++tok;

  i = str_len(tok);
  while (i--)
    if (!stralloc_append(addr,tok + i)) return 0;

  return 1;
}


#define RETURN(x)		\
do{				\
CLEAN_SA(tokens);		\
CLEAN_SA(addr);			\
CLEAN_SA(comment);		\
    return x;			\
}while(0);



int mess822_addrlist(stralloc *out,char *in)
{
  int flagwordok = 1;
  int flagphrase = 0;
  int j;
  int i;
  char ch;
  int state;
  stralloc tokens={0,0,0};
  stralloc addr={0,0,0};
  stralloc comment={0,0,0};

  CLEAN_SA(tokens);
  CLEAN_SA(addr);
  CLEAN_SA(comment);
  state = 0;
//  stralloc comment;

//  comment = {0};
  
  if (!mess822_token(&tokens,in)) RETURN(0);

  if (!stralloc_copys(out,"")) RETURN(0);
  if (!stralloc_copys(&comment,"")) RETURN(0);
  if (!stralloc_copys(&addr,"")) RETURN(0);
  state = 0;

  j = tokens.len;

  while (j) {
    while (j--) if (!j || !tokens.s[j - 1]) break;

    ch = tokens.s[j];
    if (flagphrase) {
      if ((ch != ',') && (ch != ';') && (ch != ':') && (ch != '>')) {
        if (!addcomment(tokens.s + j,&comment)) RETURN(0);
        continue;
      }
      if (!doit(out,&state,&addr,&comment)) RETURN(0);
      flagphrase = 0;
      flagwordok = 1;
    }

    switch(tokens.s[j]) {
      case ' ': case '\t':
        if (!addcomment(" ",&comment)) RETURN(0);
        break;

      case ';': case ',':
        if (!doit(out,&state,&addr,&comment)) RETURN(0);
        if (!docomment(out,&comment)) RETURN(0);
        flagwordok = 1;
        break;

      case '(':
        if (!addcomment(" ",&comment)) RETURN(0);
        if (!addcomment(tokens.s + j,&comment)) RETURN(0);
        if (!addcomment(" ",&comment)) RETURN(0);
        break;

      case '=':
        if (!flagwordok)
          if (!doit(out,&state,&addr,&comment)) RETURN(0);
        if (!addaddr(tokens.s + j,&state,&addr,&comment)) RETURN(0);
        flagwordok = 0;
        break;

      case '>':
        if (!doit(out,&state,&addr,&comment)) RETURN(0);

        if (!state) state = 1; /* <> is an address */

        for (;;) {
          if (!addaddr(tokens.s + j,&state,&addr,&comment)) RETURN(0);
          if (!j) break;
          while (j--) if (!j || !tokens.s[j - 1]) break;
          if (tokens.s[j] == ':') break;
          if (tokens.s[j] == '<') break;
        }

        if (tokens.s[j] == ':')
          for (;;) {
            if (!addcomment(tokens.s + j,&comment)) RETURN(0);
            if (!j) break;
            while (j--) if (!j || !tokens.s[j - 1]) break;
            if (tokens.s[j] == '<') break;
          }
	/* fall through */

      case ':':
        flagphrase = 1;
        break;

      default:
        if (!addaddr(tokens.s + j,&state,&addr,&comment)) RETURN(0);
        flagwordok = 1;
        break;
    }
  }

  if (!doit(out,&state,&addr,&comment)) RETURN(0);
  if (!docomment(out,&comment)) RETURN(0);

  i = 0;
  j = out->len - 1;

  while (i < j) {
    ch = out->s[i];
    out->s[i] = out->s[j];
    out->s[j] = ch;
    ++i;
    --j;
  }

  RETURN(1);
}


#if 1
// QUOTE
static int needquote(buf,len)
char *buf;
int len;
{
  int i;
  char ch;

  if (!len) return 1;
  if (buf[0] == '.') return 1;
  if (buf[len - 1] == '.') return 1;

  for (i = 0;i < len - 1;++i)
    if ((buf[i] == '.') && (buf[i + 1] == '.')) return 1;

  for (i = 0;i < len;++i) {
    ch = buf[i];
    if (ch < 33) return 1;
    if (ch > 126) return 1;
    if (ch == '@') return 1;
    if (ch == '<') return 1;
    if (ch == '>') return 1;
    if (ch == '[') return 1;
    if (ch == ']') return 1;
    if (ch == '(') return 1;
    if (ch == ')') return 1;
    if (ch == ',') return 1;
    if (ch == ';') return 1;
    if (ch == ':') return 1;
    if (ch == '"') return 1;
    if (ch == '\\') return 1;
  }

  return 0;
}

static int doit2(out,buf,len,pre,post)
stralloc *out;
char *buf;
int len;
char *pre;
char *post;
{
  char ch;

  if (!stralloc_cats(out,pre)) return 0;

  while (len--) {
    ch = *buf++;
    if (ch == '\n') ch = 0;
    if ((ch == 0) || (ch == '\r') || (ch == '"') || (ch == '\\') || (ch == '[') || (ch == ']'))
      if (!stralloc_append(out,"\\")) return 0;
    if (!stralloc_append(out,&ch)) return 0;
  }

  if (!stralloc_cats(out,post)) return 0;
  return 1;
}

int mess822_quoteplus(stralloc *out,char *addr,char *comment)
{
  int i;
  char *quote;
  int flagempty;
  int flagbracket;

  flagempty = 0;
  if (str_equal(addr,"")) flagempty = 1;
  if (str_equal(addr,"@")) flagempty = 1;

  flagbracket = flagempty;

  if (comment) {
    if (!doit2(out,comment,str_len(comment),"\"","\" ")) return 0;
    flagbracket = 1;
  }

  if (flagbracket) if (!stralloc_cats(out,"<")) return 0;

  if (!flagempty) {
    i = str_rchr(addr,'@');
    quote = needquote(addr,i) ? "\"" : "";
    if (!doit2(out,addr,i,quote,quote)) return 0;

    addr += i;
    if (*addr == '@') ++addr;

    i = str_len(addr);
    if (i) {
      if (!stralloc_append(out,"@")) return 0;

      quote = needquote(addr,i) ? "\"" : "";

      if (*quote && (i >= 2) && (addr[0] == '[') && (addr[i - 1] == ']')) {
        if (!doit2(out,addr + 1,i - 2,"[","]")) return 0;
      }
      else
        if (!doit2(out,addr,i,quote,quote)) return 0;
    }
  }

  if (flagbracket) if (!stralloc_cats(out,">")) return 0;

  return 1;
}

int mess822_quote(stralloc *out,char *addr,char *comment)
{
  if (!stralloc_copys(out,"")) return 0;
  return mess822_quoteplus(out,addr,comment);
}

int mess822_quotelist(stralloc *out,stralloc *in)
{
  int i;
  int j;
  int comment;

  if (!stralloc_copys(out,"")) return 0;

  comment = 0;

  for (j = i = 0;j < in->len;++j)
    if (!in->s[j]) {
      if (in->s[i] == '(') {
	if (comment)
          if (!doit2(out,in->s + comment,str_len(in->s + comment),"\"","\": ;,\n  ")) return 0;
	comment = i + 1;
      }
      else if (in->s[i] == '+') {
	if (!mess822_quoteplus(out,in->s + i + 1,comment ? in->s + comment : (char *) 0)) return 0;
	if (!stralloc_cats(out,",\n  ")) return 0;
	comment = 0;
      }
      i = j + 1;
    }

  if (comment)
    if (!doit2(out,in->s + comment,str_len(in->s + comment),"\"","\": ;,\n  ")) return 0;

  if (out->len && (out->s[out->len - 1] == ' ')) --out->len;
  if (out->len && (out->s[out->len - 1] == ' ')) --out->len;
  if (out->len && (out->s[out->len - 1] == '\n')) --out->len;
  if (out->len && (out->s[out->len - 1] == ',')) --out->len;

  return 1;
}


int mess822_token(stralloc *out, char* in)
{
  char ch;
  int level;

  if (!stralloc_copys(out,"")) return 0;

  for (;;)
    switch(ch = *in++) {
      case 0:
        return 1;

      case '"':
        if (!stralloc_append(out,"=")) return 0;
        while (*in) {
          ch = *in++;
          if (ch == '"') break;
          if (ch == '\\') if (*in) ch = *in++;
          if (!stralloc_append(out,&ch)) return 0;
        }
        if (!stralloc_0(out)) return 0;
        break;

      case '[':
        if (!stralloc_append(out,"=")) return 0;
        if (!stralloc_append(out,"[")) return 0;
        while (*in) {
          ch = *in++;
          if (ch == ']') break;
          if (ch == '\\') if (*in) ch = *in++;
          if (!stralloc_append(out,&ch)) return 0;
        }
        if (!stralloc_append(out,"]")) return 0;
        if (!stralloc_0(out)) return 0;
        break;
  
      case '(':
        if (!stralloc_append(out,"(")) return 0;
        level = 1;
        while (*in) {
          ch = *in++;
          if (ch == ')') {
            --level;
            if (!level) break;
            if (!stralloc_append(out,")")) return 0;
            continue;
          }
          if (ch == '(') {
	    if (level) if (!stralloc_append(out,"(")) return 0;
            ++level;
            continue;
          }
          if (ch == '\\') if (*in) ch = *in++;
          if (!stralloc_append(out,&ch)) return 0;
        }
        if (!stralloc_0(out)) return 0;
        break;
  
      case '<': case '>': case ',': case ';': case ':':
      case '@': case '.':
      case ' ': case '\t':
        if (!stralloc_append(out,&ch)) return 0;
        if (!stralloc_0(out)) return 0;
        break;
  
      default:
        if (!stralloc_append(out,"=")) return 0;
  
        for (;;) {
          if (ch == '\\') if (*in) ch = *in++;
          if (!stralloc_append(out,&ch)) return 0;
          ch = *in;
          if (!ch) break;
          if (ch == '"') break;
          if (ch == '[') break;
          if (ch == '(') break;
          if (ch == '<') break;
          if (ch == '>') break;
          if (ch == ',') break;
          if (ch == ';') break;
          if (ch == ':') break;
          if (ch == '@') break;
          if (ch == '.') break;
          if (ch == ' ') break;
          if (ch == '\t') break;
          ++in;
        }

        if (!stralloc_0(out)) return 0;
      break;
    }
}
#endif 

int alloc_re(char **xx, int old, int new)
{
    char *y = (char*)malloc(new);
    if(!y) return 0;
    memmove(y,*xx,old);
    free(*xx);
    *xx=y;
    return 1;
}

unsigned int byte_chr(s,n,c)
char *s;
register unsigned int n;
int c;
{
  register char ch;
  register char *t;

  ch = c;
  t = s;
  for (;;) {
    if (!n) break; if (*t == ch) break; ++t; --n;
    if (!n) break; if (*t == ch) break; ++t; --n;
    if (!n) break; if (*t == ch) break; ++t; --n;
    if (!n) break; if (*t == ch) break; ++t; --n;
  }
  return t - s;
}


unsigned int byte_rchr(s,n,c)
char *s;
register unsigned int n;
int c;
{
  register char ch;
  register char *t;
  register char *u;

  ch = c;
  t = s;
  u = 0;
  for (;;) {
    if (!n) break; if (*t == ch) u = t; ++t; --n;
    if (!n) break; if (*t == ch) u = t; ++t; --n;
    if (!n) break; if (*t == ch) u = t; ++t; --n;
    if (!n) break; if (*t == ch) u = t; ++t; --n;
  }
  if (!u) u = t;
  return u - s;
}
void byte_copy(to,n,from)
register char *to;
register unsigned int n;
register char *from;
{
  for (;;) {
    if (!n) return; *to++ = *from++; --n;
    if (!n) return; *to++ = *from++; --n;
    if (!n) return; *to++ = *from++; --n;
    if (!n) return; *to++ = *from++; --n;
  }
}

void byte_copyr(to,n,from)
register char *to;
register unsigned int n;
register char *from;
{
  to += n;
  from += n;
  for (;;) {
    if (!n) return; *--to = *--from; --n;
    if (!n) return; *--to = *--from; --n;
    if (!n) return; *--to = *--from; --n;
    if (!n) return; *--to = *--from; --n;
  }
}

unsigned int str_chr(s,c)
register char *s;
int c;
{
  register char ch;
  register char *t;

  ch = c;
  t = s;
  for (;;) {
    if (!*t) break; if (*t == ch) break; ++t;
    if (!*t) break; if (*t == ch) break; ++t;
    if (!*t) break; if (*t == ch) break; ++t;
    if (!*t) break; if (*t == ch) break; ++t;
  }
  return t - s;
}

unsigned int str_rchr(s,c)
register char *s;
int c;
{
  register char ch;
  register char *t;
  register char *u;

  ch = c;
  t = s;
  u = 0;
  for (;;) {
    if (!*t) break; if (*t == ch) u = t; ++t;
    if (!*t) break; if (*t == ch) u = t; ++t;
    if (!*t) break; if (*t == ch) u = t; ++t;
    if (!*t) break; if (*t == ch) u = t; ++t;
  }
  if (!u) u = t;
  return u - s;
}


int stralloc_cat(stralloc *sato,stralloc *safrom)
{
  return stralloc_catb(sato,safrom->s,safrom->len);
}

int stralloc_catb(stralloc *sa,char *s, unsigned int n)
{
  if (!sa->s) return stralloc_copyb(sa,s,n);
  if (!stralloc_readyplus(sa,n + 1)) return 0;
  byte_copy(sa->s + sa->len,n,s);
  sa->len += n;
  sa->s[sa->len] = 'Z'; /* ``offensive programming'' */
  return 1;
}


int stralloc_cats(stralloc *sa, char *s)
{
  return stralloc_catb(sa,s,str_len(s));
}

int stralloc_copy(stralloc *sato,stralloc *safrom)
{
  return stralloc_copyb(sato,safrom->s,safrom->len);
}

int stralloc_readyplus(stralloc *x, unsigned int n)
{ 
    register unsigned int i;
    if (x->s) {
	i = x->a; 
	n += x->len;
	if (n > i) {
	    x->a = 30 + n + (n >> 3);
	    if (alloc_re(&x->s,i * sizeof(char),x->a * sizeof(char))) 
		return 1;
	    x->a = i; 
	    return 0; 
	}
	return 1; 
    }
  x->len = 0;
  return !!(x->s = (char *) alloc((x->a = n) * sizeof(char))); 
}




int stralloc_ready(stralloc *x, unsigned int n) 
{
    register unsigned int i;
    if (x->s) 
	{
	    i = x->a;
	    if (n > i) {
		x->a = 30 + n + (n >> 3); 
		if (alloc_re(&x->s,i * sizeof(char),x->a * sizeof(char))) 
		    return 1; 
		x->a = i; 
		return 0; 
	    } 
	    return 1; 
	}
  x->len = 0;
  return !!(x->s = (char *) alloc((x->a = n) * sizeof(char))); 
}





int stralloc_append(stralloc *x, char *i)
{ 
    if (!stralloc_readyplus(x,1)) 
	return 0; 
    x->s[x->len++] = *i; 
    return 1; 
}


int stralloc_copyb(stralloc * sa, char *s, unsigned int n)
{
  if (!stralloc_ready(sa,n + 1)) return 0;
  byte_copy(sa->s,n,s);
  sa->len = n;
  sa->s[n] = 'Z'; /* ``offensive programming'' */
  return 1;
}

int stralloc_copys(stralloc *sa, char *s)
{
  return stralloc_copyb(sa,s,str_len(s));
}

