/* -*- C -*-
 * Spamtest connection library
 * File: msgstore.c
 *      (C)  Alex Tutubalin <lexa@lexa.ru>
 *	Apr-May 2003
 *
 * Created: Wed Apr 23 22:08:08 2003
 */
#include "heap.h"
#include "msgstore.h"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/mman.h>
#include <ctype.h>
#include <arpa/inet.h>
#include <sys/stat.h> 
#include <netinet/in.h>

#define RCPT_ALLOC		10
#define HDR_ALLOC		200
#define METADATA_ALLOC		512
#define HEADERS_ALLOC		4096
#define HDRBUF_ALLOC		4096
#define BODY_ALLOC0		20480
#define BODY_ALLOC1		102400
#define BODY_ALLOCMETA		20
#ifndef DEFAULT_ALLOC
#define DEFAULT_ALLOC		10240
#endif

#define CRLF "\r\n"
#define CR   "\n"

/* store one line (possibly incomplete) into message internal store */
int message_put_body_line(message_t *msg, char *data,size_t len);

/* MESSAGE TEXT TYPE */
message_text_t *msgtext_init(msg_heap_t *heap)
{
    message_text_t *ret = heap_allocz(heap,sizeof(message_text_t));
    if(!ret) return NULL;
    ret->array = heap_allocz(heap,(ret->alloc=BODY_ALLOCMETA)*sizeof(body_chunk_t));
    if(!ret->array) return NULL;
    ret->used = 0;
    ret->store = heap;
    if(!(ret->array[0].chunk = heap_alloc(heap,ret->array[0].alloc = BODY_ALLOC0)))
	return NULL;
    ret->array[0].used = 0;
    ret->file_fd = -1;
    return ret;
}

message_text_t *msgtext_init_with_tempfile(msg_heap_t *heap,const char *tempdir, size_t max_rss)
{
    struct stat buf;
    message_text_t *ret = msgtext_init(heap);
    if(ret && max_rss > 0 && tempdir && strlen(tempdir)>0 && !stat(tempdir, &buf) && S_ISDIR(buf.st_mode))
	{
            ret->file_fd = -1;
            ret->maxrss = max_rss;
            if(tempdir[strlen(tempdir)-1]=='/')
                ret->tempdir = heap_strdup(heap,tempdir);
            else
                ret->tempdir = heap_pstrcat(heap,tempdir,"/",NULL);
            return ret;
	}
    return ret;
}


long msgtext_totalsize(message_text_t *msgt)
{
    if(msgt) return msgt->totalsize;
    return -1;
}

int msgtext_count_chunks(message_text_t *msgt)
{
    if(msgt) return msgt->file_fd>=0? 1 :  msgt->used+1;
    return -1;
}

int msgtext_chunk_len(message_text_t *msgt, size_t chunk)
{
    if(msgt->file_fd>=0)
	{
	    return chunk==0?msgt->totalsize:-1;
	}
    else if(msgt && msgt->alloc >= chunk)
	return msgt->array[chunk].used;
    return -1;
}

void msgtext_mmap(message_text_t *msgt)
{
#if defined(OS_LINUX)
    int psize = getpagesize();
#endif
    if(!msgt) return;
    if(msgt->file_fd < 0) return;
    if(msgt->mmap_ptr)
	munmap(msgt->mmap_ptr,msgt->mmap_len);
#if defined (OS_LINUX)
    msgt->mmap_len = (msgt->totalsize/psize+1)*psize;
#else
    msgt->mmap_len = msgt->totalsize;
#endif
    msgt->mmap_ptr = mmap(NULL,msgt->mmap_len,PROT_READ|PROT_WRITE,
#ifdef MAP_NOSYNC
			  MAP_NOSYNC|MAP_SHARED,
#else
			  MAP_SHARED,
#endif
			  msgt->file_fd,0);

}


char* msgtext_get_chunk(message_text_t *msgt, size_t chunk)
{
    if(msgt && msgt->file_fd >=0)
	{
	    if(chunk==0)
		{
		    if(!msgt->mmap_ptr || msgt->totalsize!=msgt->mmap_len)
			msgtext_mmap(msgt);
		    return msgt->mmap_ptr;
		}
	    else
		return NULL;
	}
    else if(msgt && msgt->alloc >= chunk)
	return msgt->array[chunk].chunk;
    return NULL;
}

static void chunk_callback(body_chunk_t *b, writer_func_t func)
{
    if(b && b->chunk && b->used) (*func)(b->chunk,b->used);
}

static void chunk_callback_ex(body_chunk_t *b, writer_func_ex_t func, void *context)
{
    if(b && b->chunk && b->used) (*func)(context, b->chunk,b->used);
}

void msgtext_callback(message_text_t *msgt, writer_func_t func)
{
    int i;
    if(!msgt) return;

    if(msgt->file_fd>=0)
	{
	    if(!msgt->mmap_ptr || msgt->totalsize!=msgt->mmap_len)
		msgtext_mmap(msgt);
	    (*func)(msgt->mmap_ptr,msgt->mmap_len);
	}
    else
	for(i=0;i<=msgt->used;i++)
	    chunk_callback(&msgt->array[i],func);

}

void msgtext_callback_ex(message_text_t *msgt, writer_func_ex_t func, void *context)
{
    int i;
    if(!msgt) return;

    if(msgt->file_fd>=0)
	{
	    if(!msgt->mmap_ptr || msgt->totalsize!=msgt->mmap_len)
		msgtext_mmap(msgt);
	    (*func)(context, msgt->mmap_ptr,msgt->mmap_len);
	}
    else
	for(i=0;i<=msgt->used;i++)
	    chunk_callback_ex(&msgt->array[i],func, context);

}

static int msgtext_realloc_array(message_text_t *msgt)
{
    body_chunk_t *newa;
    if(!msgt) return -1;
    newa = heap_reallocz(msgt->store,msgt->array,
			sizeof(body_chunk_t)*msgt->alloc,
			sizeof(body_chunk_t)*(msgt->alloc+BODY_ALLOCMETA));
    if(newa)
	{
	    msgt->array = newa;
	    msgt->alloc+=BODY_ALLOCMETA;
	    return 0;
	}
    return -1;
}

static int msgtext_allocnew(message_text_t *msgt)
{
    body_chunk_t *bc;
    if(!msgt || !msgt->array || !msgt->alloc) return -1;
    if(msgt->used >= msgt->alloc-1)
	if(-1==msgtext_realloc_array(msgt))
	    return -1;
    msgt->used++;
    bc=&msgt->array[msgt->used];
    if(!(bc->chunk = heap_alloc(msgt->store,bc->alloc = BODY_ALLOC1)))
	return -1;
    bc->used = 0;
    return 0;
}

static int mem_msgtext_puttext(message_text_t *msgt, const char *data, size_t len)
{
    body_chunk_t *bc;
    size_t put = 0;
    if(!msgt || !msgt->array || !msgt->alloc) return -1;
    if(len<1) return 0;
    bc = &msgt->array[msgt->used];
    if(bc->used == bc->alloc)
	if(-1==msgtext_realloc_array(msgt))
	    return -1;

    if(bc->used<bc->alloc)
	{
	    put = (len > bc->alloc-bc->used) ? bc->alloc-bc->used : len;
	    memmove(bc->chunk+bc->used,data,put);
	    bc->used+=put;
	    msgt->totalsize+=put;
	}
    if(put < len)
	{
	    int ret;
	    if(-1==msgtext_allocnew(msgt))
		{
		    // recover!
		    bc->used-=put;
		    msgt->totalsize-=put;
		    return -1;
		}
	    if(-1==(ret= mem_msgtext_puttext(msgt,data+put,len-put)))
		{
		    // recover!
		    bc->used-=put;
		    msgt->totalsize-=put;
		    return -1;
		}
	}
    return 0;
}
int msgtext_puttext(message_text_t *msgt, const char *data, size_t len)
{
    if(!msgt) return -1;
    if(msgt->file_fd>=0)
	{
	    if(write(msgt->file_fd,data,len)!=len)
		return -1;
	    msgt->totalsize+=len;
	    return 0;
	}
    if(msgt->tempdir && msgt->maxrss>0 && msgt->totalsize+len > msgt->maxrss)
	{
	    // create file and put all data to it
	    char filebuf[MAXPATHLEN+1];
	    int i;
	    snprintf(filebuf,MAXPATHLEN,"%s%s",msgt->tempdir,"antispam.XXXXXXXX");
	    if((msgt->file_fd = mkstemp(filebuf))<0)
		return -1;
	    unlink(filebuf);
	    for(i=0;i<=msgt->used;i++)
		write(msgt->file_fd,msgt->array[i].chunk,msgt->array[i].used);
	    msgt->used = 0;
	    write(msgt->file_fd,data,len);
	    msgt->totalsize+=len;
	    return 0;
	}
    return mem_msgtext_puttext(msgt,data,len);
}

void msgtext_close(message_text_t *msgt)
{
    if(!msgt) return;
    if(msgt->file_fd>=0)
	{
	    close(msgt->file_fd);
	    msgt->file_fd = -1;
	}
    if(msgt->mmap_ptr && msgt->mmap_len>0)
	munmap(msgt->mmap_ptr,msgt->mmap_len);
    msgt->mmap_ptr = NULL;
    msgt->mmap_len = 0;
}




/* **** RECIPIENT LIST ***** */

recipient_list_t *rcptlist_init(msg_heap_t *heap)
{
    recipient_list_t *ret;
    if(!heap) return NULL;
    ret = heap_allocz(heap,sizeof(recipient_list_t));
    if(!ret) return NULL;
    if(!(ret->array = heap_allocz(heap,(ret->alloc=RCPT_ALLOC)*sizeof(some_text))))
	{ return NULL;};
    ret->storeptr = heap;
    return ret;
}

int rcptlist_exists(recipient_list_t *rl,const char *probe)
{
    int i;
    if(!rl || !probe) return 0;
    for(i=0;i<rcptlist_count(rl);i++)
	if(!strcmp(rcptlist_getn(rl,i),probe))
	    return 1;
    return 0;
}

int rcptlist_adduniq(recipient_list_t *rl, const char *probe)
{
    if(!rl || !probe) return -1;
    if(!rcptlist_exists(rl,probe))
	return rcptlist_add(rl,probe);
    return 0;
}


void rcptlist_clean(recipient_list_t *rl)
{
    if(rl)rl->used = 0;
}

int rcptlist_checkspace(recipient_list_t *rl)
{
    if(!rl) return -1;
    if(rl->used >= rl->alloc)
	{
	    int newsize = rl->alloc ? rl->alloc*2 : RCPT_ALLOC;
	    some_text *newalloc = heap_reallocz(rl->storeptr,rl->array,
					       rl->alloc*sizeof(some_text),
						newsize * sizeof(some_text));
	    if(newalloc)
		{
		    rl->alloc = newsize;
		    rl->array = newalloc;
		}
	    else
		return -1;
	}
    return 0;
}

int rcptlist_add(recipient_list_t *rl, const char *rcpt)
{
    size_t len;
    if(!rl || !rcpt) return -1;
    len = strlen(rcpt);
    if(-1==rcptlist_checkspace(rl))
	return -1;
    rl->array[rl->used].hptr = heap_strdupn(rl->storeptr,rcpt,len);
    rl->array[rl->used].length = len;
    rl->used++;
    return 0;
}

int rcptlist_addn(recipient_list_t *rl, const char *rcpt,size_t len)
{
    if(!rl) return -1;
    if(-1==rcptlist_checkspace(rl))
	return -1;
    rl->array[rl->used].hptr = heap_strdupn(rl->storeptr,rcpt,len+1);
    rl->array[rl->used].length = len;
    rl->used++;
    return 0;
}


int rcptlist_count(recipient_list_t *rl)
{
    return rl?rl->used:-1;
}

char * rcptlist_getn(recipient_list_t *rl,int idx)
{
    if(!rl || !rl->storeptr || !rl->array || rl->used<=idx) return NULL;
    return rl->array[idx].hptr;
}

int rcptlist_iterate(recipient_list_t *rl, writer_func_t func)
{
    int count = 0,i;
    if(!rl || !rl->storeptr || !rl->array) return -1;
    for(i=0; i< rl->used; i++)
	{
	    if(func)
		(*func)(rl->array[i].hptr,strlen(rl->array[i].hptr));
	    count++;
	}
    return count;
}

int rcptlist_iterate_ex(recipient_list_t *rl, writer_func_ex_t func, void *context)
{
    int count = 0,i;
    if(!rl || !rl->storeptr || !rl->array) return -1;
    for(i=0; i< rl->used; i++)
	{
	    if(func)
		(*func)(context, rl->array[i].hptr,strlen(rl->array[i].hptr));
	    count++;
	}
    return count;
}


/* ***** MESSAGE *****  */

message_t *message_init(void)
{
    msg_heap_t *heap = heap_init(METADATA_ALLOC);
    message_t *ret;
    if(!heap) return NULL;

    ret = heap_alloc(heap,sizeof(message_t));
    if(!ret)
	return NULL;
    bzero(ret,sizeof(*ret));
    
    ret->store = heap;

    if(NULL==(ret->recipients = rcptlist_init(ret->store)))
	{message_free(ret);return NULL;};

    if(!(ret->headers = headers_init(ret->store)))
	{message_free(ret);return NULL;};

    if (!(ret->bodystore = msgtext_init(ret->store)))
	{message_free(ret);return NULL;};
	

    ret->last_string_complete = 1;
    return ret;
}

message_t *message_init_with_heap(msg_heap_t *heap)
{
    message_t *ret;
    if(!heap) return NULL;
    ret = heap_alloc(heap,sizeof(message_t));
    if(!ret)
	return NULL;
    bzero(ret,sizeof(*ret));
    
    ret->store = heap;

    if(NULL==(ret->recipients = rcptlist_init(ret->store)))
	{message_free(ret);return NULL;};

    if(!(ret->headers = headers_init(ret->store)))
	{message_free(ret);return NULL;};

    if (!(ret->bodystore = msgtext_init(ret->store)))
	{message_free(ret);return NULL;};
	

    ret->last_string_complete = 1;
    return ret;
}

message_t *message_init_with_tempfile(const char*tempdir, size_t maxrss)
{
    msg_heap_t *heap = heap_init(METADATA_ALLOC);
    message_t *ret;
    if(!heap) return NULL;

    ret = heap_alloc(heap,sizeof(message_t));
    if(!ret)
	return NULL;
    bzero(ret,sizeof(*ret));
    
    ret->store = heap;

    if(NULL==(ret->recipients = rcptlist_init(ret->store)))
	{message_free(ret);return NULL;};

    if(!(ret->headers = headers_init(ret->store)))
	{message_free(ret);return NULL;};

    if (!(ret->bodystore = msgtext_init_with_tempfile(ret->store,tempdir,maxrss)))
	{message_free(ret);return NULL;};
	

    ret->last_string_complete = 1;
    return ret;
}


void message_free(message_t *msg)
{
    if(!msg) return;
    if(msg->recipients)rcptlist_clean(msg->recipients);
    if(msg->bodystore) msgtext_close(msg->bodystore);
    heap_free(msg->store);
}

int message_set_from(message_t *msg, const char *from)
{
    size_t len;
    if(!msg) return -1;
    len = strlen(from?from:"");
    msg->from.hptr = heap_strdupn(msg->store,from?from:"",len);
    msg->from.length = len;
    return 0;
}

void message_clean_rcpt(message_t *msg)
{
    if(msg && msg->recipients)rcptlist_clean(msg->recipients);
}

int message_add_rcpt(message_t *msg, const char *rcpt)
{
    return (msg && msg->recipients)?rcptlist_add(msg->recipients,rcpt):-1;
}

int message_set_ip(message_t *msg, const char *ip)
{
    struct in_addr addr;
    size_t len = strlen(ip?ip:"");
    if(!msg) return -1;
    if(!ip || !inet_aton(ip,&addr)) return -1;
    msg->ip_addr.hptr = heap_strdupn(msg->store,ip?ip:"",len);
    msg->ip_addr.length = len;
    return 0;
}

static int headers_checkspace(headers_list_t *hdr,int needspace)
{
    int asize = needspace > HDR_ALLOC ? (needspace/HDR_ALLOC+1)*HDR_ALLOC:HDR_ALLOC;
    if(!hdr || !hdr->storeptr) return -1;

    if(hdr->used+needspace >= hdr->alloc)
	{
	    msg_header *newarray = heap_reallocz(hdr->storeptr,
						hdr->array,
						hdr->alloc*sizeof(msg_header),
						(hdr->alloc+asize)*sizeof(msg_header));
	    if(newarray)
		{
		    hdr->alloc+=asize;
		    hdr->array = newarray;
		    return 0;
		}
	    else
		return -1;
	}
    return 0;
}


int message_put_header_char(message_t *msg,char chr,int noprocess)
{
    if(!msg || !msg->headers) return -1;
    if(headers_checkspace(msg->headers,1))
	return -1;
    if(msg->headers->hdrbuf 
       && msg->headers->bufused 
       && msg->headers->array[msg->headers->used].state == HDR_VALUEDONE)
	{
	    if(chr == ' ' || chr == '\t')
		{
		    // continuation line
		    msg->headers->array[msg->headers->used].state = HDR_VALUE;
		    
		}
	    else
		{
		    // new header start
		    // copy buffer into value field
		    msg->headers->hdrbuf[msg->headers->bufused]='\0';
		    msg->headers->array[msg->headers->used].value.hptr = heap_strdupn(msg->headers->storeptr,msg->headers->hdrbuf,msg->headers->bufused+1);
		    msg->headers->array[msg->headers->used].value.length = msg->headers->bufused;
		    msg->headers->bufused = 0;
		    msg->headers->used++;
		    bzero(&msg->headers->array[msg->headers->used],sizeof(msg_header));
		    if(noprocess) return 0;
		}
	}
    // allocate bufspace if needed
    if(!msg->headers->hdrbuf || msg->headers->bufalloc == msg->headers->bufused)
	{
	    msg->headers->hdrbuf = heap_reallocz(msg->headers->storeptr,msg->headers->hdrbuf,
					    msg->headers->bufalloc,
					    msg->headers->bufalloc+HDRBUF_ALLOC);
	    msg->headers->bufalloc+=HDRBUF_ALLOC;
	}
    if(!msg->headers->hdrbuf)
	return -1;

       switch(msg->headers->array[msg->headers->used].state)
	   {
	   case 0:
	       // just created
	       msg->headers->hdrbuf[msg->headers->bufused++] = chr;
	       msg->headers->array[msg->headers->used].state = HDR_CREATED;
	       break;
	   case HDR_CREATED:
	       // some data put in header name
	       if(chr == ':') // end of header 
		   {
		       msg->headers->hdrbuf[msg->headers->bufused]='\0';
		       msg->headers->array[msg->headers->used].state = HDR_NAMEDONE;
		       msg->headers->array[msg->headers->used].header.hptr 
			   = heap_strdupn(msg->headers->storeptr,
					  msg->headers->hdrbuf,
					  msg->headers->bufused);
		       msg->headers->array[msg->headers->used].header.length = msg->headers->bufused;
		       msg->headers->bufused=0;
		   }
	       else if(strchr(" \t[]<>",chr))
		   {
		       // ignore this header - bad symbols at left side
		       msg->headers->array[msg->headers->used].deleted = 1;
		   }
	       else
		   {
		       msg->headers->hdrbuf[msg->headers->bufused++] = chr;
		   }
	       break;
	   case HDR_NAMEDONE:
	       // complete header name read
	       if(chr != ' ')
		   {
		       msg->headers->hdrbuf[msg->headers->bufused++] = chr;
		       msg->headers->array[msg->headers->used].state = HDR_VALUE;
		       if(chr == '\n') // end of header line
			   {
			       msg->headers->array[msg->headers->used].state = HDR_VALUEDONE;
			   }
		   }
	       else
		   {
		       ; // skip leading space
		   }
	       break;
	   case HDR_VALUE:
	       // value chars
	       msg->headers->hdrbuf[msg->headers->bufused++] = chr;
	       if(chr == '\n') // end of header line
		   {
		       msg->headers->array[msg->headers->used].state = HDR_VALUEDONE;
		   }
	       break;
	   case HDR_VALUEDONE:
	       // should not happen!
               ;
	   }
    
       return 0;
}

int message_put_body_line(message_t *msg, char *data, size_t len)
{
    int partial_string;
    char *xmailbox = "X-Mailbox-Line: ";
    if(!msg || !data)
	return -1;
    if(len<1)
	return 0; // Empty line

    partial_string = data[len-1] == '\n' ? 0 : 1;
  RETRY:
    if(msg->headers_processed)
	{
	    if(-1==msgtext_puttext(msg->bodystore,data,len))
		return -1;
	}
    else
	{
	    // emtpy string - just a \n or \r\n
	    if(msg->last_string_complete && !partial_string &&
	       ((len < 2 && data[0]=='\n') ||(len < 3 && data[1]=='\n' && data[0]=='\r'))
		)
		{
		    // empty line after header
		    msg->headers_processed = 1;
		    // close last header
		    if(msg->headers->array[msg->headers->used].state == HDR_VALUEDONE)
			{
			    message_put_header_char(msg,'\0',1);
			}
		    goto RETRY; // add newline as first line of body
		}
	    // complete string
	    else if(data[len-1]=='\n')
		{
		    if(msg->headers->array[msg->headers->used].state==HDR_VALUEDONE &&
		       (data[0]==' ' || data[0]=='\t'))
			{
			    // continuation line
			    int i;
			    for(i=0;i<len;i++)
				if(-1==message_put_header_char(msg,data[i],0))
				    return -1;
			    return 0;
			}
		    // just a CR. Look at last header. If it just
		    // created or state is 0, then remove last used
		    // header or destroy it
		    else if (msg->headers->array[msg->headers->used].state == HDR_VALUEDONE &&
			     !memchr(data,':',len))
			{
			    // header completed, new header is
			    // broken. Break header processing.
			    message_put_header_char(msg,'\0',1);
			    msg->headers_processed = 1;
			    msgtext_puttext(msg->bodystore,"\n",1);
			    goto RETRY;
			}
		    else if( /* not complete header line */
			(msg->headers->array[msg->headers->used].state == HDR_CREATED ||
			 msg->headers->array[msg->headers->used].state == 0) 
			&& !memchr(data,':',len) /* no semicolon in passed string */
			)
			{
			    msg->headers_processed = 1;
			    msgtext_puttext(msg->bodystore,"\n",1);
			    msgtext_puttext(msg->bodystore,
					    msg->headers->hdrbuf,msg->headers->bufused);
			    msg->headers->bufused = 0;
			    // and clean state for current header
			    msg->headers->array[msg->headers->used].state=0;
			    goto RETRY;
			}
		    else /* last header has : or value already added */
			{
			    int i;
			    for(i=0;i<len;i++)
				if(-1==message_put_header_char(msg,data[i],0))
				    return -1;
			}
		}
	    else
		{
		    // partial string
		    int i;
		    if(msg->last_string_complete && len >=5 && !strncmp(data,"From ",5))
			{
			    char *p = xmailbox;
			    while(*p)
				{
				    if(-1==message_put_header_char(msg,*p,0))
					return -1;
				    p++;
				}
			}
		    for(i=0;i<len;i++)
			if(-1==message_put_header_char(msg,data[i],0))
			    return -1;
		}
	}
    
    msg->last_string_complete = !partial_string;
    return 0;
}

int message_put_body(message_t *msg, char *data, size_t len)
{
    char *start,*end;
    int remain = len,to_send,status;
    if(!msg || !data) return -1;
    if(msg->headers_processed)
	return msgtext_puttext(msg->bodystore,data,len);
	
    start = data;
    while(remain>0)
	{
	    end=memchr(start,'\n',remain);
	    if(!end)
		{
		    // no CRs, send entire data and break
		    return message_put_body_line(msg,start,remain);
		}
	    else
		{
		    end++;
		    to_send=end-start;
		    status = message_put_body_line(msg,start,to_send);
		    if(status!=0) return status;
		    start = end;
		    remain-=to_send;
		    if(msg->headers_processed)
			return msgtext_puttext(msg->bodystore,start,remain);
		}
	}
    return 0;
}

char *get_message_from(message_t *msg)
{
    return (msg && msg->from.hptr && msg->from.length)?msg->from.hptr:NULL;
}

char *get_message_ip(message_t *msg)
{
    return (msg && msg->ip_addr.length)?msg->ip_addr.hptr:NULL;
}


int get_message_nrcpt(message_t *msg)
{
    return (msg && msg->recipients)?rcptlist_count(msg->recipients):-1;
}

char *get_message_rcpt(message_t *msg,int n)
{
    return (msg && msg->recipients)?rcptlist_getn(msg->recipients,n):NULL;
}

int message_rcpts_iterate(message_t *msg, writer_func_t callback)
{
    if(msg && msg->recipients)
	return rcptlist_iterate(msg->recipients,callback);
    else
	return -1;
}

int message_rcpts_iterate_ex(message_t *msg, writer_func_ex_t callback, void *context)
{
    if(msg && msg->recipients)
	return rcptlist_iterate_ex(msg->recipients,callback,context);
    else
	return -1;
}


long get_message_bodysize(message_t *msgt)
{
    return msgt?msgtext_totalsize(msgt->bodystore):-1;
}

int get_message_bodychunks_count(message_t *msgt)
{
    return msgt?msgtext_count_chunks(msgt->bodystore):-1;
}

int get_message_bodychunk_len(message_t *msgt,size_t chunk)
{
    return msgt? msgtext_chunk_len(msgt->bodystore,chunk):-1;
}

char* get_message_bodychunk_ptr(message_t *msgt,size_t chunk)
{
    return msgt? msgtext_get_chunk(msgt->bodystore, chunk):NULL;
}

void message_body_callback(message_t *msgt, writer_func_t func)
{
    if(msgt) msgtext_callback(msgt->bodystore,func);
}

void message_body_callback_ex(message_t *msgt, writer_func_ex_t func, void *context)
{
    if(msgt) msgtext_callback_ex(msgt->bodystore,func, context);
}



int message_headers_iterate(message_t *msg,header_iterator_t func)
{
    return headers_iterate(msg->headers,func);
}

int message_headers_iterate_ex(message_t *msg,header_iterator_ex_t func, void *context)
{
    return headers_iterate_ex(msg->headers,func, context);
}


static void write_string(const char *str,writer_func_t writer)
{
    (*writer)(str,strlen(str));
}


int message_dump(message_t *msg,writer_func_t writer)
{
    int i;
    if(!msg || !writer) return -1;
    if(msg->ip_addr.length)
	{
	    write_string("HELO: ",writer);
	    write_string(msg->ip_addr.hptr,writer);
	    write_string("\n",writer);
	}
    if(msg->from.length)
	{
	    write_string("MAIL FROM: ",writer);
	    write_string(msg->from.hptr,writer);
	    write_string("\n",writer);
	}
    if(msg->recipients)
	rcptlist_iterate(msg->recipients,writer);
    write_string("DATA\n",writer);
    // write headers
    for(i=0;i<msg->headers->used;i++)
	{
	    write_string(msg->headers->array[i].header.hptr,writer);
	    write_string(": ",writer);
	    write_string(msg->headers->array[i].value.hptr,writer);
	}
    write_string("\n",writer);
    message_body_callback(msg,writer);
    write_string(".\n",writer);
    return 0;
}

/*    HEADERS */

headers_list_t* headers_init(msg_heap_t *heap)
{
    headers_list_t *ret = heap_allocz(heap,sizeof(headers_list_t));
    if(!ret) return NULL;
    ret->storeptr = heap;
    if(-1==headers_checkspace(ret,HDR_ALLOC))
	return NULL;
    return ret;
}

int headers_iterate(headers_list_t *hdrs, header_iterator_t func) 
{
    int count=0,i;
    if(!hdrs || !func) return -1;
    for(i=0;i<hdrs->used;i++)
	if(!hdrs->array[i].deleted)
	    {
		(*func)(hdrs->array[i].header.hptr,hdrs->array[i].value.hptr);
		count++;
	    }
    return count;
}

int headers_iterate_ex(headers_list_t *hdrs, header_iterator_ex_t func, void *context)
{
    int count=0,i;
    if(!hdrs || !func) return -1;
    for(i=0;i<hdrs->used;i++)
	if(!hdrs->array[i].deleted)
	    {
		(*func)(context, hdrs->array[i].header.hptr,hdrs->array[i].value.hptr);
		count++;
	    }
    return count;
}


static int headers_trivial_add(headers_list_t *hdr,char *h,int hl,char *v,int vl, int deleted)
{
    if(-1==headers_checkspace(hdr,1))
	return -1;
    if(!(hdr->array[hdr->used].header.hptr = heap_strdupn(hdr->storeptr,h,hl)))
	return -1;
    hdr->array[hdr->used].header.length = hl;
    if(!(hdr->array[hdr->used].value.hptr = heap_strdupn(hdr->storeptr,v,vl)))
	return -1;
    hdr->array[hdr->used].value.length = vl;
    hdr->array[hdr->used].deleted = deleted;
    hdr->used++;
    return 0;
}


headers_list_t *headers_clone(message_t *msg)
{
    int i;
    headers_list_t *ret;
    if(!msg || !msg->store) return NULL;

    ret = heap_allocz(msg->store,sizeof(headers_list_t));
    if(!ret) return NULL;
    ret->storeptr = msg->store;
    ret->array=heap_allocz(msg->store,
			  (ret->alloc = msg->headers->alloc)*sizeof(msg_header));
    if(!ret->array)
	return NULL;
    /* create copy of values */
    for(i=0;i<msg->headers->used;i++)
	if(-1==headers_trivial_add(ret,
				   msg->headers->array[i].header.hptr,
				   msg->headers->array[i].header.length,
				   msg->headers->array[i].value.hptr,
				   msg->headers->array[i].value.length,
				   msg->headers->array[i].deleted))
	    return NULL;

    return ret;
}

int headers_del_hdr(headers_list_t *hdr,const char *header)
{
    int i,count=0;
    if(!hdr) return -1;
    if(!header || strlen(header) <1)
	return -1;

    for(i=0;i<hdr->used;i++)
	if(!strcasecmp(hdr->array[i].header.hptr,header))
	    {
		hdr->array[i].deleted = 1;
		count++;
	    }
    return count;
}

int headers_chg_hdr(headers_list_t *hdr,const  char *header,const  char *value)
{
    if(-1 != headers_del_hdr(hdr,header))
	return headers_add_hdr(hdr,header,value);
    return -1;
}

static char *crlf=CRLF;
static char *cr=CR;
static char *empty="";

int headers_add_hdr(headers_list_t *hdr, const char *header, const char *value)
{
    int i,lastdelidx=-1,lastidx=-1;
    int lenh,lenv;
    const char *valuecr;
    const char *newline;
    size_t      newline_length;

    if(!hdr || !header) return -1;

    newline = (hdr->flags & HDR_USE_CRLF) ? crlf : cr;
    newline_length = strlen(newline);

    lenh = strlen(header);
    if(lenh < 1) return 0;

    if(!value) value = newline;
    lenv = strlen(value);
    if(lenv < 1)
	{
	    value = newline;
	    lenv=1;
	}
    if(value[lenv-1] == '\n')
	valuecr = value;
    else
	{
	    valuecr = heap_pstrcat(hdr->storeptr,value,newline,NULL);
	    if(!valuecr) return -1;
	    lenv+=newline_length;
	}

    if(headers_checkspace(hdr,1))
	return -1;

    // find first deleted or first existing header
    for(i=0;i<hdr->used;i++)
	if(!strcasecmp(hdr->array[i].header.hptr,header))
	    {
		if(hdr->array[i].deleted)
		    {
			if(lastdelidx<0)
			    lastdelidx=i;
		    }
		else 
		    {
			if(lastidx<0)
			    lastidx = i;
		    }
	    }
    if(lastidx>-1)
	{
	    // header already exists, append value to it
	    if(hdr->array[lastidx].value.length == 1  &&
	       *(hdr->array[lastidx].value.hptr)=='\n')
		{
		    
		    hdr->array[lastidx].value.hptr = heap_strdupn(hdr->storeptr,valuecr,lenv);
		    hdr->array[lastidx].value.length=lenv;
		}
	    else
		{
		    hdr->array[lastidx].value.hptr = heap_pstrcat(hdr->storeptr,
								  hdr->array[lastidx].value.hptr,
								  "  ",valuecr,NULL);
		    hdr->array[lastidx].value.length+=lenv+2;
		}
	}
    else if(lastdelidx>-1)
	{

	    hdr->array[lastdelidx].value.hptr = heap_strdupn(hdr->storeptr,valuecr,lenv);
	    hdr->array[lastdelidx].value.length=lenv;
	    hdr->array[lastdelidx].deleted = 0;
	}
    else
	{
	    // complete new header
	    i = hdr->used;
	    hdr->used++;
	    hdr->array[i].header.hptr = heap_strdupn(hdr->storeptr,header,lenh);
	    hdr->array[i].header.length = lenh;

	    hdr->array[i].value.hptr = heap_strdupn(hdr->storeptr,valuecr,lenv);
	    hdr->array[i].value.length = lenv;
	    hdr->array[i].deleted = 0;
	}
    return 0;
    
}

int headers_prepend_hdr(headers_list_t *hdr, const char *header, const char *value)
{
    int i,lastdelidx=-1,lastidx=-1;
    int lenh,lenvZ;
    char *valueZ;
    const char *newline;
    size_t      newline_length;


    if(!hdr || !header) return -1;

    newline = (hdr->flags & HDR_USE_CRLF) ? crlf : cr;
    newline_length = strlen(newline);

    lenh = strlen(header);
    if(lenh < 1) return 0;

    
    // we need to have valueZ without \n at end
    if(!value)
	{
	    valueZ = empty;
	    lenvZ = 0;
	}
    else
	{
	    lenvZ = strlen(value);
	    if(lenvZ>0) 
		{
		    if(value[lenvZ-1] == '\n')
			{
			    valueZ = heap_strdupn(hdr->storeptr,value,lenvZ-1);
			    valueZ[lenvZ-1]='\0';
			    lenvZ--;
			    // eat extra NLs if any
			    while(lenvZ>0 && valueZ[lenvZ-1]=='\n')
				{
				    valueZ[lenvZ-1]='\0';
				    lenvZ--;
				}
			}
		    else
			valueZ = (char*)value;
		}
	    else
		{
		    valueZ = empty;
		    lenvZ=0;
		}
	}
    // so valueZ is null-terminated without extra NL

    if(headers_checkspace(hdr,1))
	return -1;

    // first deleted or first existing
    for(i=0;i<hdr->used;i++)
	if(!strcasecmp(hdr->array[i].header.hptr,header))
	    {
		if(hdr->array[i].deleted)
		    {
			if(lastdelidx<0)
			    lastdelidx=i;
		    }
		else 
		    {
			if(lastidx<0)
			    lastidx = i;
		    }
	    }
    if(lastidx>-1)
	{
	    // header already exists, append value to it (1st one!)
	    if(hdr->array[lastidx].value.length == 1  &&
	       *(hdr->array[lastidx].value.hptr)=='\n')
		{
		    // empty header, duplicate passed value
		    hdr->array[lastidx].value.hptr = heap_pstrcat(hdr->storeptr,valueZ,newline,NULL);
		    hdr->array[lastidx].value.length=lenvZ+newline_length;
		}
	    else
		{
		    // not empty header, really prepend
		    hdr->array[lastidx].value.hptr = heap_pstrcat(hdr->storeptr,valueZ," ",
								  hdr->array[lastidx].value.hptr,
								  NULL);
		    hdr->array[lastidx].value.length+=lenvZ+1; // value and space
		}
	}
    else if(lastdelidx>-1)
	{
	    hdr->array[lastdelidx].value.hptr = heap_pstrcat(hdr->storeptr,valueZ,newline,NULL);
	    hdr->array[lastdelidx].value.length=lenvZ+newline_length;
	    hdr->array[lastdelidx].deleted = 0;
	}
    else
	{
	    // complete new header
	    i = hdr->used;
	    hdr->used++;
	    hdr->array[i].header.hptr = heap_strdupn(hdr->storeptr,header,lenh);
	    hdr->array[i].header.length = lenh;

	    hdr->array[i].value.hptr = heap_pstrcat(hdr->storeptr,valueZ,newline,NULL);
	    hdr->array[i].value.length = lenvZ + newline_length;
	    hdr->array[i].deleted = 0;
	}
    return 0;
    
}


int headers_new_hdr(headers_list_t *hdr,const  char *header,const  char *value)
{
    int i,lastdelidx=-1,lastidx=-1;
    int lenh;
    int lenv;
    const char *newline;
    size_t      newline_length;

    const char *valuecr;
    
    if(!hdr) return -1;

    newline = (hdr->flags & HDR_USE_CRLF) ? crlf : cr;
    newline_length = strlen(newline);

    if(headers_checkspace(hdr,1)) return -1;

    if(!header) return -1;
    lenh = strlen(header);
    if(lenh<1) return -1;

    if(!value) value = newline;
    lenv = strlen(value);
    if(lenv < 1)
	{
	    value = newline;
	    lenv=1;
	}

    if(value[lenv-1] == '\n')
	valuecr = value;
    else
	{
	    if(!(valuecr = heap_pstrcat(hdr->storeptr,value,newline,NULL)))
		return -1;
	    lenv+=newline_length;
	}

    // last existing or first deleted
    for(i=0;i<hdr->used;i++)
	if(!strcasecmp(hdr->array[i].header.hptr,header))
	    {
		if(hdr->array[i].deleted)
		    {
			if(lastdelidx<0)
			    lastdelidx=i;
		    }
		else
		    lastidx = i;
	    }

    if(lastidx>-1)
	{
	    // header already exists, move over headers down and
	    // insert new below last exists
	    i = lastidx+1;
	    memmove(&hdr->array[i+1],&hdr->array[i],sizeof(hdr->array[0])*(hdr->used-i));
	    hdr->used++;
	    hdr->array[i].header.hptr = heap_strdupn(hdr->storeptr,header,lenh);
	    hdr->array[i].header.length = lenh;
	    hdr->array[i].value.hptr = heap_strdupn(hdr->storeptr,valuecr,lenv);
	    hdr->array[i].value.length = lenv;
	    hdr->array[i].deleted = 0;
	}
    else if(lastdelidx>-1)
	{
	    // create new value
	    i = lastdelidx;
	    hdr->array[i].value.hptr = heap_strdupn(hdr->storeptr,valuecr,lenv);
	    hdr->array[i].value.length = lenv;
	    hdr->array[i].deleted = 0;
	}
    else
	{
	    // complete new header below all headers
	    i = hdr->used;
	    hdr->used++;

	    hdr->array[i].header.hptr = heap_strdupn(hdr->storeptr,header,lenh);
	    hdr->array[i].header.length = lenh;
	    hdr->array[i].value.hptr = heap_strdupn(hdr->storeptr,valuecr,lenv);
	    hdr->array[i].value.length = lenv;
	    hdr->array[i].deleted = 0;
	}
    return 0;
    
}

#ifdef MSGTEST
void  writer(const char *d, size_t len)
{
    fwrite(d,1,len,stdout);
}

void header_writer(char *hdr,char *val)
{
    printf("%s: %s",hdr,val);
}

void recipient_writer(char *rcpt)
{
    printf("RCPT TO: %s\n",rcpt);
}



int main(int ac,char *av[])
{
    message_t *msg;
    FILE *f;
    char buf[5120],*p;
    int i,reps=1,j;
    headers_list_t *hdr;
    int RR=10;
    if(ac> 1 && !strncmp(av[1],"-r",2))
	reps = atoi(av[1]+2);

    for(i=0;i<reps;i++)
	{
	    if(!(msg = message_init()))
		exit(1);
	    message_set_from(msg,"vasyapupkin@rambler.ru");
	    message_set_ip(msg,"127.0.0.1");
	    for(j=0;j<RR;j++)
		message_add_rcpt(msg,"lexa@lexa.ru");
	    message_rcpts_iterate(msg,recipient_writer);
	    message_add_rcpt(msg,"lexa@lexa.ru");
	    message_add_rcpt(msg,"lexa@lexa.ru");
	    message_clean_rcpt(msg);
	    message_add_rcpt(msg,"lexa@lexa.ru");
	    message_add_rcpt(msg,"lexa2@lexa.ru");
	    message_add_rcpt(msg,"lexa1@lexa.ru");
	    message_add_rcpt(msg,"lexa@lexa.ru");
	    if(ac>1 && ((f = fopen(av[ac-1],"r"))!=NULL))
		{
		    while(fgets(buf,sizeof(buf),f))
			{
			    int len = strlen(buf);
			    message_put_body_line(msg,buf,len);
			}
		    fclose(f);
		}
	    else
		printf("Cannot read message!\n");
	    printf("Body len: %ld\n",get_message_bodysize(msg));
	    for(j=0;j<get_message_bodychunks_count(msg);j++)
		{
		    fwrite(get_message_bodychunk_ptr(msg,j),1,
			   get_message_bodychunk_len(msg,j),stdout);
		}
	    if((p = get_message_ip(msg)))
		printf("HELO: %s\n",p);
	    if((p = get_message_from(msg)))
		printf("MAIL FROM: %s\n",p);
	    printf("DATA\n");
	    message_headers_iterate(msg,header_writer);
	    message_body_callback(msg,writer);
#if 1
	    hdr = headers_clone(msg);
	    printf("\n======= another copy of headers === (%x)\n",(int)hdr);
	    if(hdr)
		{
		    headers_del_hdr(hdr,"To");
		    headers_del_hdr(hdr,"To2");
		    
		    headers_prepend_hdr(hdr,"X-Status","prepend0");
		    headers_add_hdr(hdr,"X-Status","xxxxx");
		    headers_add_hdr(hdr,"X-Status","yyy");
		    headers_new_hdr(hdr,"X-Status","zzz");
		    headers_prepend_hdr(hdr,"X-Status","prepend");
		    headers_add_hdr(hdr,"X-Status","add");
		    headers_iterate(hdr,header_writer);
		}
	    printf(".\n");
#endif
	    message_free(msg);
	}
    return 0;
}
#endif

#ifdef MSGTEST3
void  writer(const char *d, size_t len)
{
    fwrite(d,1,len,stdout);
}

void header_writer(char *hdr,char *val)
{
    printf("%s: %s",hdr,val);
}

void recipient_writer(char *rcpt)
{
    printf("RCPT TO: %s\n",rcpt);
}
int main(int ac,char *av[])
{
    message_t *msg;
    char *header="From: lexa@lexa.ru\nTo: lexa@lexa.ru\nSubject: test\n\n";
    char *bodyline = "quick brown foxquick brown foxquick brown foxquick brown foxquick brown foxquick brown foxquick brown foxquick brown foxquick brown foxquick brown foxquick brown fox\n";
    int i;
    size_t sz=1024*100;
    int repeat = 100000;
    if(ac>3)
	repeat = atoi(av[3]);
    if(ac>2)
	sz = 1024*atoi(av[2]);
    if(ac>1)
	msg = message_init_with_tempfile(av[1],sz);
    else
	msg = message_init_with_tempfile("/tmp",10);
    
    message_put_body(msg,header,strlen(header));
    for(i=0;i<repeat;i++)
	message_put_body(msg,bodyline,strlen(bodyline));
    printf("DATA\n");
    message_headers_iterate(msg,header_writer);
    message_body_callback(msg,writer);
    return 0;
}

#endif
#ifdef MSGTEST2
void  writer(const char *d, size_t len)
{
    fwrite(d,1,len,stdout);
}

void header_writer(char *hdr,char *val)
{
    printf("%s: %s",hdr,val);
}

void recipient_writer(char *rcpt)
{
    printf("RCPT TO: %s\n",rcpt);
}
int main(int ac,char *av[])
{
    message_t *msg;
    unsigned char buf1;
    char bufl[2048];
    FILE *f;
    if(ac<2)
	{
	    printf("Usage: %s message-file\n",av[0]);
	    exit(1);
	}
    f = fopen(av[1],"r");
    if(!f) { perror("fopen"); exit(1);}
    msg = message_init();
    if(!msg){printf("Cannot init message\n");exit(1);}
#if 1
    while(!feof(f))
	{
	    if(fread(&buf1,sizeof(buf1),1,f))
		message_put_body(msg,&buf1,1);
	}
# else
    while(fgets(bufl,2048,f))
	{
	    message_put_body(msg,bufl,strlen(bufl));
	}
#endif
    message_headers_iterate(msg,header_writer);
    printf("+++");
    message_body_callback(msg,writer);
    return 0;
}

#endif
