/* -*- C -*-
 * Spamtest connection library
 * File: record2.c
 *      (C)  Alex Tutubalin <lexa@lexa.ru>
 *	Apr-May 2003
 *
 * Created: Wed Apr 23 22:08:08 2003
 */

#include <unistd.h>
#include <stdarg.h>
#include <stdio.h>
#include <errno.h>
#include <syslog.h>
#include <string.h>
#include "record2.h"
#ifdef AP_BUILD
#include "smtp-filter/common/rec_types.h"
#else
#include "rec_types.h"
#endif
#include "spamtest.h"
#include "spamtest_internal.h"



int     rec_put(spamtest_session_t *s, int type, const char *data, int len)
{
    int     len_rest;
    int     len_byte;
    int written,w_now;
    int status;

    if((status=check_init(s))!=STS_ERR_NO_ERR)
	return STS_ERR_INVALID_PARAMS;

    s->logp(s,LOG_DEBUG,"put rec type %d(%c), size=%d",type,type,len);
    if (type < 0 || type > 255)
	{
	    s->logp(s,LOG_ERR,"rec_put: bad record type %d", type);
	    return STS_ERR_INVALID_PARAMS;
	}

    /*
     * Write the record type, one byte.
     */
    if(BUFIO_OK != (status=bufio_putbyte(s->bufio,type,s)))
	return status==BUFIO_TIMEOUT?STS_ERR_SND_RCV_TIMEOUT:STS_ERR_WRITE_ERR;
    /*
     * Write the record data length in 7-bit portions, using the 8th bit to
     * indicate that there is more. Use as many length bytes as needed.
     */
    len_rest = len;
    do {
	len_byte = len_rest & 0177;
	if (len_rest >>= 7)
	    len_byte |= 0200;
	if(BUFIO_OK != (status=bufio_putbyte(s->bufio,len_byte,s)))
	    return status==BUFIO_TIMEOUT?STS_ERR_SND_RCV_TIMEOUT:STS_ERR_WRITE_ERR;
    } while (len_rest != 0);
    /*
     * Write the record data portion. Use as many length bytes as needed.
     */
    s->logp(s,LOG_DEBUG+2,"Record data: %s",data);

    written = 0;
    while(written < len)
	{
	    w_now = bufio_write(s->bufio,(const char*)(data+written),len-written,s);
	    if(w_now<0)
		return w_now==BUFIO_TIMEOUT?STS_ERR_SND_RCV_TIMEOUT:STS_ERR_WRITE_ERR;
	    written+=w_now;
	}
    
    if(type == REC_TYPE_END || type == REC_TYPE_BLOCKFIN)
	if(BUFIO_OK != (status=bufio_flush(s->bufio,s)))
	    return status==BUFIO_TIMEOUT?STS_ERR_SND_RCV_TIMEOUT:STS_ERR_WRITE_ERR;
    return (type);
}

#define NBBY 8


int read_byte(bufio_t *buf,spamtest_session_t *s)
{
    unsigned char byte;
    int ret;
    if((ret=bufio_getbyte(buf,&byte,s))!=BUFIO_OK)
	return -1;
    return byte;
}

int     rec_get(spamtest_session_t *s, unsigned char  *buf, int maxsize,int *gotlen)
{
    int     type;
    int     len;
    int     len_byte;
    int     shift;
    int     to_read,has_read;
    int	    status;
    char    *bufp;

    if((status=check_init(s))!=STS_ERR_NO_ERR)
	return STS_ERR_INVALID_PARAMS;

    *gotlen = 0;
    if((type = read_byte(s->bufio,s))<0)
	{
	    return s->bufio->rbuf.pending_error ==BUFIO_TIMEOUT?STS_ERR_SND_RCV_TIMEOUT:STS_ERR_READ_ERR;
	}

    /*
     * Find out the record data length. Return an error result when the
     * record data length is malformed or when it exceeds the acceptable
     * limit.
     */
    for (len = 0, shift = 0; /* void */ ; shift += 7) {

	if (shift >= (int) (NBBY * sizeof(int))) 
 	    return STS_ERR_READ_ERR;

	if ((len_byte = read_byte(s->bufio,s)) == -1)
	    return s->bufio->rbuf.pending_error ==BUFIO_TIMEOUT
		?STS_ERR_SND_RCV_TIMEOUT
		:STS_ERR_READ_ERR;

	len |= (len_byte & 0177) << shift;
	if ((len_byte & 0200) == 0)
	    break;
    }
    s->logp(s,LOG_DEBUG,"Record type=%d(%c), size=%d",type,type,len);
    if (len < 0 || (maxsize > 0 && len > maxsize)) 
	{
	    s->logp(s,LOG_ERR,"Record too large for %d byte buffer",maxsize);
	    while (len-- > 0 && read_byte(s->bufio,s) != -1)
		/* void */ ;
	    return s->bufio->rbuf.pending_error ==BUFIO_TIMEOUT?STS_ERR_SND_RCV_TIMEOUT:STS_ERR_READ_ERR;
	}
    bufp = buf;
    to_read = len;
    while(to_read>0) 
	{
	    has_read = bufio_read(s->bufio,buf,to_read,s);
	    if(has_read <= 0)
		{
		    s->logp(s,LOG_ERR,"Read error: %d",has_read);
		    break;
		}
	    to_read-=has_read;
	    bufp+=has_read;
	}
    if(to_read >0)
	return s->bufio->rbuf.pending_error ==BUFIO_TIMEOUT?STS_ERR_SND_RCV_TIMEOUT:STS_ERR_READ_ERR;
    *gotlen = len;
    s->logp(s,LOG_DEBUG+2,"Got Record: %s",buf);
    return (type);
}
