/* -*- C -*-
 * Spamtest connection library
 * File: spamtest.c
 *      (C)  Alex Tutubalin <lexa@lexa.ru>
 *	and
 *	Mike Patutin <patutin@vist-v.ru>
 *	Apr-May 2003
 *
 * Created: Wed Apr 23 22:08:08 2003
 */

#include <stdio.h>
#include <syslog.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <syslog.h>
#include <errno.h>
#include <fcntl.h>
#include <ctype.h>

#include <sys/types.h>
#include <poll.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <sys/time.h>
#include <time.h>

#if (defined (_THREAD_SAFE) || defined (_REENTRANT)) && defined __USE_LOCKING
#include <pthread.h>
static pthread_mutex_t sts_mutex;
static pthread_mutex_t attr_mutex;

void local_sts_library_init(void)
{
    pthread_mutex_init(&sts_mutex,NULL);
    pthread_mutex_init(&attr_mutex,NULL);
}

static void __lock(void)
{
    pthread_mutex_lock(&sts_mutex);
}
static void __locka(void)
{
    pthread_mutex_lock(&attr_mutex);
}
static void __unlock(void)
{
    pthread_mutex_unlock(&sts_mutex);
}
static void __unlocka(void)
{
    pthread_mutex_unlock(&attr_mutex);
}
#else
static void __unlock(void){}
static void __lock(void){}
static void __unlocka(void){}
static void __locka(void){}
void local_sts_library_init(void){}

#endif


#ifdef AP_BUILD
#include "smtp-filter/common/rec_types.h"
#else
#include "rec_types.h"
#endif
#include "record2.h"
#include "spamtest.h"
#include "spamtest_internal.h"

/* send and recieve buffer size */
#define DFL_BUFFER_SIZE		8192

#ifndef AF_LOCAL
#define AF_LOCAL AF_UNIX
#endif



#define ATTRIBUTE_END 2
#define ATTRIBUTE_OK  1
#define ATTRIBUTE_WAIT 0
#define ATTRIBUTE_ERROR -1

#define STS_MAGIC 0x1f7a67bc

#ifdef NO_INET_ATON
int
inet_aton(cp, addr)
	register const char *cp;
	struct in_addr *addr;
{
	u_long parts[4];
	in_addr_t val;
	char *c;
	char *endptr;
	int gotend, n;

	c = (char *)cp;
	n = 0;
	/*
	 * Run through the string, grabbing numbers until
	 * the end of the string, or some error
	 */
	gotend = 0;
	while (!gotend) {
		errno = 0;
		val = strtoul(c, &endptr, 0);

		if (errno == ERANGE)	/* Fail completely if it overflowed. */
			return (0);
		
		/* 
		 * If the whole string is invalid, endptr will equal
		 * c.. this way we can make sure someone hasn't
		 * gone '.12' or something which would get past
		 * the next check.
		 */
		if (endptr == c)
			return (0);
		parts[n] = val;
		c = endptr;

		/* Check the next character past the previous number's end */
		switch (*c) {
		case '.' :
			/* Make sure we only do 3 dots .. */
			if (n == 3)	/* Whoops. Quit. */
				return (0);
			n++;
			c++;
			break;

		case '\0':
			gotend = 1;
			break;

		default:
			if (isspace((unsigned char)*c)) {
				gotend = 1;
				break;
			} else
				return (0);	/* Invalid character, so fail */
		}

	}

	/*
	 * Concoct the address according to
	 * the number of parts specified.
	 */

	switch (n) {
	case 0:				/* a -- 32 bits */
		/*
		 * Nothing is necessary here.  Overflow checking was
		 * already done in strtoul().
		 */
		break;
	case 1:				/* a.b -- 8.24 bits */
		if (val > 0xffffff || parts[0] > 0xff)
			return (0);
		val |= parts[0] << 24;
		break;

	case 2:				/* a.b.c -- 8.8.16 bits */
		if (val > 0xffff || parts[0] > 0xff || parts[1] > 0xff)
			return (0);
		val |= (parts[0] << 24) | (parts[1] << 16);
		break;

	case 3:				/* a.b.c.d -- 8.8.8.8 bits */
		if (val > 0xff || parts[0] > 0xff || parts[1] > 0xff ||
		    parts[2] > 0xff)
			return (0);
		val |= (parts[0] << 24) | (parts[1] << 16) | (parts[2] << 8);
		break;
	}

	if (addr != NULL)
		addr->s_addr = htonl(val);
	return (1);
}
#endif

const char *sts_strerror(int ercode)
{
    switch(ercode)
	{
	case STS_ERR_NO_ERR:
	    return "No error";
	case STS_ERR_OUT_OF_MEMORY:
	    return "Cannot allocate memory";
	case STS_ERR_UNABLE_TO_CONNECT:
	    return "Cannot connect to specified address";
	case STS_ERR_CONNECT_TIMEOUT:
	    return "Cannot connect: timeout";
	case STS_ERR_SND_RCV_TIMEOUT:
	    return "I/O timeout";
	case STS_ERR_INCORRECT_PROTOCOL:
	    return "Incorrect I/O protocol: unexpected response from filtering service";
	case STS_ERR_UNABLE_TO_RESOLVE:
	    return "Unable to resolve specified filter service address";
	case STS_ERR_READ_ERR:
	    return "Socket read error";
	case STS_ERR_ILLEGAL_CONFIG:
	    return "Illegal configuration data specified by user";
	case STS_ERR_WRITE_ERR:
	    return "Socket write error";
	case STS_ERR_INVALID_PARAMS:
	    return "sts_* function called with invalid parameter";
	case STS_ERR_NOT_INITED:
	    return "sts_* function called before sts_init()";
        case STS_ERR_ILLEGAL_ADDRESS:
            return "Illegal address";
        case STS_ERR_LIBRARY_ERR:
            return "Library error";
        case STS_ERR_SELECT_FD_ERR:
            return "Error with selecting file descriptor";
        case STS_ERR_BUFFER_NOT_PRESENT:
            return "Buffer is not present";
        case STS_ERR_BUFFER_IS_FULL:
            return "Buffer is full";
        case STS_ERR_GET_RECORD_SIZE:
            return "Cannot get record size";
        case STS_ERR_RECORD_TOO_LARGE:
            return "Record size is too large";
	default:
	    return "Unknown or undefined error";
	}
}


int sts_default_logger_log_level=LOG_ERR;

int check_init(spamtest_session_t *s)
{
    if(!s)
	return STS_ERR_INVALID_PARAMS;
    if(!s->store || !s->bufio || !s->rcpts || !s->attrs)
	return STS_ERR_OUT_OF_MEMORY;
    if(s->magic != STS_MAGIC)
	return STS_ERR_NOT_INITED;
    return STS_ERR_NO_ERR;
}

int calc_timeout(spamtest_session_t *s, int get_timediff)
{
    struct timeval tv;
    long w;
    if(get_timediff)
	{
	    gettimeofday(&tv,NULL);
	    w= s->rw_timeout-(tv.tv_sec-s->tvs.tv_sec)*1000+(tv.tv_usec-s->tvs.tv_usec)/1000;
	}
    else
	w = s->rw_timeout;
    if(w<=0) return STS_ERR_SND_RCV_TIMEOUT;
    s->timeout_nq = w > TIMEOUT_QUANT * MIN_TIMEOUT ? TIMEOUT_QUANT : (w/MIN_TIMEOUT)+1;
    s->timeout_curr = w/s->timeout_nq;
    return 0;

}

void init_timeout(spamtest_session_t *s)
{
    gettimeofday(&s->tvs,NULL);
    calc_timeout(s,0);
}



/*this function called as s->logp and can work with any user defined function */
static int default_logger(spamtest_session_t *s, int loglevel, char *fmt, ...)
{
    char logbuf[512];
    char logbuf2[128];
    va_list ap;
    if(loglevel>sts_default_logger_log_level) return 0;
    va_start(ap,  fmt);
    vsnprintf(logbuf,512, fmt, ap);
    va_end(ap);
    snprintf(logbuf2,128,"%s:",s->message_id?s->message_id:"No MSGID");
    syslog(loglevel>LOG_DEBUG?LOG_DEBUG:loglevel,"%s %s",logbuf2,logbuf);
    return 0;
}


/* translating given name to sockaddr_in structure */
/* on errors sin_port will be -1 */
static int normalize_inet_name(spamtest_session_t *s, const char *xname, struct sockaddr_in *addr) {

    struct hostent *h;
    struct servent *serv;
    char *host, *port, *first, *second,*name;
    
    if(!(name = heap_strdup(s->store,xname)))
	return STS_ERR_OUT_OF_MEMORY;

    if(!(first=strchr(name, ':')) || !(second=strrchr(name, ':')) || !(first == second))
        return STS_ERR_UNABLE_TO_RESOLVE;

    *first='\0';
    host = (char *)name;
    port = first+1;

    addr->sin_family = AF_INET;

    /* check for numeric addresses by normal way. now possible bugs */
    if (*host>='0' && *host<='9') 
	addr->sin_addr.s_addr = inet_addr(host);
    else {
	  h = gethostbyname(host);
	  if (!h)
	      return  STS_ERR_UNABLE_TO_RESOLVE;
	  addr->sin_addr = *((struct in_addr *) h->h_addr);
    }

    if (*port>='0' && *port<='9')
	addr->sin_port=htons (atoi(port));
    else {	
	  serv=getservbyname(port, (char*)0);
	  if (serv)
	      addr->sin_port=serv->s_port;
	  else
	      return STS_ERR_UNABLE_TO_RESOLVE;
    }

    return STS_ERR_NO_ERR;
}

static int wait_for_fd(int fd, long timeout_ms,spamtest_session_t *s)
{
    fd_set rfd;
    struct pollfd pfd;
    struct timeval tv;
    int state;

    if(s->flags & STS_USE_POLL)
	{
	    pfd.fd = fd;
#ifdef POLLDRNORM
	    pfd.events =(POLLIN|POLLRDNORM|POLLRDBAND);
#else
	    pfd.events =(POLLIN);
#endif
	    pfd.revents = 0;
	    state = poll(&pfd,1,timeout_ms);
	    if(state <0)
		{
		    s->logp(s, LOG_ERR, "Failed in poll: %s",strerror(errno));
		    return STS_ERR_UNABLE_TO_CONNECT;
		}
	    else if (state == 0)
		return STS_ERR_CONNECT_TIMEOUT;
	    else if (
#ifdef POLLDRNORM
		pfd.revents &(POLLIN|POLLRDNORM|POLLRDBAND)
#else
		pfd.revents & (POLLIN)
#endif
		)
		s->logp(s,LOG_DEBUG,"Connected, fd=%d",s->filter_fd);
	    return fd;
	}
    else
	{
	    FD_ZERO(&rfd);
	    FD_SET(fd,&rfd);
	    bzero(&tv,sizeof(tv));
	    tv.tv_sec = timeout_ms/1000;
	    tv.tv_usec = (timeout_ms%1000)*1000;
	    
	    state = select(s->filter_fd+1,&rfd,NULL,NULL,&tv);
	    if(state >0 && FD_ISSET(fd,&rfd))
		{
		    s->logp(s,LOG_DEBUG,"Connected, fd=%d",s->filter_fd);
		    return fd;
		}
	    else if(state == 0)
		{
		    return STS_ERR_CONNECT_TIMEOUT;
		}
	    else
		{
		    s->logp(s, LOG_ERR, "Failed in select: %s",strerror(errno));
		    return STS_ERR_UNABLE_TO_CONNECT;
		}
	}
}

/* connecting to TCP socket */
static int inet_connect(spamtest_session_t *s, const struct sockaddr_in *addr) {
    int retval;
    int flags;
    int state;
    int t=1;
    struct linger l;

    if((s->filter_fd = socket (AF_INET,  SOCK_STREAM,  0))<0)
	{
	    s->logp(s, LOG_ERR, "Cannot allocate socket: %s",strerror(errno));
	    return STS_ERR_UNABLE_TO_CONNECT;
	}

    if(-1==(flags = fcntl(s->filter_fd,  F_GETFL,  0)))
	{
	    close(s->filter_fd);
	    s->logp(s, LOG_ERR, "Cannot get flags for socket %d: %s",
		    s->filter_fd,strerror(errno));
	    return STS_ERR_UNABLE_TO_CONNECT;
	}
    if(-1== fcntl(s->filter_fd,  F_SETFL,  flags | O_NONBLOCK))
	{
	    close(s->filter_fd);
	    s->logp(s, LOG_ERR, "Cannot set NONBLOCK on socket %d: %s",
		    s->filter_fd,strerror(errno));
	    return STS_ERR_UNABLE_TO_CONNECT;
	}
    
    l.l_onoff=1;
    l.l_linger = 0;
    if(-1==setsockopt(s->filter_fd,SOL_SOCKET,SO_LINGER,&l,sizeof(l)))
	{
	    close(s->filter_fd);
	    s->logp(s, LOG_ERR, "Cannot set lingering close on socket %d: %s",
		    s->filter_fd,strerror(errno));
	    return STS_ERR_UNABLE_TO_CONNECT;
	}
    if(-1==setsockopt(s->filter_fd,IPPROTO_TCP,TCP_NODELAY,&t,sizeof(t))) 
	{
	    s->logp(s,LOG_ERR,"Could not set TCP_NODELAY on socket: %s",strerror(errno));
	    close(s->filter_fd);
	    return STS_ERR_UNABLE_TO_CONNECT;
	}

    if ((state = connect (s->filter_fd,(struct sockaddr *) addr, sizeof(*addr))) < 0) 
	{ 
	    if (errno != EINPROGRESS) 
		{
		    s->logp(s, LOG_ERR, "Failed to connect: %s",strerror(errno));
		    close(s->filter_fd);
		    s->filter_fd=-1;
		    return STS_ERR_UNABLE_TO_CONNECT;
		} 
	    else 
		{
		    retval= wait_for_fd(s->filter_fd,s->connect_timeout,s);
		    if(retval<0 && s->filter_fd >0)
		    {
		     close(s->filter_fd);
		     s->filter_fd=-1;
		    }
		    return retval;
		}
	}
    else 
	{
	    s->logp(s, LOG_DEBUG+1, "Connected immediately, fd=%d",s->filter_fd);
	    return s->filter_fd;
	}
}

/* connecting to unix socket */
static int unix_connect(spamtest_session_t *s, char *sockname) {

    int flags;
    int state;
    struct sockaddr_un port;

    port.sun_family = AF_LOCAL;
    strncpy(port.sun_path, sockname, sizeof(port.sun_path)-1);
    if((s->filter_fd = socket (AF_LOCAL,  SOCK_STREAM,  0))<0) 
	{
	    s->logp(s, LOG_ERR, "Cannot allocate socket: %s",strerror(errno));
	    return STS_ERR_UNABLE_TO_CONNECT;
	}

    if(-1==(flags = fcntl(s->filter_fd,  F_GETFL,  0)))
	{
	    close(s->filter_fd);
	    s->logp(s, LOG_ERR, "Cannot get flags for socket %d:%s",
		    s->filter_fd,strerror(errno));
	    return STS_ERR_UNABLE_TO_CONNECT;
	}
    if(-1== fcntl(s->filter_fd,  F_SETFL,  flags | O_NONBLOCK))
	{
	    close(s->filter_fd);
	    s->logp(s, LOG_ERR, "Cannot set NONBLOCK on socket %d: %s",
		    s->filter_fd,strerror(errno));
	    return STS_ERR_UNABLE_TO_CONNECT;
	}

    if ((state = connect (s->filter_fd, (struct sockaddr *) &port,  sizeof port)) < 0) 
	{
	    if (errno != EINPROGRESS) 
		{
		    s->logp(s, LOG_ERR, "Failed to connect: %s",strerror(errno));
		    close(s->filter_fd);
		    s->filter_fd=-1;
		    return STS_ERR_UNABLE_TO_CONNECT;
		} 
	    else 
		{
		    return wait_for_fd(s->filter_fd,s->connect_timeout,s);
		}
	}
    else
	{
	    s->logp(s, LOG_DEBUG, "Connected immediately, fd=%d",s->filter_fd);
	    return s->filter_fd;
	}
}


static void dump_buffer(spamtest_session_t *st)
{
    int i,len;
    char logbuf[10240];
    snprintf(logbuf,10240,"Buffer(%d): ",st->rb_ptr);
    len=strlen(logbuf);
    for(i=0;i<st->rb_ptr;i++)
	if(!st->attr_buf[i])
	    {
		strcat(logbuf,"\\0");
		len+=2;
	    }
	else
	    {
		snprintf(logbuf+len,10240-len,"%c",st->attr_buf[i]);
		len++;
	    }
    st->logp(st,LOG_DEBUG+2,"%s",logbuf);
}

static char *check_string(char *ptr,int len)
{
    int i;
    for(i=0;i<len;i++)
	if(!ptr[i]) return ptr+i;
    return NULL;
}

static int read_attribute(spamtest_session_t *st,char **aname,char **aval)
{
    unsigned char *p;
    int rstatus;
    
    rstatus = bufio_getbyte(st->bufio,st->attr_buf+st->rb_ptr,st);
    __locka();
    *aname = *aval = NULL;
    if(rstatus == BUFIO_OK)
	st->rb_ptr++;
    else
	{
	    __unlocka();
	    return ATTRIBUTE_ERROR;
	}

    /* checking for attributes */

    if (st->attr_buf[0] == '\0')
	  {
	    st->logp(st, LOG_INFO, "+Zero attribute, end of attribute exchange");
	    *aname = "";
	    st->rb_ptr = 0;
	    __unlocka();
	    return ATTRIBUTE_END;
	  }

    if (st->rb_ptr < 4)
	{
	    __unlocka();
	    return ATTRIBUTE_WAIT; /* at least one-char aname and one-char aval + two \0. */
	}
    if (st->rb_ptr >= STS_ABUF_S - 1) /* buffer overflow guard */
	{
	    __unlocka();
	    return ATTRIBUTE_ERROR;
	}

    if(sts_default_logger_log_level >= LOG_DEBUG+1)
	dump_buffer(st);


    p = check_string(st->attr_buf, st->rb_ptr);
    if (!p)
	{
	    __unlocka();
	    return ATTRIBUTE_WAIT; /* incomplete aname */
	}
    *aval = p + 1;
    st->logp(st, LOG_DEBUG + 2, "check_string() #2: total %d first %d last %d",
                st->rb_ptr, strlen(st->attr_buf) + 1,
                st->rb_ptr - (p + 1 - st->attr_buf));
    p = check_string(p + 1, st->rb_ptr - (p + 1 - st->attr_buf));
    if (!p)
      {
        *aval = NULL;
	__unlocka();
        return ATTRIBUTE_WAIT; /* incomplete aval */
      }
    *aname = st->attr_buf;

    st->rb_ptr = 0;
    st->logp(st, LOG_INFO, "<= GotAttr: %s|%s", *aname, *aval);

    __unlocka();
    return ATTRIBUTE_OK; /* one attribute read */
}


/**********************************************************************************/
/**********************  SESSION REALIZATION  *************************************/
/**********************************************************************************/


/* initialize spam test session */
int sts_init(spamtest_session_t *s) { 
    struct sockaddr_in addr;
    char *aname;
    char *aval;
    int ra,status;

    ra = check_init(s);
    if(ra == STS_ERR_INVALID_PARAMS)
	return ra; // invalid s pointer
    if(ra == STS_ERR_NO_ERR)
	return STS_ERR_NO_ERR; // already inited;

    if (!s->access_address || strlen(s->access_address) < 6 ) 
	return STS_ERR_ILLEGAL_CONFIG;


    /* install logger procedure */
    if ( s->logp == NULL )
	s->logp = (logger_proc)default_logger;

    if ( !s->bufsize )
	s->bufsize = DFL_BUFFER_SIZE;

    if(!(s->store = heap_init(2*s->bufsize+32*1024)))
	return STS_ERR_OUT_OF_MEMORY;
    if(!(s->rcpts = rcptlist_init(s->store)))
	{
	    heap_free(s->store);
	    s->store = NULL;
	    return STS_ERR_OUT_OF_MEMORY;
	}
    if(!(s->attrs = rcptlist_init(s->store)))
	{
	    heap_free(s->store);
	    s->store = NULL;
	    return STS_ERR_OUT_OF_MEMORY;
	}

    bzero(&(s->status),sizeof(s->status));
    s->status.storeptr = s->store;
    s->status.status = STS_CONTINUE;

    if ( s->rw_timeout == 0 ) 
	s->rw_timeout = STS_DEFAULT_SND_RCV_TIMEOUT;
    if ( s->connect_timeout == 0 ) 
	s->connect_timeout = STS_DEFAULT_CONNECT_TIMEOUT;

    s->recomended_size=1024000;
    s->filter_mode = 0;
    s->should_store_message = 1;
    s->message_id=NULL;
    s->filter_fd = -1;
    s->mail_from = NULL;
    s->envelope_sent = s->header_sent = s->body_sent = s->chunk_sent = 0;
    s->rb_ptr = 0;
    s->line_complete = 1;
    bzero(s->attr_buf,STS_ABUF_S);
    init_timeout(s);

    if(!strncmp(s->access_address,"unix:",5))
	s->filter_fd=unix_connect(s, s->access_address+5);
    else if(!strncmp(s->access_address,"tcp:",4))
	{
	    if (( status = normalize_inet_name(s,s->access_address+4, &addr))<0) 
		{
		    heap_free(s->store);
		    s->store = NULL;
		    return status;
		}
	    s->filter_fd=inet_connect(s, &addr);		
	}
    else
	{
	    heap_free(s->store);
	    s->store = NULL;
	    return STS_ERR_ILLEGAL_CONFIG;
	}

    if ( s->filter_fd <0 ) 
	{
	    s->logp(s, LOG_DEBUG, "Unable to connect ");
	    heap_free(s->store);
	    s->store = NULL;
	    return s->filter_fd;
	}

    if(NULL==(s->bufio = bufio_init(s->store,s->filter_fd,s->bufsize,s->bufsize)))
	{
	    heap_free(s->store);
	    s->store = NULL;
	    close(s->filter_fd);
	    s->filter_fd = -1;
	    return STS_ERR_OUT_OF_MEMORY;
	}
    if(s->flags & STS_USE_POLL)
	{
	    unsigned int fl = bufio_getflags(s->bufio);
	    bufio_setflags(s->bufio,fl|BUFIO_USE_POLL);
	}
    /* at this point we have guaranted connect,  now trying to get
     * filter state and recommended size ( if possible ) */    
    while((ra=read_attribute(s,&aname,&aval))==ATTRIBUTE_WAIT)
	;
    if(ra!=ATTRIBUTE_OK)
	{
	    s->logp(s, LOG_DEBUG, "Unable to read attr ");
	    heap_free(s->store);
	    s->store = NULL;
	    close(s->filter_fd);
	    s->filter_fd = -1;
	    return(STS_ERR_INCORRECT_PROTOCOL);
	}

    if(!strcmp(aname,"queue_id"))
	{
	    s->filter_mode = STS_SMTP_MODE;
	    s->should_store_message=0;
	    if(!(s->message_id = heap_strdup(s->store,aval)))
		{
		    heap_free(s->store);
		    s->store = NULL;
		    close(s->filter_fd);
		    s->filter_fd = -1;
		    return STS_ERR_OUT_OF_MEMORY;
		}
	    s->logp(s, LOG_DEBUG, "Session in SMTP mode\n");
	}
    else if (!strcmp(aname,"xqueue_id"))
	{
	    s->logp(s, LOG_DEBUG, "Session in action mode\n");
	    s->filter_mode = STS_ACTIONS_MODE;
	    s->should_store_message=1;
	    if(!(s->message_id = heap_strdup(s->store,aval)))
		{
		    heap_free(s->store);
		    s->store = NULL;
		    close(s->filter_fd);
		    s->filter_fd = -1;
		    return STS_ERR_OUT_OF_MEMORY;
		}
	    while((ra=read_attribute(s,&aname,&aval))==ATTRIBUTE_WAIT)
		;
	    if(ra == ATTRIBUTE_OK && !strcmp(aname,"continue"))
		{
		    s->recomended_size = atoi(aval);
		    s->logp(s, LOG_DEBUG, "Chunk size %d\n",s->recomended_size);
		}
	    else
		{
		    heap_free(s->store);
		    s->store = NULL;
		    close(s->filter_fd);
		    s->filter_fd = -1;
		    return(STS_ERR_INCORRECT_PROTOCOL);
		}
	}
    else
	{
	    s->logp(s, LOG_ERR, "Unknown initial attribute %s\n",aname);
	    heap_free(s->store);
	    s->store = NULL;
	    close(s->filter_fd);
	    s->filter_fd = -1;
	    return(STS_ERR_INCORRECT_PROTOCOL);
	}

    /* waiting for zero attr */  
    while((ra=read_attribute(s,&aname,&aval))==ATTRIBUTE_WAIT)
	;
    if(ra!=ATTRIBUTE_END)
	{
	    s->logp(s, LOG_ERR, "Cannot read zero attribute, %d\n",ra);
	    heap_free(s->store);
	    s->store = NULL;
	    close(s->filter_fd);
	    s->filter_fd = -1;
	    return(STS_ERR_INCORRECT_PROTOCOL);
	}
    s->magic = STS_MAGIC;
    return STS_ERR_NO_ERR;
}


/* already connected socket, just reinit all fields */

int sts_reuse(spamtest_session_t *s) { 
    char *aname;
    char *aval;
    int ra;
    msg_heap_t *newheap;
    bufio_t *newbufio;
    int i;
    ra = check_init(s);
    if(ra == STS_ERR_INVALID_PARAMS)
	return ra; // invalid s pointer
    
    if(ra ==  STS_ERR_OUT_OF_MEMORY)
	return STS_ERR_OUT_OF_MEMORY; // buffers are not set

    if(ra == STS_ERR_NOT_INITED)
	return STS_ERR_NOT_INITED; // try to use on not inited connection

    if(!(newheap = heap_init(2*s->bufsize+32*1024)))
	{
	    heap_free(s->store);
	    s->store = NULL;
	    close(s->filter_fd);
	    return STS_ERR_OUT_OF_MEMORY;
	}
    if(NULL==(newbufio = bufio_init(newheap,s->filter_fd,s->bufsize,s->bufsize)))
	{
	    heap_free(s->store);
	    heap_free(newheap);
	    s->store = NULL;
	    close(s->filter_fd);
	    s->filter_fd = -1;
	    return STS_ERR_OUT_OF_MEMORY;
	}
    for(i=s->bufio->rbuf.firstbyte;i<s->bufio->rbuf.lastbyte;i++)
	newbufio->rbuf.buf[newbufio->rbuf.lastbyte++]=s->bufio->rbuf.buf[i];

    heap_free(s->store);
    s->store = newheap;
    s->bufio = newbufio;


    if(!(s->rcpts = rcptlist_init(s->store)))
	{
	    heap_free(s->store);
	    close(s->filter_fd);
	    s->filter_fd=-1;
	    s->store = NULL;
	    return STS_ERR_OUT_OF_MEMORY;
	}
    if(!(s->attrs = rcptlist_init(s->store)))
	{
	    heap_free(s->store);
	    close(s->filter_fd);
	    s->filter_fd=-1;
	    s->store = NULL;
	    return STS_ERR_OUT_OF_MEMORY;
	}

    bzero(&(s->status),sizeof(s->status));
    s->status.storeptr = s->store;
    s->status.status = STS_CONTINUE;
    s->recomended_size=1024000;
    s->filter_mode = 0;
    s->should_store_message = 1;
    s->message_id=NULL;
    s->mail_from = NULL;
    s->envelope_sent = s->header_sent = s->body_sent = s->chunk_sent = 0;
    s->rb_ptr = 0;
    s->line_complete = 1;
    bzero(s->attr_buf,STS_ABUF_S);
    init_timeout(s);

    /* at this point we have guaranted connect,  now trying to get
     * filter state and recommended size ( if possible ) */    
    while((ra=read_attribute(s,&aname,&aval))==ATTRIBUTE_WAIT)
	;
    if(ra!=ATTRIBUTE_OK)
	{
	    s->logp(s, LOG_DEBUG, "Unable to read attr ");
	    heap_free(s->store);
	    s->store = NULL;
	    close(s->filter_fd);
	    s->filter_fd = -1;
	    return(STS_ERR_INCORRECT_PROTOCOL);
	}

    if(!strcmp(aname,"queue_id"))
	{
	    s->filter_mode = STS_SMTP_MODE;
	    s->should_store_message=0;
	    if(!(s->message_id = heap_strdup(s->store,aval)))
		{
		    heap_free(s->store);
		    s->store = NULL;
		    close(s->filter_fd);
		    s->filter_fd = -1;
		    return STS_ERR_OUT_OF_MEMORY;
		}
	    s->logp(s, LOG_DEBUG, "Session in SMTP mode\n");
	}
    else if (!strcmp(aname,"xqueue_id"))
	{
	    s->logp(s, LOG_DEBUG, "Session in action mode\n");
	    s->filter_mode = STS_ACTIONS_MODE;
	    s->should_store_message=1;
	    if(!(s->message_id = heap_strdup(s->store,aval)))
		{
		    heap_free(s->store);
		    s->store = NULL;
		    close(s->filter_fd);
		    s->filter_fd = -1;
		    return STS_ERR_OUT_OF_MEMORY;
		}
	    while((ra=read_attribute(s,&aname,&aval))==ATTRIBUTE_WAIT)
		;
	    if(ra == ATTRIBUTE_OK && !strcmp(aname,"continue"))
		{
		    s->recomended_size = atoi(aval);
		    s->logp(s, LOG_DEBUG, "Chunk size %d\n",s->recomended_size);
		}
	    else
		{
		    heap_free(s->store);
		    s->store = NULL;
		    close(s->filter_fd);
		    s->filter_fd = -1;
		    return(STS_ERR_INCORRECT_PROTOCOL);
		}
	}
    else
	{
	    s->logp(s, LOG_ERR, "Unknown initial attribute %s\n",aname);
	    heap_free(s->store);
	    s->store = NULL;
	    close(s->filter_fd);
	    s->filter_fd = -1;
	    return(STS_ERR_INCORRECT_PROTOCOL);
	}

    /* waiting for zero attr */  
    while((ra=read_attribute(s,&aname,&aval))==ATTRIBUTE_WAIT)
	;
    if(ra!=ATTRIBUTE_END)
	{
	    s->logp(s, LOG_ERR, "Cannot read zero attribute, %d\n",ra);
	    heap_free(s->store);
	    s->store = NULL;
	    close(s->filter_fd);
	    s->filter_fd = -1;
	    return(STS_ERR_INCORRECT_PROTOCOL);
	}
    s->magic = STS_MAGIC;
    return STS_ERR_NO_ERR;
}


int sts_set_sender(spamtest_session_t *s, const char *mail_from)
{
    int rc;
    if (!s || !mail_from)
        return STS_ERR_INVALID_PARAMS;
    if((rc = check_init(s)))
	return rc;
    if ( mail_from )
	{
	    if ( NULL == (s->mail_from = heap_strdup(s->store,mail_from)) )
		return STS_ERR_OUT_OF_MEMORY;
	    s->logp(s, LOG_DEBUG, "mail from: %s",s->mail_from);
	} 
    return STS_ERR_NO_ERR;    
}

int sts_add_recipient(spamtest_session_t *s, const char *rcpt)
{
    int rc;
    if (!s || !rcpt)
        return STS_ERR_INVALID_PARAMS;
    if((rc = check_init(s)))
	return rc;
    if (  rcpt )
	{
	    s->logp(s, LOG_DEBUG, "add one more recipient %s", rcpt);
	    if(-1==rcptlist_add(s->rcpts,rcpt))
		return STS_ERR_OUT_OF_MEMORY;
	}
    return STS_ERR_NO_ERR;    
}

int sts_add_attribute(spamtest_session_t *s, const char *attrname, const char *attrval)
{
    int rc;

    if (!s || !attrname || !attrval)
        return STS_ERR_INVALID_PARAMS;
    if((rc = check_init(s)))
	return rc;
    if ( attrname && attrval )
	{
	    char *v = heap_pstrcat(s->store,attrname,"=",attrval,NULL);
	    s->logp(s, LOG_DEBUG, "add one more attribute %s", v);
	    if(-1==rcptlist_add(s->attrs,v))
		return STS_ERR_OUT_OF_MEMORY;
	}
    return STS_ERR_NO_ERR;    
}

int sts_set_relay(spamtest_session_t *s, const char *relay)
{
    struct in_addr addr;
    int rc;
    if((rc = check_init(s)))
	return rc;
    if(relay && !strncmp(relay,"IP:",3))
	relay+=3;
    if(relay && *relay && inet_aton(relay,&addr))
	{
	    return  sts_add_attribute(s,"client_address",relay);
	}
    else
	return STS_ERR_ILLEGAL_ADDRESS;
}

/* sending to filter header of message */
int sts_prepare_envelope(spamtest_session_t *s, 
			 const char* mail_from,  
			 const char * rcpt,
			 const char *relay_IP,
			 int *stat) 
{ 
    int rc;
    if(stat)
	*stat = 0;

    if((rc = check_init(s)))
       return rc;
    if(mail_from)
	if((rc=sts_set_sender(s,mail_from))!=0)
	    return rc;
    if (rcpt)
	if((rc=sts_add_recipient(s,rcpt))!=0)
	    return rc;
    if(relay_IP && *relay_IP)
	{
	    rc = sts_set_relay(s,relay_IP);
	    if (rc == STS_ERR_ILLEGAL_ADDRESS && stat) 
		*stat|=STS_ERR_ILLEGAL_RELAY;
	    if(rc && rc != STS_ERR_ILLEGAL_ADDRESS)
		return rc;
	}
    return STS_ERR_NO_ERR;
}


static int process_action_line(spamtest_session_t *s, message_status_t *msg,char *buf,int buflen)
{
    char *p1,*p2;
    int status;
    buf[buflen]='\0';
    s->logp(s,LOG_DEBUG,"Action: %s",buf);
    if(!strncmp(buf,"NewHeader:",10))
	{
	    p1 = buf+10;
	    if(p1 && *p1 && (p2=strchr(p1,':')))
		{
		    *(p2++)='\0';
		    if((status=add_action(msg,s->store,STS_HEADER_NEW,p1,p2))<0)
			return status;
		    s->logp(s,LOG_DEBUG,"New Header: %s|%s",p1,p2);
		    return 0;
		}
	}
    else if(!strncmp(buf,"AddHeader:",10))
	{
	    p1 = buf+10;
	    if(p1 && *p1 && (p2=strchr(p1,':')))
		{
		    *(p2++)='\0';
		    if((status=add_action(msg,s->store,STS_HEADER_ADD,p1,p2))<0)
			return status;
		    s->logp(s,LOG_DEBUG,"Add Header: %s|%s",p1,p2);
		    return 0;
		}
	}
    else if(!strncmp(buf,"PrependHeader:",14))
	{
	    p1 = buf+14;
	    if(p1 && *p1 && (p2=strchr(p1,':')))
		{
		    *(p2++)='\0';
		    if((status=add_action(msg,s->store,STS_HEADER_PREPEND,p1,p2))<0)
			return status;
		    s->logp(s,LOG_DEBUG,"Prepend Header: %s|%s",p1,p2);
		    return 0;
		}
	}
    else if(!strncmp(buf,"DelHeader:",10))
	{
	    p1 = buf+10;
	    if(*(p1+1))
		{
		    if((status=add_action(msg,s->store,STS_HEADER_DEL,p1,NULL))<0)
			return status;
		    s->logp(s,LOG_DEBUG,"Del Header: %s",p1);
		    return 0;
		}
	}
#if 0
    else if(!strncmp(buf,"DoSyslog:",9))
	{
	    p1 = buf+10;
	    if(*(p1+1))
		{
		    if((status=add_action(msg,s->store,STS_DOSYSLOG,p1,NULL))<0)
			return status;
		    s->logp(s,LOG_DEBUG,"Syslog: %s",p1);
		    return 0;
		}
	}
#endif
    else if(!strncmp(buf,"ChangeHeader:",13))
	{
	    p1 = buf+13;
	    if(p1 && *p1 && (p2=strchr(p1,':')))
		{
		    *(p2++)='\0';
		    if((status=add_action(msg,s->store,STS_HEADER_DEL,p1,NULL))<0)
			return status;
		    if((status=add_action(msg,s->store,STS_HEADER_ADD,p1,p2))<0)
			return status;
		    s->logp(s,LOG_DEBUG,"Change Header: %s|%s",p1,p2);
		    return 0;
		}
	}
    else if(!strncmp(buf,"ChgHeader:",10))
	{
	    p1 = buf+10;
	    if(p1 && *p1 && (p2=strchr(p1,':')))
		{
		    *(p2++)='\0';
		    if((status=add_action(msg,s->store,STS_HEADER_DEL,p1,NULL))<0)
			return status;
		    if((status=add_action(msg,s->store,STS_HEADER_ADD,p1,p2))<0)
			return status;
		    s->logp(s,LOG_DEBUG,"Change Header: %s|%s",p1,p2);
		    return 0;
		}
	}
    else
	{
	    s->logp(s,LOG_WARNING,"Cannot perform action %s, unknown type",buf);
	    return 0;
	}
    
    /* known action, but cannot parse */
    return -1;
}

static void read_final_status(spamtest_session_t *st)
{
    char *aname,*aval;
    int  rc;
    while((rc=read_attribute(st,&aname,&aval))!=ATTRIBUTE_ERROR)
	{
	    if(rc == ATTRIBUTE_END)
		break;
	}
}



/* after gettind status,  reason and count we need to recieve processed messages */
int read_processed_messages(spamtest_session_t *s, int count) { 

    int type=-1;
    int res = 0;
    message_status_t *curmsg=NULL;
    int gotln;
    if(count<1)
	return STS_ERR_NO_ERR;

    __lock();
    if(!s->record_buf)
	s->record_buf = heap_alloc(s->store,MAX_RECORD_LEN+1);
    
    s->logp(s, LOG_DEBUG, "read_proccessed_messages started --->");

    while ((count > 0) && ((type = rec_get(s, s->record_buf, MAX_RECORD_LEN, &gotln)) > 0)  ) 
	{

	    s->record_buf[gotln]='\0';
	    s->logp(s,LOG_DEBUG,"read_processed_messages: read one record");
	    
	    switch (type)
		{
		case REC_TYPE_END:
		    if(--count == 0)
			{
			    read_final_status(s);
			    __unlock();
			    return STS_ERR_NO_ERR;
			}
		    break;
		    curmsg = NULL;
		case REC_TYPE_SIZE:
		    curmsg = make_new_status(&s->status);
		    curmsg->action = STS_ACTION_CHANGE;
		    break;
	        case REC_TYPE_ACTION:
		    if(curmsg)
			{
			    if(!strncmp("Bounce",s->record_buf,6))
				{
				    curmsg->action = STS_ACTION_BOUNCE;
				}
			    else
				{
				    if((res=process_action_line(s,curmsg,s->record_buf,gotln)))
					s->logp(s,LOG_INFO,"Cannot process action %s",s->record_buf);
				}
			}
		    else
			s->logp(s,LOG_WARNING,"Records out of order: ACTION before SIZE");
		    break;
		case REC_TYPE_RCPT:
		    if(curmsg)
			{
			    rcptlist_addn(curmsg->rcpts,s->record_buf,gotln);
			    s->logp(s,LOG_DEBUG+2,"rcpt: count=%i type=%i",count,type);
			}
		    else
			s->logp(s,LOG_WARNING,"Records out of order: RCPT before SIZE");
		    break;
		case REC_TYPE_FROM:
		    if(curmsg)
			{
			    curmsg->mailfrom = heap_strdupn(s->store,s->record_buf,gotln);
			    s->logp(s,LOG_DEBUG+2,"FROM: count=%i type=%i",count,type);
			}
		    else
			s->logp(s,LOG_WARNING,"Records out of order: FROM before SIZE");
		    break;
		case REC_TYPE_NORM:
		    if(curmsg)
			{
			    if(curmsg->action == STS_ACTION_BOUNCE)
				{
				    if(-1==msgstat_addbody(curmsg,s->store,s->record_buf,gotln))
					{
					    __unlock();
					    return STS_ERR_OUT_OF_MEMORY;
					}
				    if(-1==msgstat_addbody(curmsg,s->store,"\n",1))
					{
					    __unlock();
					    return STS_ERR_OUT_OF_MEMORY;
					}
				}
			    else
				s->logp(s,LOG_DEBUG,"Ignoring NORM record for Action-type message");
			}
		    else
			s->logp(s,LOG_WARNING,"Records out of order: NORM before SIZE");
		    break;
		case REC_TYPE_CONT:
		    if(curmsg)
			{
			    if(curmsg->action == STS_ACTION_BOUNCE)
				{
				    if(-1==msgstat_addbody(curmsg,s->store,s->record_buf,gotln))
					{
					    __unlock();
					    return STS_ERR_OUT_OF_MEMORY;
					}
				}
			    else
				s->logp(s,LOG_DEBUG,"Ignoring CONT record for Action-type message");
			}
		    else
			s->logp(s,LOG_WARNING,"Records out of order: CONT before SIZE");
		    break;
		default:
		    s->logp(s,LOG_DEBUG,"Unexpected record type %d(%c)",type,type);
	    }
	}
    if ( type < 0 ) 
	{
	    s->logp(s,LOG_ERR,"get_rec got negative type %d",type);
	    __unlock();
	    return type;
	}
    /* should not be there  */
    s->logp(s,LOG_WARNING,"Protocol misorder when reading final message");
    __unlock();
    read_final_status(s);
    return STS_ERR_NO_ERR;
}

/* after sending information to filter,  get filter directives to
 * proceed message */
/* called only in Actions mode */
int get_message_status(spamtest_session_t *s) {

    int  rc;
    char *aname,*aval;

    __lock();
    if(!s|| !s->bufio)
	{
	    __unlock();
	    return STS_ERR_INVALID_PARAMS;
	}
    if(s->filter_mode != STS_ACTIONS_MODE)
	{
	    s->logp(s,LOG_ERR,"get_message_status called not in Actions mode:");
	    __unlock();
	    return STS_ERR_INCORRECT_PROTOCOL;
	}
    
    s->logp(s, LOG_DEBUG, "Waiting for 1st attribute");
    __unlock();
    while((rc=read_attribute(s,&aname,&aval))==ATTRIBUTE_WAIT)
	;
    __lock();
    if(rc!=ATTRIBUTE_OK)
	{
	    s->logp(s,LOG_ERR,"Cannot read attribute\n");
	    __unlock();
	    return STS_ERR_READ_ERR;
	}
    if(!strcmp(aname,"continue"))
	{
	    s->status.status = STS_CONTINUE;
	    s->recomended_size = atoi(aval);
	    s->logp(s, LOG_DEBUG, "Got continue attr, size=%d",s->recomended_size);
	    
	    s->logp(s, LOG_DEBUG, "Waiting for zero attribute");
	    while((rc=read_attribute(s,&aname,&aval))==ATTRIBUTE_WAIT)
		;
	    if(rc!=ATTRIBUTE_END)
		{
		    s->logp(s,LOG_ERR,"Cannot read attribute: %d\n",rc);
		    __unlock();
		    return STS_ERR_READ_ERR;
		}
	    __unlock();
	    return STS_ERR_NO_ERR;
	}
    else if(!strcmp(aname,"xstatus"))
	{
	    // xstatus got, process remaining of message
	    s->status.status = atoi(aval);
	    // read remaining attributes

	    s->logp(s,LOG_DEBUG,"xstatus got: Waiting for final attribute in ACTIONS mode");
	    while((rc=read_attribute(s,&aname,&aval))!=ATTRIBUTE_ERROR)
		{
		    if(rc == ATTRIBUTE_WAIT)
			continue;
		    else if (rc == ATTRIBUTE_OK)
			{
			    s->logp(s,LOG_DEBUG,"Got attribute %s %d",aname,s->filter_fd);
			    if(0==strcmp(aname,"reason"))
				{
				    s->logp(s,LOG_DEBUG,"Got reason %s",aval);
				    s->status.reason = heap_strdup(s->store,aval);
				}
			    else if(!strcmp(aname,"count"))
				{
				    s->logp(s,LOG_DEBUG,"Got count %s",aval);
				    s->status.count = atoi(aval);
				    if(s->status.status == STS_ACCEPT && s->status.count==0)
					s->status.status = STS_BLACKHOLE;
				}
			    else
				{
				    s->logp(s,LOG_ERR,
					    "Unexpected final attribute %s=%s in ACTIONS mode",
					    aname,aval);
				    s->status.status = STS_SENDERROR;
				    __unlock();
				    return  STS_ERR_INCORRECT_PROTOCOL;
				}
			}
		    else if(rc == ATTRIBUTE_END)
			{
			    s->logp(s,LOG_DEBUG,"Got final attribute in ACTIONS mode");
			    __unlock();
			    return read_processed_messages(s, s->status.count);
			}
		}
	}
    else
	{
	    s->logp(s,LOG_ERR, "Unexpected final attribute %s=%s in ACTIONS mode",
		    aname,aval);
	    s->status.status = STS_SENDERROR;
	    __unlock();
	    return STS_ERR_INCORRECT_PROTOCOL;
	}
    s->status.status = STS_SENDERROR;
    __unlock();
    return STS_ERR_INCORRECT_PROTOCOL;
}

/* real sending message header to filter */

static
spamtest_status_t errvar = {STS_SENDERROR,"Invalid Params",STS_ERR_INVALID_PARAMS,0,NULL,0,NULL};


spamtest_status_t *sts_send_envelope(spamtest_session_t *s) {

    int opstatus =0;
    int i;
    int rc;
    if(!s)
	return &errvar;
    if((rc = check_init(s)))
	{
	    s->status.status = STS_SENDERROR;
	    s->status.errorcode = rc;
	    return &(s->status);
	}

    __lock();
    if(!s->record_buf)
	s->record_buf = heap_alloc(s->store,MAX_RECORD_LEN+1);
    s->logp(s, LOG_DEBUG, "envelope start");
    snprintf(s->record_buf, MAX_RECORD_LEN, REC_TYPE_SIZE_FORMAT, 0L, 0L, 0L);
    if ( (opstatus = rec_put(s, REC_TYPE_SIZE, s->record_buf, strlen(s->record_buf))) < 0 ) {
	s->status.status = STS_SENDERROR;
	s->status.errorcode = opstatus;
	__unlock();
	return &(s->status);
    }
    snprintf(s->record_buf,MAX_RECORD_LEN,"%ld",time(NULL));
    if ( (opstatus = rec_put(s, REC_TYPE_TIME, s->record_buf, strlen(s->record_buf))) < 0 ) {
	s->status.status = STS_SENDERROR;
	s->status.errorcode = opstatus;
	__unlock();
	return &(s->status);
    }

    s->logp(s,LOG_DEBUG+1,"Sent envelope format = %s",s->record_buf);

    // sending attributes
    for(i=0;i<rcptlist_count(s->attrs);i++)
	{
	    char *attr = rcptlist_getn(s->attrs,i);
	    if(!attr||!*attr) continue;
	    if ( (opstatus = rec_put(s, REC_TYPE_ATTR, attr, strlen(attr)))  < 0 ) 
		{
		s->status.status = STS_SENDERROR;
		s->status.errorcode = opstatus;
		__unlock();
		return &(s->status);
		}
	    s->logp(s,LOG_DEBUG+1,"Sent attribute %s",attr);
	}

    if(1){ // we'll always send From address
	char *tosend = s->mail_from?s->mail_from:"";
	char *unq;
	int l;
	if(s->mail_domain && strlen(tosend)>0)
	    {
		unq = heap_alloc(s->store,l=strlen(s->mail_from)+strlen(s->mail_domain)+16);
		if(unq && !rfc822_unquote_one(unq,l,s->mail_from,s->mail_domain))
		    tosend = unq;
	    }
	if ( (opstatus = rec_put(s, REC_TYPE_FROM, tosend, strlen(tosend))) < 0 ) {
	    s->status.status = STS_SENDERROR;
    	    s->status.errorcode = opstatus;
	    __unlock();
    	    return &(s->status);
	}
    }


    s->logp(s,LOG_DEBUG+1,"Sent envelope mail from = %s",s->mail_from);
    {
	recipient_list_t *tosend = s->rcpts;
	recipient_list_t *unq;
	if(s->mail_domain && (unq=rfc822_unquote_list(s->rcpts,s->mail_domain)))
	    tosend = unq;
	
	for(i=0;i<rcptlist_count(tosend);i++)
	{
	    char *rcpt = rcptlist_getn(tosend,i);
	    if ( ( opstatus = rec_put(s, REC_TYPE_RCPT, rcpt, strlen(rcpt))) < 0 ) 
		{
		    s->status.status = STS_SENDERROR;
		    s->status.errorcode = opstatus;
		    __unlock();
		    return &(s->status);
		}
	    s->logp(s,LOG_DEBUG,"Sent envelope rcpt = %s",rcpt);
	}
    }

    s->envelope_sent = 1;
    s->logp(s, LOG_DEBUG, "finish send envelope");
    __unlock();
    return &(s->status);
}

/* sending message's body (part of body),  if we need - will send header */
spamtest_status_t* sts_send_body(spamtest_session_t *s,
				 const char *msgbody, 
				 size_t msglen, 
				 int flags) 
{ 

    const char *start, *end;
    size_t tosend,recsize;
    char *xmx = "X-Mailbox-Line: ";

    int status = 0;
    spamtest_status_t *st;
    int rc;
    if(!s)
	return &errvar;
    if((rc = check_init(s)))
	{
	    s->status.status = STS_SENDERROR;
	    s->status.errorcode = rc;
	    return &(s->status);
	}
    
    if(s->status.status != STS_CONTINUE)
	return &(s->status);

    __lock();
    if(flags &STS_REINIT_TIMEOUT)
	init_timeout(s);

    if ( s->body_sent == 1 ) 
	{
	    __unlock();
	    return &(s->status);
	}
    
    if(!s->envelope_sent)
	{
	    st = sts_send_envelope(s);
	    if(st->errorcode)
		{
		    __unlock();
		    return st;
		}
	}

    s->logp(s, LOG_DEBUG, "send_body: (%d) %s",msglen,msgbody);

    s->chunk_sent += msglen;
    
    start = msgbody;
    tosend = msglen;
    while (tosend > 0)
	{
	    end = memchr(start,'\n',tosend>MAX_RECORD_LEN-1?MAX_RECORD_LEN-1:tosend);
	    if(end)
		{
		    recsize = end-start;
		    if(recsize>=0)
			{
			    if(!s->header_sent && s->line_complete && !strncmp(start,"From ",5))
				{
				    /* hiding From line */
				    if ((status = rec_put(s,REC_TYPE_CONT,xmx,strlen(xmx))) < 0 ) {
					s->status.status = STS_SENDERROR;
					s->status.errorcode = status;
					__unlock();
					return &(s->status);
				    }
				}
			    if ((status = rec_put(s, REC_TYPE_NORM, start, recsize)) < 0 ) {
				s->status.status = STS_SENDERROR;
				s->status.errorcode = status;
				__unlock();
				return &(s->status);
			    }
			}
		    if(recsize == 0 && !s->header_sent)
			{
			    s->header_sent = 1;
			}
		    else if (recsize==1 && !s->header_sent && *start=='\r')
			{
			    s->header_sent = 1;
			}
		    tosend-=recsize+1;
		    start+=recsize+1;
		    s->chunk_sent += recsize+1;
		    s->line_complete = 1; // was _NORM record */
		}
	    else
		{
		    recsize=tosend>MAX_RECORD_LEN-1?MAX_RECORD_LEN-1:tosend;
		    if ((status = rec_put(s, REC_TYPE_CONT, start, recsize)) < 0 ) {
			s->status.status = STS_SENDERROR;
			s->status.errorcode = status;
			__unlock();
			return &(s->status);
		    }
		    tosend-=recsize;
		    start+=recsize;
		    s->chunk_sent += recsize;
		    s->line_complete = 0; // was _CONT record */
		}
	}
    __unlock();
    if ( flags & STS_FINAL_CHUNK ) 
	{
	    if ( sts_body_end(s,flags)->errorcode < 0 )
		    return &(s->status);
	} 

    if ( (s->filter_mode == STS_SMTP_MODE ) ) 
	{
	    s->status.status = STS_CONTINUE;
	    return &(s->status);
        } 

    if ( ! s->header_sent ) 
	{
	    s->logp(s, LOG_DEBUG+1, "Not sending REC_TYPE_BLOCKFIN before headers were sent\n");
	    s->status.status = STS_CONTINUE;
	    return &(s->status);
	}
    
    if(! (flags & STS_FLUSH_DATA) && (s->chunk_sent < s->recomended_size))
	{
	    s->logp(s, LOG_DEBUG, "Not sending REC_TYPE_BLOCKFIN, not much data sent\n");
	    s->status.status = STS_CONTINUE;
	    return &(s->status);
	    
	}
    s->logp(s, LOG_DEBUG, "Sending REC_TYPE_BLOCKFIN!\n");
    s->chunk_sent = 0;
    __lock();
    if ( (status = rec_put(s, REC_TYPE_BLOCKFIN, "", 0)) < 0 ) 
	{
	    s->logp(s, LOG_DEBUG, "Sent blockfin, got %d",status);
	    s->status.status = STS_SENDERROR;
	    s->status.errorcode = status;
	    __unlock();
	    return &(s->status);
	}
    __unlock();
    if ( (status = get_message_status(s)) < 0 ) 
	{
	    s->logp(s, LOG_DEBUG, "message status got %d",status);
	    s->status.status = STS_SENDERROR;
	    s->status.errorcode = status;
	    return &(s->status);
	}

    s->logp(s, LOG_DEBUG+1, "ending to send body");
    return &(s->status);
}


static spamtest_status_t *read_smtp_status(spamtest_session_t *s)
{
    char *aname,*aval;
    int  rc;
    s->logp(s,LOG_DEBUG,"Waiting for final attribute in SMTP mode");
    while((rc=read_attribute(s,&aname,&aval))!=ATTRIBUTE_ERROR)
	{
	    if(rc == ATTRIBUTE_WAIT)
		continue;
	    else if (rc == ATTRIBUTE_OK)
		{
		    if(!strcmp(aname,"status"))
			{
			    s->logp(s,LOG_DEBUG,"Got status %s",aval);
			    if(0==(s->status.status = atoi(aval)))
				s->status.status = STS_SMTP_ACCEPT;
			}
		    else if(!strcmp(aname,"reason"))
			{
			    s->logp(s,LOG_DEBUG,"Got reason %s",aval);
			    s->status.reason = heap_strdup(s->store,aval);
			}
		    else
			{
			    s->logp(s,LOG_ERR,
				    "Unexpected final attribute in SMTP mode: %s/%s",
				    aname,aval);
			    s->status.errorcode = STS_ERR_INCORRECT_PROTOCOL;
			    s->status.status = STS_SENDERROR;
			    return &(s->status);
			}
		}
	    else if(rc == ATTRIBUTE_END)
		{
		    s->logp(s,LOG_DEBUG,"Got final attribute in SMTP mode");
		    s->status.errorcode = 0;
		    return &(s->status);
		}
	}
    s->logp(s,LOG_ERR,"Cannot get final attribute in SMTP mode");
    s->status.errorcode = STS_ERR_READ_ERR;
    s->status.status = STS_SENDERROR;
    return &(s->status);
}

spamtest_status_t* sts_body_end(spamtest_session_t *s,int flags) { 
    
    int status = 0;
    int rc;
    __lock();
    if(!s)
	{
	    __unlock();
	    return &errvar;
	}
    if((rc = check_init(s)))
	{
	    s->status.status = STS_SENDERROR;
	    s->status.errorcode = rc;
	    __unlock();
	    return &(s->status);
	}

    if(s->status.status != STS_CONTINUE)
	{
	    __unlock();
	    return &(s->status);
	}

    if(!(flags & STS_REINIT_TIMEOUT))
	init_timeout(s);

    if ( s->body_sent == 0 ) {
	s->body_sent = 1;
	s->logp(s, LOG_DEBUG, "Sending REC_TYPE_END!\n");

	if ( ( status = rec_put(s, REC_TYPE_END, "", 0)) < 0 ) {
	    s->status.errorcode = status;
	    __unlock();
	    return &(s->status);
	}

	if(s->filter_mode == STS_SMTP_MODE)
	    {
		__unlock();
		return read_smtp_status(s);
	    }
	else
	    {
		__unlock();
		if ( (status = get_message_status(s)) < 0 ) {
		    s->status.status = STS_SENDERROR;
		    s->status.errorcode = status;
		    return &(s->status);
		}
	    }
    }
    __unlock();
    return &(s->status);
}



int sts_close(spamtest_session_t *s) 
{ 
    int off;
    if((off = check_init(s))) return off;
    s->magic=0;
    close(s->filter_fd);
    s->filter_fd = -1;
    bzero(&s->status,sizeof(s->status)); 
    heap_free(s->store);
    return STS_ERR_NO_ERR;
}

void heap_library_init(void);
void local_bufio_library_init(void);

void sts_library_init(void)
{
    heap_library_init();
    local_bufio_library_init();
    local_sts_library_init();
}
