/* -*- C -*-
 * File: spamtest_async.c
 *
 * Created: Wed Sep 29 12:43:00 2004
 */

// Single session objects

#include <sys/time.h>
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <stdlib.h>
#ifdef SOLARIS
#include <alloca.h>
#endif
#include <sys/stat.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include <stdio.h>
#include <poll.h>
#include <time.h>
#include <fcntl.h>


#include "kas_internal.h"
#include "kas_defs.h"
#include "spamtest_async.h"

EXTERN_C_BEGIN
#ifndef AF_LOCAL
#define AF_LOCAL AF_UNIX
#endif

void clear_metadata(kas_metadata_t *m);

#define PROGRESS_CALLBACK_SIMPLE(sesptr,st) do{	\
	kas_progress_info_t cpi; \
	bzero(&cpi,sizeof(cpi));\
	cpi.pid = sesptr->pid;\
	cpi.state=st;\
	if(sesptr->progress_cb)\
	    (*sesptr->progress_cb)(sesptr,&cpi,sesptr->progress_cb_param);\
    }while(0)




#define STATUS_CALLBACK(sesptr) do{ \
	sesptr->current_status.scaner_pid = sesptr->pid;	       \
	if(sesptr->status_cb)					       \
	    (*sesptr->status_cb)(sesptr,&sesptr->old_status,	       \
				 &sesptr->current_status,sesptr->status_cb_param); \
	memmove(&sesptr->old_status,&sesptr->current_status,sizeof(sesptr->old_status)); \
    }while(0)


static int diff_msec (struct timeval *old, struct timeval *now)
{
    return (now->tv_sec-old->tv_sec)*1000+(now->tv_usec-old->tv_usec)/1000;
}

void STATUS_CALLBACK_EVENT(kas_session_t *sesptr,enum kas_status_value status,
			   enum kas_extdstatus_value estatus)
{
    kas_status_t kstatus;
    kstatus.status = status;
    kstatus.extd_status = estatus;
    kstatus.scaner_pid = sesptr->pid;
    if(status==KAS_STATUS_SCAN_END && sesptr->scan_start.tv_sec>0)
	{
	    struct timeval now;
	    size_t msec;
	    gettimeofday(&now,NULL);
	    msec = diff_msec(&sesptr->scan_start,&now);
	    kstatus.elapsed_time.tv_sec = msec/1000;
	    kstatus.elapsed_time.tv_usec = msec%1000;
	    bzero(&sesptr->scan_start,sizeof(sesptr->scan_start));
	}
    else
	{
	    kstatus.elapsed_time.tv_sec = 0;
	    kstatus.elapsed_time.tv_usec = 0;
	}
    if(sesptr->status_cb)
	(*sesptr->status_cb)(sesptr,&sesptr->old_status,       
		     &kstatus,sesptr->status_cb_param);
}
// private cb's
// read
static int read_connected(kas_session_t *sesptr);
static int read_greeting(kas_session_t *sesptr);
//static int read_filter_response(kas_session_t *sesptr);
static int read_pingresponse(kas_session_t *sesptr);
static int read_response_attrs(kas_session_t *sesptr);

// write
static int write_ping(kas_session_t *sesptr);

// pvt func
static int KasCheckSession(kas_session_t *sesptr);
static int KasSReadData(kas_session_t *sesptr);
static int KasSWriteData(kas_session_t *sesptr);
static int  check_timeout(kas_session_t *sesptr);
static int  check_timeout2(kas_session_t *sesptr,struct timeval *now);
static int KasSProcessIOResult(kas_session_t *sesptr,int socket_read, int socket_write);
static kas_metadata_t *clone_metadata(kas_metadata_t *m);
static int     rec_get(kas_session_t *s,unsigned char *buf, int maxsize, int *gotlen);



kas_session_t *
KasSCreateSession(kas_session_t *sesptr,
		  unsigned int buf_size,
		  int read_timeout,
		  int write_timeout,
		  unsigned int max_spam_size_in_bytes,
		  unsigned int scanner_data_limit_in_kbytes,
		  kas_cb_single_status_t *status_cb,
		  void *status_cb_data)
{
    int was_malloc=0;
    int singlebuf_sz;

    if(!sesptr)
	{
	    sesptr = malloc(sizeof(kas_session_t));
	    was_malloc++;
	}
    if(!sesptr)
	{
	    return NULL;
	}
    bzero(sesptr,sizeof(kas_session_t));

    if(buf_size<8192)
	buf_size = 8192;
    singlebuf_sz = (buf_size/3/1024+1)*1024;
    sesptr->wr_buf.data=malloc(singlebuf_sz);
    sesptr->rd_buf.data=malloc(singlebuf_sz);
    sesptr->fi_buf.data=malloc(singlebuf_sz);
    if(!sesptr->wr_buf.data || !sesptr->rd_buf.data || !sesptr->fi_buf.data)
	{
	    if(sesptr->wr_buf.data)
		free(sesptr->wr_buf.data);
	    if(sesptr->rd_buf.data)
		free(sesptr->rd_buf.data);
	    if(sesptr->fi_buf.data)
		free(sesptr->fi_buf.data);
	    if(was_malloc) free(sesptr);
	    return NULL;
	}
    sesptr->read_func = sesptr->write_func = NULL;
    sesptr->wr_buf.alloc = sesptr->rd_buf.alloc= sesptr->fi_buf.alloc = singlebuf_sz;
    sesptr->wr_buf.used = sesptr->rd_buf.used = sesptr->fi_buf.used = 0;
    sesptr->status_cb = status_cb;
    sesptr->status_cb_param = status_cb_data;
    sesptr->datafile_fd = sesptr->socket = -1;
    sesptr->wasmalloc = was_malloc;
    sesptr->max_spam_size = max_spam_size_in_bytes;
    sesptr->scanner_rss_limit = scanner_data_limit_in_kbytes;
    sesptr->read_timeout = read_timeout;
    sesptr->write_timeout = write_timeout;
    sesptr->standard_proto = 0;

    sesptr->old_status.status = sesptr->current_status.status = KAS_STATUS_DISCONNECTED;
    bzero(&sesptr->last_io,sizeof(sesptr->last_io));
    sesptr->magick = KAS_MAGIC;
    STATUS_CALLBACK(sesptr);
    return sesptr;
}

static int KasCheckSession(kas_session_t *sesptr)
{
    if(!sesptr) return KASerrPARAMS_NULLPTR;
    if(sesptr->magick != KAS_MAGIC) return KASerrPARAMS_NOTINITED;
    if(!sesptr->wr_buf.alloc || !sesptr->wr_buf.data) return KASerrPARAMS_NOTINITED;
    if(!sesptr->rd_buf.alloc || !sesptr->rd_buf.data) return KASerrPARAMS_NOTINITED;
    if(!sesptr->fi_buf.alloc || !sesptr->fi_buf.data) return KASerrPARAMS_NOTINITED;
    return KASerrOK;
}


static void KasSCleanup(kas_session_t *sesptr)
{
    clear_metadata(sesptr->metadata);
    sesptr->metadata = NULL;
    if(sesptr->datafile_fd>=0)
	{
	    if(sesptr->file_pos != (off_t)-1)
		{
		    lseek(sesptr->datafile_fd,SEEK_SET,sesptr->file_pos);
		}
	    close(sesptr->datafile_fd);
	    sesptr->datafile_fd = -1;
	}
    sesptr->file_pos = (off_t)-1;
    sesptr->data_size=0;
    sesptr->metadata_start = 0;
    sesptr->progress_cb = NULL;
    sesptr->progress_cb_param = NULL;
    sesptr->wr_buf.used = sesptr->rd_buf.used = sesptr->fi_buf.used = 0;
    sesptr->header_sent = sesptr->line_complete = sesptr->end_sent = 0;
    sesptr->xstatus = sesptr->resp_count = sesptr-> final_status = 0;
#define CLEAR_STRING(str) do{ if(sesptr->str){free(sesptr->str); sesptr->str=NULL;}}while(0)
    CLEAR_STRING(reason);
    CLEAR_STRING(content_categories);
    CLEAR_STRING(detection_methods);
    CLEAR_STRING(extended_info);
#undef CLEAR_STRING
}

int KasSDisconnect(kas_session_t *sesptr)
{
    int rc;
    if((rc = KasCheckSession(sesptr))!=KASerrOK)
	return rc;
    if(sesptr->socket<0 && sesptr->current_status.status == KAS_STATUS_DISCONNECTED)
	return KASerrOK; // already disconnected
    // close socket
    if(sesptr->socket>=0)
	close(sesptr->socket);
    sesptr->socket=-1;
    // clear buffers
    // close datafile
    KasSCleanup(sesptr);
    // report closing
    if(sesptr->current_status.status>KAS_STATUS_CONNECTING)
	{
	    STATUS_CALLBACK_EVENT(sesptr,KAS_STATUS_CLOSING,0);
	}
    sesptr->read_func = sesptr->write_func = NULL;
    sesptr->pid = sesptr->filter_rss = 0;
    bzero(&sesptr->last_io,sizeof(sesptr->last_io));
    // report new state
    if(sesptr->current_status.status>=KAS_STATUS_CONNECTING)
	{
	    sesptr->current_status.status = KAS_STATUS_DISCONNECTED;
	    STATUS_CALLBACK(sesptr);
	}
    return KASerrOK;
}

void KasSSessionDestroy(kas_session_t *sesptr)
{
    int rc;
    rc = KasCheckSession(sesptr);
    if(rc==KASerrPARAMS_NOTINITED && sesptr->wasmalloc)
	{
	    free(sesptr);
	    return;
	}
    else if (rc!=KASerrOK)
	return;
    if(sesptr->current_status.status>KAS_STATUS_CONNECTING && sesptr->progress_cb)
	{
	    PROGRESS_CALLBACK_SIMPLE(sesptr,KAS_INFO_ERROR);
	    KasSDisconnect(sesptr);
	}
    free(sesptr->wr_buf.data);
    free(sesptr->rd_buf.data);
    free(sesptr->fi_buf.data);
    if(sesptr->wasmalloc)
	free(sesptr);
    return;
}

/* connecting to TCP socket */
static int inet_connect(kas_session_t *s, const struct sockaddr_in *addr) {
    int flags;
    int state;
    int t=1;
    struct linger l;
    int rc;
    if((rc = KasCheckSession(s))!= KASerrOK)
	return rc;
    if((s->socket = socket (AF_INET,  SOCK_STREAM,  0))<0)
	return KASerrUNABLE_TO_CONNECT;

    if(-1==(flags = fcntl(s->socket,  F_GETFL,  0)))
	{
	    close(s->socket);
	    s->socket = -1;
	    return KASerrUNABLE_TO_CONNECT;
	}
    if(-1== fcntl(s->socket,  F_SETFL,  flags | O_NONBLOCK))
	{
	    close(s->socket);
	    s->socket = -1;
	    return KASerrUNABLE_TO_CONNECT;
	}
    
    l.l_onoff=1;
    l.l_linger = 0;
    if(-1==setsockopt(s->socket,SOL_SOCKET,SO_LINGER,&l,sizeof(l)))
	{
	    close(s->socket);
	    s->socket = -1;
	    return KASerrUNABLE_TO_CONNECT;
	}
    if(-1==setsockopt(s->socket,IPPROTO_TCP,TCP_NODELAY,&t,sizeof(t))) 
	{
	    close(s->socket);
	    s->socket = -1;
	    return KASerrUNABLE_TO_CONNECT;
	}

    if ((state = connect (s->socket,(struct sockaddr *) addr, sizeof(*addr))) < 0) 
	{ 
	    if (errno != EINPROGRESS) 
		{
		    close(s->socket);
		    s->socket = -1;
		    return KASerrUNABLE_TO_CONNECT;
		}
	    else 
		return KASerrIN_PROGRESS;
	}
    else 
	    return KASerrOK;
}

	
static int unix_connect(kas_session_t *s, const char *sockname) 
{
    int flags;
    int state;
    struct sockaddr_un port;
    int rc;
    if((rc = KasCheckSession(s))!= KASerrOK)
	return rc;

    port.sun_family = AF_LOCAL;
    strncpy(port.sun_path, sockname, sizeof(port.sun_path)-1);
    if((s->socket = socket (AF_LOCAL,  SOCK_STREAM,  0))<0) 
	    return KASerrUNABLE_TO_CONNECT;

    if(-1==(flags = fcntl(s->socket,  F_GETFL,  0)))
	{
	    close(s->socket);
	    s->socket = -1;
	    return KASerrUNABLE_TO_CONNECT;
	}
    if(-1== fcntl(s->socket,  F_SETFL,  flags | O_NONBLOCK))
	{
	    close(s->socket);
	    s->socket = -1;
	    return KASerrUNABLE_TO_CONNECT;
	}
    if ((state = connect (s->socket, (struct sockaddr *) &port,  sizeof port)) < 0) 
	{
	    if (errno != EINPROGRESS) 
		{
		    close(s->socket);
		    s->socket = -1;
		    return KASerrUNABLE_TO_CONNECT;
		}
	    else 
		return KASerrIN_PROGRESS;
	}
    else
	return KASerrOK;
}


static int normalize_inet_name(kas_session_t *s, const char *xname, struct sockaddr_in *addr) {

    struct hostent *h;
    struct servent *serv;
    char *host, *port, *first, *second,*name;
    
    if(!(name = alloca(strlen(xname)+1)))
	return KASerrUNABLE_TO_CONNECT;
    strcpy(name,xname);

    if(!(first=strchr(name, ':')) || !(second=strrchr(name, ':')) || !(first == second))
	    return KASerrUNABLE_TO_CONNECT;

    *first='\0';
    host = (char *)name;
    port = first+1;
    addr->sin_family = AF_INET;

    /* check for numeric addresses by normal way. now possible bugs */
    if (*host>='0' && *host<='9') 
	addr->sin_addr.s_addr = inet_addr(host);
    else {
	  h = gethostbyname(host);
	  if (!h)
	      return KASerrUNABLE_TO_CONNECT;
	  addr->sin_addr = *((struct in_addr *) h->h_addr);
    }
    if (*port>='0' && *port<='9')
	addr->sin_port=htons (atoi(port));
    else {	
	  serv=getservbyname(port, (char*)0);
	  if (serv)
	      addr->sin_port=serv->s_port;
	  else
	      return KASerrUNABLE_TO_CONNECT;
    }
    return KASerrOK;
}


			
int KasSConnect(kas_session_t *sesptr,
		const char *daemon_addr,
		int timeout)
{
    int rc;
    if((rc = KasCheckSession(sesptr))!= KASerrOK)
	return rc;
    if(sesptr->current_status.status > KAS_STATUS_DISCONNECTED)
	return KASerrALREADY_CONNECTED;

    sesptr->connect_timeout = timeout;
    sesptr->standard_proto = sesptr->chunked_write = 0;
    gettimeofday(&sesptr->last_io,NULL);
    if(!strncasecmp(daemon_addr,"unix:",5))
	rc = unix_connect(sesptr, daemon_addr+5);
    else if(!strncasecmp(daemon_addr,"tcp:",4))
	{
	    struct sockaddr_in addr;
	    rc=normalize_inet_name(sesptr,daemon_addr+4, &addr);
	    if(rc ==KASerrOK)
		rc = inet_connect(sesptr,&addr);
	}
    else
	return KASerrUNABLE_TO_CONNECT;

    sesptr->wr_buf.used = sesptr->rd_buf.used = sesptr->fi_buf.used = 0;
    
    switch (rc)
	{
	default:
	    STATUS_CALLBACK_EVENT(sesptr,KAS_STATUS_CONNECTERROR,KAS_ES_CONNERROR);
	    return KASerrUNABLE_TO_CONNECT;
	case KASerrOK:
	    sesptr->current_status.status = KAS_STATUS_CONNECTED;
	    STATUS_CALLBACK(sesptr);
	    sesptr->read_func = read_greeting;
	    return rc;
	case KASerrIN_PROGRESS:
	    sesptr->current_status.status = KAS_STATUS_CONNECTING;
	    STATUS_CALLBACK(sesptr);
	    sesptr->read_func = read_connected;
	    return rc;
	}
}

static int KasSReadData(kas_session_t *sesptr)
{
    int rc;
    if((rc = KasCheckSession(sesptr))!= KASerrOK)
	return rc;
    if(sesptr->socket<0 || sesptr->current_status.status<KAS_STATUS_CONNECTING)
	// nothing to read
	return KASerrDISCONNECTED;
    if(sesptr->rd_buf.used>= sesptr->rd_buf.alloc)
	return KASerrREAD_ERROR;
    rc = read(sesptr->socket,
	      sesptr->rd_buf.data+sesptr->rd_buf.used,
	      sesptr->rd_buf.alloc-sesptr->rd_buf.used);
    if(rc<=0)
	return KASerrREAD_ERROR; /* 0 bytes read after non-zero poll - error! */
    else
	{
	    sesptr->rd_buf.used+=rc;
	    sesptr->rd_buf.data[sesptr->rd_buf.used]=0;
	    gettimeofday(&sesptr->last_io,NULL);
	    return KASerrOK;
	}
}

static int KasSWriteData(kas_session_t *sesptr)
{
    int rc;
    if((rc = KasCheckSession(sesptr))!= KASerrOK)
	return rc;
    if(sesptr->socket<0 || sesptr->current_status.status<KAS_STATUS_CONNECTED)
	// nothing to read
	return KASerrDISCONNECTED;
    if(sesptr->wr_buf.used < 1)
	return KASerrWRITE_ERROR;
    rc = write(sesptr->socket,
	       sesptr->wr_buf.data,
	       sesptr->wr_buf.used);
    if(rc<0)
	return KASerrWRITE_ERROR;
    else if(rc==0)
	{
	    return __KASerrNODATA;
	}
    else
	{
	    memmove(sesptr->wr_buf.data,sesptr->wr_buf.data+rc,sesptr->wr_buf.used-rc);
	    sesptr->wr_buf.used-=rc;
	    gettimeofday(&sesptr->last_io,NULL);
	    return KASerrOK;
	}
}


int KasSPreparePoll(kas_session_t *sesptr,struct pollfd *pollfds,unsigned int index_to_use)
{
    int rc;
    if(!pollfds||!sesptr)
	return KASerrPARAMS_NULLPTR;
    if((rc = KasCheckSession(sesptr))!= KASerrOK)
	return rc;
    if(sesptr->socket<0 || sesptr->current_status.status<KAS_STATUS_CONNECTING)
	// nothing to read
	return KASerrDISCONNECTED;
    pollfds[index_to_use].fd = sesptr->socket;
    if(sesptr->read_func)
#ifdef POLLDRNORM
	pollfds[index_to_use].events |=(POLLIN|POLLRDNORM|POLLRDBAND);
#else
	pollfds[index_to_use].events |=(POLLIN);
#endif
    if(sesptr->write_func)
#ifdef POLLWRNORRM
	pollfds[index_to_use].events |=(POLLOUT|POLLWRNORM|POLLWRBAND);
#else
	pollfds[index_to_use].events |=(POLLOUT);
#endif
    return KASerrOK;
}

int KasSPrepareSelect(kas_session_t *sesptr,fd_set *readfd, fd_set *writefd)
{
    int rc;
    if(!readfd || !writefd || !sesptr)
	return KASerrPARAMS_NULLPTR;
    if((rc = KasCheckSession(sesptr))!= KASerrOK)
	return rc;
    if(sesptr->socket<0 || sesptr->current_status.status<KAS_STATUS_CONNECTING)
	// nothing to read
	return KASerrDISCONNECTED;
    if(sesptr->read_func)
	FD_SET(sesptr->socket,readfd);
    if(sesptr->write_func)
	FD_SET(sesptr->socket,writefd);
    return KASerrOK;
}



char *check_string(char *ptr,int len)
{
    int i;
    for(i=0;i<len;i++)
	if(!ptr[i]) return ptr+i;
    return NULL;
}


#define ATTRIBUTE_WAIT		0
#define ATTRIBUTE_OK		1
#define ATTRIBUTE_ZERO		2

static int read_one_attribute(kas_session_t *st, unsigned char **aname, unsigned char **aval)
{
    unsigned char *p;
    
    *aname = *aval = NULL;

    if(st->rd_buf.used <1)
	return ATTRIBUTE_WAIT;

    if (st->rd_buf.data[0] == '\0')
	  {
	    *aname = "";
	    return ATTRIBUTE_ZERO;
	  }

    if (st->rd_buf.used < 4)
        return ATTRIBUTE_WAIT; /* at least one-char aname and one-char aval + two \0. */

    p = check_string(st->rd_buf.data, st->rd_buf.used);
    if (!p)
        return ATTRIBUTE_WAIT; /* incomplete aname */
    *aval = p + 1;
    p = check_string(p + 1, st->rd_buf.used - (p + 1 - st->rd_buf.data));
    if (!p)
      {
        *aval = NULL;
        return ATTRIBUTE_WAIT; /* incomplete aval */
      }
    *aname = st->rd_buf.data;
    return ATTRIBUTE_OK; /* one attribute read */
}

static void move_buffer(kas_buffer_t *buf, unsigned int to_move)
{
    if(!buf || !buf->data || !buf->used) return;
    if(to_move>buf->used) to_move = buf->used;
    memmove(buf->data,buf->data+to_move,buf->used-to_move);
    buf->used-=to_move;
}

static int parse_xq_id(kas_session_t *sesptr,char *aval)
{
    char *p,*pp;
    if(strncmp(aval,"AAPx",4))
	{
	    if(!strncmp(aval,"AAP",3))
		{
		    char pidbuf[8];
		    strncpy(pidbuf,aval+3,7);
		    pidbuf[7]=0;
		    sesptr->pid=strtol(pidbuf,&pp,16);
		    if(sesptr->pid<0) sesptr->pid=0;
		    sesptr->filter_rss=0;
		    sesptr->standard_proto=1;
		    return KASerrOK;
		}
	    else
		return __KASerrINCORRECT_PROTOCOL;		
	}
    p = aval+4;
    p = strchr(p,'x');
    if(!p)
	return __KASerrINCORRECT_PROTOCOL;		
    *p=0;
    if((sesptr->pid = strtol(aval+4,&pp,16))<1)
	return __KASerrINCORRECT_PROTOCOL;		
    p++;
    pp = strchr(p,'x');
    if(!pp)
	return __KASerrINCORRECT_PROTOCOL;		
    *pp=0;
    pp=p;
    if((sesptr->filter_rss = strtol(p,&pp,16))<1)
	{
	    return __KASerrINCORRECT_PROTOCOL;		
	}
    
    return KASerrOK;
}


static int read_greeting(kas_session_t *sesptr)
{
    int rc;
    unsigned char *aname,*aval;
    if((rc = KasCheckSession(sesptr))!= KASerrOK)
	return rc;
    if(sesptr->current_status.status != KAS_STATUS_CONNECTED && sesptr->current_status.status != KAS_STATUS_RECEIVING)
	return __KASerrINCORRECT_PROTOCOL;
    while(1)
	{
	    rc = read_one_attribute(sesptr,&aname,&aval);
	    if(rc==ATTRIBUTE_ZERO)
		{
		    // change to IDLE mode
		    move_buffer(&sesptr->rd_buf,1);
		    KasSCleanup(sesptr);
		    sesptr->read_func = NULL;
		    sesptr->current_status.status=KAS_STATUS_IDLE;
		    STATUS_CALLBACK(sesptr);
		    return KASerrOK;
		}
	    else if (rc == ATTRIBUTE_OK)
		{
		    int movelen;
		    movelen=aval - sesptr->rd_buf.data;
		    movelen += strlen(aval)+1;
		    if(!strcasecmp(aname,"queue_id"))
			{
			    STATUS_CALLBACK_EVENT(sesptr,KAS_STATUS_INCORRECT_PROTO,0);
			    KasSDisconnect(sesptr);
			    return __KASerrINCORRECT_PROTOCOL;		
			}
		    else if(!strcasecmp(aname,"xqueue_id"))
			{
			    if(parse_xq_id(sesptr,aval)!=KASerrOK)
				{
				    STATUS_CALLBACK_EVENT(sesptr,KAS_STATUS_INCORRECT_PROTO,0);
				    KasSDisconnect(sesptr);
				    return __KASerrINCORRECT_PROTOCOL;		
				}
			    // callback if too large
			    if(sesptr->scanner_rss_limit > 0 && 
			       sesptr->scanner_rss_limit < sesptr->filter_rss)
				STATUS_CALLBACK_EVENT(sesptr,KAS_STATUS_WANTRESTART,0);

			}
		    else if (!strcasecmp(aname,"continue"))
			{
			    // continue attr not used
			}
		    else
			{
			    // unknown attribute
			    STATUS_CALLBACK_EVENT(sesptr,KAS_STATUS_INCORRECT_PROTO,0);
			    KasSDisconnect(sesptr);
			    return __KASerrINCORRECT_PROTOCOL;		
			}
		    // move buffer
		    move_buffer(&sesptr->rd_buf,movelen);
		}
	    else 
		break;
	}
    return KASerrIN_PROGRESS; // not all attributes read
}

static int read_pingresponse(kas_session_t *sesptr)
{
    unsigned char *aname,*aval;
    int rc;
    if((rc = KasCheckSession(sesptr))!= KASerrOK)
	return rc;
    while(1)
	{
	    rc = read_one_attribute(sesptr,&aname,&aval);
	    if(rc==ATTRIBUTE_ZERO)
		{
		    // change to IDLE mode
		    move_buffer(&sesptr->rd_buf,1);
		    sesptr->read_func = NULL;
		    sesptr->current_status.status=KAS_STATUS_IDLE;
		    STATUS_CALLBACK(sesptr);
		    return KASerrOK;
		}
	    else if (rc == ATTRIBUTE_OK)
		{
		    int movelen;
		    if (!strcasecmp(aname,"continue"))
			{
			    // continue attr not used
			}
		    else
			{
			    // unknown attribute
			    STATUS_CALLBACK_EVENT(sesptr,KAS_STATUS_INCORRECT_PROTO,0);
			    KasSDisconnect(sesptr);
			    return __KASerrINCORRECT_PROTOCOL;		
			}
		    // move buffer
		    movelen=aval - sesptr->rd_buf.data;
		    movelen += strlen(aval)+1;
		    move_buffer(&sesptr->rd_buf,movelen);
		}
	    else
		break;
	}
    return KASerrIN_PROGRESS; // not all attributes read
}

static int read_skip_zero_attr(kas_session_t *sesptr)
{
    unsigned char *aname,*aval;
    int rc;
    if((rc = KasCheckSession(sesptr))!= KASerrOK)
	return rc;
    while(1)
	{
	    rc = read_one_attribute(sesptr,&aname,&aval);
	    if(rc==ATTRIBUTE_ZERO)
		{
		    move_buffer(&sesptr->rd_buf,1);
		    sesptr->read_func = read_response_attrs;
		    return read_response_attrs(sesptr);
		}
	    else if (rc == ATTRIBUTE_OK)
		{
		    // unknown/unwanted attribute, 
		    PROGRESS_CALLBACK_SIMPLE(sesptr,KAS_INFO_ERROR);
		    STATUS_CALLBACK_EVENT(sesptr,KAS_STATUS_INCORRECT_PROTO,0);
		    KasSDisconnect(sesptr);
		    return __KASerrINCORRECT_PROTOCOL;		
		}
	    else
		break;
	}
    return KASerrIN_PROGRESS; // not all attributes read
}
static int read_final_attr(kas_session_t *sesptr)
{
    unsigned char *aname,*aval;
    int rc;
    if((rc = KasCheckSession(sesptr))!= KASerrOK)
	return rc;
    while(1)
	{
	    rc = read_one_attribute(sesptr,&aname,&aval);
	    if(rc==ATTRIBUTE_ZERO)
		{
		    // change to IDLE mode
		    move_buffer(&sesptr->rd_buf,1);
		    STATUS_CALLBACK_EVENT(sesptr,KAS_STATUS_SCAN_END,0);
		    if(!sesptr->standard_proto)
			{
			    sesptr->read_func = read_greeting;
			    return read_greeting(sesptr);
			}
		    else
			return KasSDisconnect(sesptr);
		}
	    else if (rc == ATTRIBUTE_OK)
		{
		    PROGRESS_CALLBACK_SIMPLE(sesptr,KAS_INFO_ERROR);
		    STATUS_CALLBACK_EVENT(sesptr,KAS_STATUS_INCORRECT_PROTO,0);
		    KasSDisconnect(sesptr);
		    return __KASerrINCORRECT_PROTOCOL;		
		}
	    else
		break;
	}
    return KASerrOK; // not all attributes read
}

static int read_response_records(kas_session_t *sesptr)
{
    char rec_buf[MAX_RECORD_LEN+4];
    int type,gotlen;
    char *action;
    // read was called before

  RETRY:
    if(sesptr->rd_buf.used<1)
	return KASerrIN_PROGRESS;
    // try to read one record
    if(sesptr->resp_count<1)
	{
	    // all read
	    sesptr->read_func = read_final_attr;
	    return read_final_attr(sesptr);
	}
    // read and parse records
    if((type=rec_get(sesptr,rec_buf,sizeof(rec_buf),&gotlen))>0)
	{
	    switch(type)
		{
		case REC_TYPE_ACTION:
		    action=rec_buf;
		    rec_buf[gotlen]=0;
		    if(!strncasecmp(action,"Bounce",6))
			{
			    sesptr->final_status = KAS_EINFO_BOUNCE;
			    if(sesptr->progress_cb)
				{
				    char *rcptlist,*end;
				    kas_progress_info_t cpi;
				    bzero(&cpi,sizeof(cpi));
				    cpi.pid = sesptr->pid;
				    cpi.state = KAS_INFO_DONE;
				    cpi.message_status = KAS_EINFO_BOUNCE;
				    rcptlist=strchr(action,':');
				    if(!rcptlist) goto RETRY; // skip
				    *rcptlist=0; rcptlist++;
				    while(rcptlist)
					{
					    end = strchr(rcptlist,';');
					    if(end)
						{
						    *end=0;
						    end++;
						}
					    if(strlen(rcptlist)>0)
						{
						    cpi.rcpt = rcptlist;
						    (*sesptr->progress_cb)(sesptr,&cpi,
									  sesptr->progress_cb_param);
						}
					    rcptlist = end;
					}
				}
			    
			}
		    else if(sesptr->final_status != KAS_EINFO_BOUNCE
			    && (!strncasecmp(action,"AddHeader",9)
				|| !strncasecmp(action,"NewHeader",9))
			)
			{
			    char *header;
			    char *value;
			    header=strchr(action,':');
			    if(!header) goto RETRY; // skip
			    *header=0; header++;
			    value=strchr(header,':');
			    if(!value) goto RETRY;
			    *value=0; value++;
#define APPEND_STRING(field,val) do{					\
				if(!sesptr->field){sesptr->field=strdup(val);} \
				else { int l; sesptr->field=realloc(sesptr->field,l=strlen(sesptr->field)+strlen(val)+3); strncat(sesptr->field,"|",l-1);strncat(sesptr->field,val,l-1);}}while(0)
			    if(!strcasecmp(header,"X-SpamTest-Status"))
				{
				    if(!strcasecmp(value,"SPAM"))
				       sesptr->final_status = KAS_EINFO_ACCEPT_SPAM;
				    else if(!strcasecmp(value,"Probable Spam"))
				       sesptr->final_status = KAS_EINFO_ACCEPT_PROBABLESPAM;
				    else if(!strcasecmp(value,"Not detected"))
				       sesptr->final_status = KAS_EINFO_ACCEPT_CLEAN;
				    else if(!strcasecmp(value,"Trusted"))
				       sesptr->final_status = KAS_EINFO_ACCEPT_TRUSTED;
				    else if(!strncasecmp(value,"Formal",6))
				       sesptr->final_status = KAS_EINFO_ACCEPT_FORMAL;
				    else if(!strncasecmp(value,"Blacklisted",11))
				       sesptr->final_status = KAS_EINFO_ACCEPT_BLACKLISTED;
				    else
				       sesptr->final_status = KAS_EINFO_ACCEPT_USERDEFINED;
				}
			    else if(!strcasecmp(header,"X-SpamTest-Method"))
				APPEND_STRING(detection_methods,value);
			    else if(!strcasecmp(header,"X-SpamTest-Categories"))
				APPEND_STRING(content_categories,value);
			    else if(!strcasecmp(header,"X-SpamTest-Info"))
				APPEND_STRING(extended_info,value);
#undef APPEND_STRING
			}
		    else
			{
			    // unknown action, just skip
			}
		    goto RETRY;
		    break;
		case REC_TYPE_RCPT:
		    if(sesptr->final_status != KAS_EINFO_BOUNCE)
			{
			    if(sesptr->progress_cb)
				{
				    kas_progress_info_t cpi;
				    bzero(&cpi,sizeof(cpi));
				    cpi.pid = sesptr->pid;
				    cpi.state = KAS_INFO_DONE;
				    cpi.message_status = sesptr->final_status;
				    rec_buf[gotlen]=0;
				    cpi.rcpt=rec_buf;
				    cpi.content_categories = sesptr->content_categories;
				    cpi.detection_methods = sesptr->detection_methods;
				    cpi.extended_info = sesptr->extended_info;
				    (*sesptr->progress_cb)(sesptr,&cpi, sesptr->progress_cb_param);
				}
			}
		    goto RETRY;
		    break;
		case REC_TYPE_END:
		    // end of record cycle
		    sesptr->resp_count--;
		    // clean message status
#define CLEAN_STRING(ss) do{if(sesptr->ss){free(sesptr->ss);sesptr->ss=NULL;}}while(0)
		    CLEAN_STRING(detection_methods);
		    CLEAN_STRING(extended_info);
		    CLEAN_STRING(content_categories);
		    sesptr->final_status=0;
		    goto RETRY;
		default:
		    goto RETRY;
		}
	    goto RETRY;
	}
    return KASerrIN_PROGRESS; // not ready!
    
}

static int read_response_attrs(kas_session_t *sesptr)
{
    unsigned char *aname,*aval;
    int rc;
    if((rc = KasCheckSession(sesptr))!= KASerrOK)
	return rc;
    while(1)
	{
	    rc = read_one_attribute(sesptr,&aname,&aval);
	    if(rc==ATTRIBUTE_ZERO)
		{
		    move_buffer(&sesptr->rd_buf,1);
		    // distinguish between count>0 and count=0
		    if(sesptr->xstatus!=0 || sesptr->resp_count==0)
			{
			    kas_progress_info_t cpi;
			    bzero(&cpi,sizeof(cpi));
			    cpi.state = KAS_INFO_DONE;
			    cpi.pid = sesptr->pid;
			    if(sesptr->xstatus)
				switch(sesptr->xstatus)
				    {
				    case 1: cpi.message_status = KAS_EINFO_ERROR; break;
				    case 8: cpi.message_status = KAS_EINFO_REJECT;  break;
				    case 16: cpi.message_status = KAS_EINFO_BLACKHOLE; break;
				    default: cpi.message_status = KAS_EINFO_UNKNOWN; break;
				    }
			    else
				cpi.message_status = KAS_EINFO_BLACKHOLE;
			    if(sesptr->progress_cb)
				(*sesptr->progress_cb)(sesptr,&cpi,sesptr->progress_cb_param);
			    sesptr->read_func = read_final_attr;
			    return read_final_attr(sesptr);
			}
		    else
			{
			    sesptr->read_func = read_response_records;
			    return read_response_records(sesptr);
			}
		}
	    else if (rc == ATTRIBUTE_OK)
		{
		    int movelen;
		    if (!strcasecmp(aname,"continue"))
			{
			    // got continue, should read zero attr
			    movelen=aval - sesptr->rd_buf.data;
			    movelen += strlen(aval)+1;
			    move_buffer(&sesptr->rd_buf,movelen);
			    sesptr->read_func = read_skip_zero_attr;
			    return read_skip_zero_attr(sesptr);
			}
		    else if(!strcasecmp(aname,"xstatus"))
			{
			    sesptr->xstatus=atoi(aval);
			}
		    else if(!strcasecmp(aname,"count"))
			{
			    sesptr->resp_count=atoi(aval);
			}
		    else if(!strcasecmp(aname,"reason"))
			{
			    sesptr->reason = strdup(aval);
			}
		    else
			{
			    // unknown attribute
			    PROGRESS_CALLBACK_SIMPLE(sesptr,KAS_INFO_ERROR);
			    STATUS_CALLBACK_EVENT(sesptr,KAS_STATUS_INCORRECT_PROTO,0);
			    KasSDisconnect(sesptr);
			    return __KASerrINCORRECT_PROTOCOL;		
			}
		    // move buffer
		    movelen=aval - sesptr->rd_buf.data;
		    movelen += strlen(aval)+1;
		    move_buffer(&sesptr->rd_buf,movelen);
		}
	    else
		break;
	}
    return KASerrOK; // not all attributes read
}





static int  check_timeout2(kas_session_t *sesptr,struct timeval *now)
{
    int tdiff;
    kas_status_t kstatus;

    bzero(&kstatus,sizeof(kstatus));
    tdiff=diff_msec(&sesptr->last_io,now);
    if(sesptr->read_func)
	{
	    if(sesptr->current_status.status == KAS_STATUS_CONNECTING)
		{
		    if(sesptr->connect_timeout > 0 && sesptr->connect_timeout<tdiff)
			{
			    if(sesptr->progress_cb)
				PROGRESS_CALLBACK_SIMPLE(sesptr,KAS_INFO_ERROR);
			    STATUS_CALLBACK_EVENT(sesptr,KAS_STATUS_CONNECTERROR,KAS_ES_CONNTIMEOUT);
			    KasSDisconnect(sesptr);
			    return KASerrUNABLE_TO_CONNECT;
			}
		}
	    else
		{
		    if(sesptr->read_timeout > 0 && sesptr->read_timeout<tdiff)
			{
			    if(sesptr->progress_cb)
				PROGRESS_CALLBACK_SIMPLE(sesptr,KAS_INFO_ERROR);
			    STATUS_CALLBACK_EVENT(sesptr,KAS_STATUS_IOERROR,KAS_ES_IOTIMEOUT);
			    KasSDisconnect(sesptr);
			    return __KASerrTIMEOUT;
			}
		}
	}
    if(sesptr->write_func)
	{
	    if(sesptr->write_timeout > 0 && sesptr->write_timeout<tdiff)
		{
		    if(sesptr->progress_cb)
			PROGRESS_CALLBACK_SIMPLE(sesptr,KAS_INFO_ERROR);
		    STATUS_CALLBACK_EVENT(sesptr,KAS_STATUS_IOERROR,KAS_ES_IOTIMEOUT);
		    KasSDisconnect(sesptr);
		    return __KASerrTIMEOUT;
		}
	}
    return KASerrOK;
}

static int  check_timeout(kas_session_t *sesptr)
{
    struct timeval tv;
    gettimeofday(&tv,NULL);
    return check_timeout2(sesptr,&tv);
}

static int read_connected(kas_session_t *sesptr)
{
    int rc;
    if((rc = KasCheckSession(sesptr))!= KASerrOK)
	return rc;
    if(sesptr->current_status.status != KAS_STATUS_CONNECTING)
	return __KASerrINCORRECT_PROTOCOL;
    // we're called because some data arrived, so not check timeouts
    sesptr->read_func = read_greeting;
    sesptr->current_status.status = KAS_STATUS_CONNECTED;
    STATUS_CALLBACK(sesptr);
    return (sesptr->read_func)(sesptr);
}



#define WRITE_OK	0
#define WRITE_NOSPACE  -2
static int     rec_put(kas_session_t *s, int type, const char *data, int len)
{
    int     len_rest;
    int     len_byte;

    if(s->wr_buf.alloc-s->wr_buf.used < len+6)
	return WRITE_NOSPACE;

    s->wr_buf.data[s->wr_buf.used++]=type;

    /*
     * Write the record data length in 7-bit portions, using the 8th bit to
     * indicate that there is more. Use as many length bytes as needed.
     */
    len_rest = len;
    do {
	len_byte = len_rest & 0177;
	if (len_rest >>= 7)
	    len_byte |= 0200;
	s->wr_buf.data[s->wr_buf.used++]=len_byte;
    } while (len_rest != 0);

    memmove(s->wr_buf.data+s->wr_buf.used,data,len);
    s->wr_buf.used+=len;
    return WRITE_OK;
}

#define READ_INCOMPLETE -1

// just read a byte, keeping state
int read_byte(kas_session_t *s, int *pos)
{
    if(*pos < s->rd_buf.used)
	    return s->rd_buf.data[(*pos)++];
    else
	return -1;
}

#ifndef NBBY
#define NBBY 8
#endif

static int     rec_get(kas_session_t *s,unsigned char *buf, int maxsize, int *gotlen)
{
    int     len,len_byte, shift, to_read, type;
    char    *bufp;
    int		pos = 0,byte;

    *gotlen = 0;
    if((type = read_byte(s,&pos))<0)
	    return READ_INCOMPLETE;

    for (len = 0, shift = 0; /* void */ ; shift += 7) 
	{
	    if (shift >= (int) (NBBY * sizeof(int)))
		return READ_INCOMPLETE;
	    if ((len_byte = read_byte(s,&pos)) == -1)
		return READ_INCOMPLETE;
	    len |= (len_byte & 0177) << shift;
	    if ((len_byte & 0200) == 0)
		break;
	}
    // too small buffer, skip data
    if (len < 0 || (maxsize > 0 && len > maxsize)) 
	{
	    while (len-- > 0 && read_byte(s,&pos) != -1)
		/* void */ ;
	    // now we've total number of skipped bytes in pos; if all
	    // skipped, than move buffer one step
	    if(!len)
		{
		    memmove(s->rd_buf.data,s->rd_buf.data+pos, s->rd_buf.used-pos);
		    s->rd_buf.used-=pos;
		}
	    return READ_INCOMPLETE;
	}
    bufp = buf;
    to_read = len;
    while(to_read>0) 
	{
	    byte = read_byte(s,&pos);
	    if(byte<0)
		return READ_INCOMPLETE;
	    *(bufp++)=byte;
	    to_read--;
	}
    // now we've full data in buf and readed data in pos
    memmove(s->rd_buf.data,s->rd_buf.data+pos, s->rd_buf.used-pos);
    s->rd_buf.used-=pos;
    *gotlen = len;
    return type;
}


int KasSPing(kas_session_t *s)
{
    int rc;
    if((rc = KasCheckSession(s))!= KASerrOK)
	return rc;
    if(s->current_status.status != KAS_STATUS_IDLE)
	return KASerrBUSY;
    if(s->wr_buf.used>0)
	return KASerrBUSY;
    if(WRITE_OK!=rec_put(s,REC_TYPE_BLOCKFIN,"",0))
	return KASerrBUSY;
    s->current_status.status = KAS_STATUS_PINGWRITE;
    s->write_func = write_ping;
    STATUS_CALLBACK(s);
    return KASerrIN_PROGRESS;
}

static int write_ping(kas_session_t *sesptr)
{
    int rc;
    if((rc = KasCheckSession(sesptr))!= KASerrOK)
	return rc;
    // buffer filled before!
    rc = KasSWriteData(sesptr);
    switch(rc)
	{
	    // no data sent
	case __KASerrNODATA:
	    return KASerrIN_PROGRESS;
	    // incorrect call to write
	case KASerrDISCONNECTED:
	    return KASerrWRITE_ERROR;
	case KASerrWRITE_ERROR:
	    // socket error, send event and disconnect
	    STATUS_CALLBACK_EVENT(sesptr,KAS_STATUS_IOERROR,KAS_ES_IOERROR);
	    KasSDisconnect(sesptr);
	    return KASerrWRITE_ERROR;
	case KASerrOK:
	    // check, if all written
	    if(sesptr->wr_buf.used == 0)
		{
		    // switch to ping response
		    sesptr->current_status.status = KAS_STATUS_PINGRESPONSE;
		    STATUS_CALLBACK(sesptr);
		    sesptr->write_func = NULL;
		    sesptr->read_func = read_pingresponse;
		    return KASerrOK;
		}
	    return KASerrIN_PROGRESS;
	}
    return KASerrOK;
}

int fill_file_data(kas_session_t *s)
{
    int rc;
    unsigned char *start,*end;
    size_t tosend,recsize;
    int retval;
    char *xmx = "X-Mailbox-Line: ";

    if((rc = KasCheckSession(s))!= KASerrOK)
	return rc;
    if(s->current_status.status != KAS_STATUS_SENDING)
	return KASerrBUSY;
    if(s->datafile_fd<0 && s->fi_buf.used<1)
	return KASerrPARAMS_NOTINITED;
    
    if(s->data_size<1 &&s->fi_buf.used<1)
	return KASerrOK; // all read

  REREAD:
    // space in buf ?
    if(s->fi_buf.used<s->fi_buf.alloc && s->datafile_fd >=0 && s->data_size>0)
	{
	    // read something to buffer
	    if((rc=read(s->datafile_fd,s->fi_buf.data+s->fi_buf.used,s->fi_buf.alloc-s->fi_buf.used))>0)
		{
		    s->fi_buf.used+=rc;
		    s->data_size-=rc;
		}
	    if(rc<1 || s->data_size<1)
		{
		    if(s->file_pos!=(off_t)-1)
			{
			    lseek(s->datafile_fd,SEEK_SET,s->file_pos);
			    s->file_pos=(off_t)-1;
			}
		    close(s->datafile_fd);
		    s->datafile_fd=-1;
		    s->data_size=0;
		}
	}
    // nothing in file buffer
    if(s->fi_buf.used<1)
	{
	    if(!s->end_sent)
		{
		    if (rec_put(s, REC_TYPE_END, "",0)!=WRITE_OK )
			return KASerrIN_PROGRESS;
		    else
			{
			    s->end_sent=1;
			    return KASerrOK;
			}
		}
	}
    // put something to buffer
    retval = 0;
    tosend=s->fi_buf.used;
    start = s->fi_buf.data;
    while(tosend>0)
	{
	    end=memchr(start,'\n',tosend>MAX_RECORD_LEN-1?MAX_RECORD_LEN-1:tosend);
	    if(end)
		{
		    // send NORM rec
		    recsize = end-start;
		    if(!s->header_sent && s->line_complete && !strncmp(start,"From ",5))
			{
			    if(s->wr_buf.alloc-s->wr_buf.used < recsize+strlen(xmx)+10)
				{
				    retval = 1;
				    break;
				}
			    /* hiding From line */
			    if (rec_put(s,REC_TYPE_CONT,xmx,strlen(xmx))!=WRITE_OK)
				{
				    retval=1;
				    break;
				}
			}
		    if (rec_put(s, REC_TYPE_NORM, start, recsize)!=WRITE_OK )
			{
			    retval=1;
			    break;
			}
		    if(recsize == 0 && !s->header_sent)
			s->header_sent = 1;
		    s->line_complete = 1; 
		    // move buffer to recsize+1;
		    recsize++;
		    
		    start+=recsize;
		    tosend-=recsize;
		}
	    else
		{
		    // send CONT rec
		    recsize=tosend>MAX_RECORD_LEN-1?MAX_RECORD_LEN-1:tosend;
		    if (rec_put(s, REC_TYPE_CONT, start, recsize)!=WRITE_OK)
			{
			    retval=1;
			    break;
			}
		    start+=recsize;
		    tosend-=recsize;
		}
	}
    if(!tosend)
	{
	    //all sent
	    s->fi_buf.used = 0;
	    goto REREAD;
	}
	    
    tosend = start-s->fi_buf.data;
    if(tosend>0)
	{
	    memmove(s->fi_buf.data,s->fi_buf.data+tosend,s->fi_buf.used-tosend);
	    s->fi_buf.used-=tosend;
	    if(retval)
		return KASerrIN_PROGRESS;
	    else
		goto REREAD;
	}
    return KASerrIN_PROGRESS;
}

int write_file(kas_session_t *s)
{
    int rc;
    if((rc = KasCheckSession(s))!= KASerrOK)
	return rc;
    if(s->current_status.status != KAS_STATUS_SENDING)
	return KASerrBUSY;
    // buffer filled before, at fill_file_data
    rc = KasSWriteData(s);
    switch(rc)
	{
	    // no data sent
	case __KASerrNODATA:
	    return KASerrIN_PROGRESS;
	    // incorrect call to write
	case KASerrDISCONNECTED:
	    return KASerrWRITE_ERROR;
	case KASerrWRITE_ERROR:
	    // socket error, send event and disconnect
	    STATUS_CALLBACK_EVENT(s,KAS_STATUS_IOERROR,KAS_ES_IOERROR);
	    KasSDisconnect(s);
	    return KASerrWRITE_ERROR;
	case KASerrOK:
	    // check, if all written
	    fill_file_data(s);
	    if(s->wr_buf.used == 0) // nothing to write
		{
		    // switch to file write
		    s->write_func = NULL;
		    s->read_func = read_response_attrs;
		    PROGRESS_CALLBACK_SIMPLE(s,KAS_INFO_WAITSTATUS);
		    s->current_status.status = KAS_STATUS_RECEIVING;
		    STATUS_CALLBACK(s);
		    return KASerrOK;
		}
	    return KASerrIN_PROGRESS;
	}
    return KASerrOK;
}

static int KasSProcessIOResult(kas_session_t *sesptr,int socket_read, int socket_write)
{
    int rc;
    kas_status_t kstatus;

    bzero(&kstatus,sizeof(kstatus));
    if((rc = KasCheckSession(sesptr))!= KASerrOK)
	return rc;
    if(sesptr->socket<0 || sesptr->current_status.status<KAS_STATUS_CONNECTING)
	// nothing to read
	return KASerrDISCONNECTED;
    if(sesptr->read_func && socket_read)
	{
	    rc = KasSReadData(sesptr);
	    switch(rc)
		{
		case KASerrOK:
		    return sesptr->read_func(sesptr);
		case KASerrREAD_ERROR:
		    //   -      
		    // 
		    if(sesptr->progress_cb)
			PROGRESS_CALLBACK_SIMPLE(sesptr,KAS_INFO_ERROR);
		    if(sesptr->current_status.status == KAS_STATUS_CONNECTING)
			    STATUS_CALLBACK_EVENT(sesptr,KAS_STATUS_CONNECTERROR,KAS_ES_CONNERROR);
		    else 
			    STATUS_CALLBACK_EVENT(sesptr,KAS_STATUS_IOERROR,KAS_ES_IOERROR);
		    KasSDisconnect(sesptr);
		    return KASerrREAD_ERROR;
		case __KASerrNODATA:
		    // check timeout
		    return check_timeout(sesptr);
		}
	}
    else if (sesptr->write_func && socket_write)
	{
	    return sesptr->write_func(sesptr); // will fill buffer than write
	}
    else
	return check_timeout(sesptr);
    /* to make stupid compiler happy */
    return KASerrOK;
}

int KasSProcessPollResult(kas_session_t *sesptr, struct pollfd* pollfds, unsigned int index)
{
    int socket_read=0,socket_write=0;

    if(!pollfds ||!sesptr)
	return KASerrPARAMS_NULLPTR;
    if(sesptr->socket<0 || sesptr->current_status.status<KAS_STATUS_CONNECTING)
	// nothing to read
	return KASerrDISCONNECTED;
    if(pollfds[index].fd != sesptr->socket) return KASerrOK; //??
#ifdef POLLRDNORM
    if(pollfds[index].revents & (POLLIN|POLLRDNORM|POLLRDBAND))
#else
    if(pollfds[index].revents & (POLLIN))
#endif
	socket_read = 1;
#ifdef POLLWRNORM
    if(pollfds[index].revents & (POLLOUT|POLLWRNORM|POLLWRBAND))
#else
    if(pollfds[index].revents & (POLLOUT))
#endif
	socket_write = 1;
    if(pollfds[index].revents & (POLLERR|POLLHUP))
	socket_write = 1;
    return KasSProcessIOResult(sesptr,socket_read,socket_write);
}

int KasSProcessSelectResult(kas_session_t *sesptr, fd_set *readfd, fd_set *writefd)
{
    int socket_read=0,socket_write=0;

    if(!readfd||!writefd||!sesptr)
	return KASerrPARAMS_NULLPTR;
    if(sesptr->socket<0 || sesptr->current_status.status<KAS_STATUS_CONNECTING)
	// nothing to read
	return KASerrDISCONNECTED;
    if(FD_ISSET(sesptr->socket,readfd))
	socket_read = 1;
    if(FD_ISSET(sesptr->socket,writefd))
	socket_write = 1;
    return KasSProcessIOResult(sesptr,socket_read,socket_write);
}


static int minimal_timeout(int to1, int to2)
{
    if(to1<0 && to2<0) return -1; // both negative
    if(to1<0) return to2; // to2 not negative
    if(to2<0) return to1; // to1 not negative
    return to1<to2?to1:to2; // minimal value, both not negative
}


// ,   
static int active_timeout(kas_session_t *sessp)
{
   if(KasCheckSession(sessp)!= KASerrOK)
	return 1<<20;
    if(sessp->current_status.status == KAS_STATUS_CONNECTING)
	    return sessp->connect_timeout;
    else
	{
	    if (sessp->read_func)
		return sessp->read_timeout;
	    else if(sessp->write_func)
		return sessp->write_timeout;
	    else
		return 1<<20;
	}
    return 1<<20;
}

static int remaining_time(kas_session_t *sessp, struct timeval *now)
{
    int full_timeout,elapsed;
   if(KasCheckSession(sessp)!= KASerrOK)
	return 1<<20;
    full_timeout = active_timeout(sessp);
    elapsed = diff_msec(&sessp->last_io,now);
    return full_timeout > elapsed? full_timeout-elapsed:0;
}

static int KasSNearestTimeout(kas_session_t *sessp, struct timeval *now, int timeout)
{
   int io_timeout,real_timeout;
   io_timeout = active_timeout(sessp);
   real_timeout = minimal_timeout(io_timeout,timeout);
   if(real_timeout>0)
       real_timeout = minimal_timeout(remaining_time(sessp,now),timeout);
   return real_timeout;
}


static int KasSEventLoop(kas_session_t *sessp, int timeout, int once)
{
    struct pollfd pfd[1];
    int rc;
    struct timeval now,start;
    int real_timeout,timed;
    int try_retry;
    if((rc = KasCheckSession(sessp))!= KASerrOK)
	return rc;
    
    if(sessp->current_status.status == KAS_STATUS_IDLE)
	return KASerrOK;
    if(sessp->current_status.status == KAS_STATUS_WRIDLE)
	return KASerrOK;
    if(sessp->current_status.status == KAS_STATUS_DISCONNECTED)
	return KASerrDISCONNECTED;
    gettimeofday(&start,NULL);
    real_timeout = KasSNearestTimeout(sessp,&start,timeout);
    while(1)
	{
	    try_retry = 0;
	    if(sessp->current_status.status == KAS_STATUS_DISCONNECTED)
		return KASerrDISCONNECTED;
	    bzero(pfd,sizeof(pfd));
	    if(KASerrOK!=(rc=KasSPreparePoll(sessp,pfd,0)))
		return rc;
	    rc = poll(pfd,1,real_timeout);
	    gettimeofday(&now,NULL);
	    if(rc<0)
		{
		    if(errno==EINTR)
			try_retry = 1;
		}
	    else if(rc==0)
		{
		    if(check_timeout2(sessp,&now)!=KASerrOK)
			{
			    return KASerrIN_PROGRESS;
						  // Internal TO elapsed
			}
		}
	    else
		{
		    rc = KasSProcessPollResult(sessp,pfd,0);
		    if(rc != KASerrOK && rc != KASerrIN_PROGRESS)
			return rc;
		    if(rc == KASerrOK && 
		       sessp->standard_proto && sessp->current_status.status == KAS_STATUS_DISCONNECTED)
			return KASerrOK;
		    if(sessp->current_status.status == KAS_STATUS_IDLE)
			return KASerrOK;
		    if(sessp->current_status.status == KAS_STATUS_WRIDLE)
			return KASerrOK;
		    if(check_timeout2(sessp,&now)!=KASerrOK)
			    return KASerrIN_PROGRESS;
						  // Internal TO elapsed
		}
	    if(!once || try_retry)
		{
		    timed=diff_msec(&start,&now);
		    if(timeout>0)
			{
			    if(timed>0)
				{
				    if( timeout<timed)
					return KASerrIN_PROGRESS;
				    else
					timeout-=timed;
				}
			}
		    real_timeout = KasSNearestTimeout(sessp,&now,timeout);
		    if(real_timeout==0)
			return KASerrIN_PROGRESS;
		}
	    else
		return KASerrIN_PROGRESS;
	}
}

int KasSProcessEvents(kas_session_t *sessp, int timeout)
{
    return KasSEventLoop(sessp,timeout,1);
}

int KasSWaitIdle(kas_session_t *sessp, int timeout)
{
    return KasSEventLoop(sessp,timeout,0);
}

void clear_metadata(kas_metadata_t *m)
{
    if(!m) return;
#define CLEAR_STRING(s) do{if(m->s){free(m->s);m->s=NULL;}}while(0)
    CLEAR_STRING(relay_ip);
    CLEAR_STRING(helo_name);
    CLEAR_STRING(mail_from);
    if(m->rcptlist && m->nrcpts)
	{
	    int i;
	    for(i=0;i<m->nrcpts;i++)
		CLEAR_STRING(rcptlist[i]);
	    CLEAR_STRING(rcptlist);
	    m->nrcpts=0;
	}
    if(m->attrlist && m->nattrs)
	{
	    int i;
	    for(i=0;i<m->nattrs;i++)
		{
		    CLEAR_STRING(attrlist[i].aname);
		    CLEAR_STRING(attrlist[i].avalue);
		}
	    CLEAR_STRING(attrlist);
	    m->nattrs=0;
	}
#undef CLEAR_STRING
    free(m);
}

static kas_metadata_t *clone_metadata(kas_metadata_t *m)
{
    kas_metadata_t *ret ;
    
    ret= malloc(sizeof(kas_metadata_t));
    if(!ret) 
	{
	    return NULL;
	}
    bzero(ret,sizeof(kas_metadata_t));
    if(m->relay_ip)
	ret->relay_ip=strdup(m->relay_ip);
    if(m->helo_name)
	ret->helo_name=strdup(m->helo_name);
    if(m->mail_from)
	ret->mail_from=strdup(m->mail_from);
    else
	ret->mail_from=strdup("");
    if(m->rcptlist && m->nrcpts>0)
	{
	    int i;
	    if((ret->rcptlist=malloc(m->nrcpts*sizeof(char*))))
		{
		    for(i=0;i<m->nrcpts;i++)
			{
			    if(m->rcptlist[i])
				ret->rcptlist[i] = strdup(m->rcptlist[i]);
			    else
				ret->rcptlist[i] = strdup("");
			}
		    ret->nrcpts = m->nrcpts;
		}
	}
    if(m->attrlist && m->nattrs > 0)
	{
	    if((ret->attrlist=malloc(m->nattrs*sizeof(kas_attribute_t))))
		{
		    int i;
		    for(i=0;i<m->nattrs;i++)
			{
			    if(m->attrlist[i].aname)
				{
				    ret->attrlist[i].aname = strdup(m->attrlist[i].aname);
				    ret->attrlist[i].avalue = strdup(m->attrlist[i].avalue?m->attrlist[i].avalue:"");
				}
			    else
				{
				    ret->attrlist[i].aname=ret->attrlist[i].avalue=0;
				}
			}
		    ret->nattrs = m->nattrs;
		}

	}
    return ret;
}

int fill_metadata(kas_session_t *s)
{
    char msgbuf[MAX_RECORD_LEN+1];

    int rc,savelen;

    if((rc = KasCheckSession(s))!= KASerrOK)
	return rc;
    if(s->current_status.status != KAS_STATUS_SENDING)
	return KASerrBUSY;
    
    savelen = s->wr_buf.used;
    if(!s->metadata_start)
	{
	    snprintf(msgbuf, MAX_RECORD_LEN, REC_TYPE_SIZE_FORMAT, 0L, 0L, 0L);
	    if(rec_put(s,REC_TYPE_SIZE, msgbuf, strlen(msgbuf))!=WRITE_OK)
		goto RECOVER;
	    snprintf(msgbuf,MAX_RECORD_LEN,"%ld",time(NULL));
	    if(rec_put(s,REC_TYPE_TIME, msgbuf, strlen(msgbuf))!=WRITE_OK)
		goto RECOVER;
	    s->metadata_start = 1;
	}
    // if relay_ip exists
    if(s->metadata->relay_ip)
	{
	    snprintf(msgbuf,MAX_RECORD_LEN,"client_address=%s",s->metadata->relay_ip);
	    if((rc=rec_put(s,REC_TYPE_ATTR, msgbuf, strlen(msgbuf)))==WRITE_OK)
		{
		    free(s->metadata->relay_ip);
		    s->metadata->relay_ip=NULL;
		}
	}
    if(s->metadata->helo_name)
	{
	    snprintf(msgbuf,MAX_RECORD_LEN,"helo_name=%s",s->metadata->helo_name);
	    if((rc=rec_put(s,REC_TYPE_ATTR, msgbuf, strlen(msgbuf)))==WRITE_OK)
		{
		    free(s->metadata->helo_name);
		    s->metadata->helo_name=NULL;
		}
	}
    if(s->metadata->mail_from)
	{
	    if(rec_put(s,REC_TYPE_FROM, s->metadata->mail_from, strlen(s->metadata->mail_from))==WRITE_OK)
		{
		    free(s->metadata->mail_from);
		    s->metadata->mail_from=NULL;
		}
	}
    if(s->metadata->rcptlist && s->metadata->nrcpts>0)
	{
	    int i;
	    for(i=0;i<s->metadata->nrcpts;i++)
		if(s->metadata->rcptlist[i])
		    {
			if(rec_put(s,REC_TYPE_RCPT, 
				   s->metadata->rcptlist[i], 
				   strlen(s->metadata->rcptlist[i]))==WRITE_OK)
			    {
				free(s->metadata->rcptlist[i]);
				s->metadata->rcptlist[i]=NULL;
			    }
			else
			    break;
		    }
	}
#if 1
    if(s->metadata->attrlist && s->metadata->nattrs>0)
	{
	    int i;
	    for(i=0;i<s->metadata->nattrs;i++)
		if(s->metadata->attrlist[i].aname && s->metadata->attrlist[i].avalue)
		    {
			snprintf(msgbuf,MAX_RECORD_LEN,"%s=%s",
				 s->metadata->attrlist[i].aname,
				 s->metadata->attrlist[i].avalue);
			if(rec_put(s,REC_TYPE_ATTR, 
				   msgbuf,strlen(msgbuf))==WRITE_OK)
			    {
				free(s->metadata->attrlist[i].aname);
				s->metadata->attrlist[i].aname=NULL;
				free(s->metadata->attrlist[i].avalue);
				s->metadata->attrlist[i].avalue=NULL;
			    }
			else
			    break;
		    }
		else
		    {
			if(s->metadata->attrlist[i].aname)
			    {
				free(s->metadata->attrlist[i].aname);
				s->metadata->attrlist[i].aname=NULL;
			    }
			if(s->metadata->attrlist[i].avalue)
			    {
				free(s->metadata->attrlist[i].avalue);
				s->metadata->attrlist[i].avalue=NULL;
			    }
		    }
	}
#endif
    return KASerrOK;
  RECOVER:;
    s->wr_buf.used = savelen;
    return KASerrBUSY;
}
// can write at end of func (or switch to data send if nothing to write)
int write_metadata(kas_session_t *s)
{
    int rc;
    if((rc = KasCheckSession(s))!= KASerrOK)
	return rc;
    if(s->current_status.status != KAS_STATUS_SENDING)
	return KASerrBUSY;
    // buffer filled before, at fill_metadata
    rc = KasSWriteData(s);
    switch(rc)
	{
	    // no data sent
	case __KASerrNODATA:
	    return KASerrOK;
	    // incorrect call to write
	case KASerrDISCONNECTED:
	    return KASerrWRITE_ERROR;
	case KASerrWRITE_ERROR:
	    // socket error, send event and disconnect
	    STATUS_CALLBACK_EVENT(s,KAS_STATUS_IOERROR,KAS_ES_IOERROR);
	    KasSDisconnect(s);
	    return KASerrWRITE_ERROR;
	case KASerrOK:
	    // check, if all written
	    fill_metadata(s);
	    if(s->wr_buf.used == 0) // nothing to write
		{
		    // switch to file write
		    if(s->chunked_write)
			{
			    s->write_func = NULL;
			    s->read_func = NULL;
			    s->current_status.status = KAS_STATUS_WRIDLE;
			    STATUS_CALLBACK(s);
			    return KASerrOK;
			}
		    else
			{
			    fill_file_data(s);
			    s->write_func = write_file;
			    s->read_func = NULL;
			}
		}
	    return KASerrOK;
	}
    return KASerrOK;
}

int KasSScanFile(kas_session_t *s,
		 kas_metadata_t * message_metadata,
		 char *filename_to_scan,
		 kas_cb_single_progress_t *progress_callback,
		 void *callback_param)
{
    int rc;
    struct stat st;
    
    if((rc = KasCheckSession(s))!= KASerrOK)
	return rc;
    if(!message_metadata || !filename_to_scan)
	return KASerrPARAMS_NULLPTR;
    if(s->current_status.status != KAS_STATUS_IDLE)
	return KASerrBUSY;
    if(stat(filename_to_scan,&st) || S_ISDIR(st.st_mode))
	return KASerrFILEERROR;
    
    if(s->max_spam_size>0 && st.st_size >s->max_spam_size)
	{
	    if(progress_callback)
		{
		    kas_progress_info_t cpi;
		    bzero(&cpi,sizeof(cpi));
		    cpi.pid = s->pid;
		    cpi.state = KAS_INFO_DONE;
		    cpi.message_status = KAS_EINFO_ACCEPT_CLEAN;
		    for(rc=0;rc<message_metadata->nrcpts;rc++)
			{
			    cpi.rcpt = message_metadata->rcptlist[rc];
			    (*progress_callback)(s,&cpi,callback_param);
			}
		}
	    return KASerrOK;
	}
    if((s->datafile_fd = open(filename_to_scan,O_RDONLY))<0)
	return KASerrFILEERROR;
    s->data_size = st.st_size;
    s->metadata = clone_metadata(message_metadata);
    if(!s->metadata)
	return KASerrPARAMS_NULLPTR;
    s->read_func = NULL;
    s->header_sent = 0;
    s->line_complete = 1;
    s->end_sent = 0;
    s->current_status.status = KAS_STATUS_SENDING;
    if((rc=fill_metadata(s))!=KASerrOK)
	{
	    return rc;
	}
    s->write_func = write_metadata;
    s->progress_cb = progress_callback;
    s->progress_cb_param = callback_param;
    s->current_status.status = KAS_STATUS_SENDING;
    gettimeofday(&s->scan_start,NULL);
    PROGRESS_CALLBACK_SIMPLE(s,KAS_INFO_STARTSEND);
    STATUS_CALLBACK(s);
    return KASerrIN_PROGRESS;
}

// Pass opened file, seek-ed to message start position
int KasSScanOpenFile(kas_session_t *s,
		       kas_metadata_t * message_metadata,
		       int   file_fd,
		       off_t file_data_size,
		       kas_cb_single_progress_t *progress_callback,
		       void *callback_param)
{
    int rc;
    
    if((rc = KasCheckSession(s))!= KASerrOK)
	return rc;
    if(!message_metadata || file_fd <0 || file_data_size<1)
	return KASerrPARAMS_NULLPTR;
    if(s->current_status.status != KAS_STATUS_IDLE)
	return KASerrBUSY;
    
    if(s->max_spam_size>0 && file_data_size >s->max_spam_size)
	{
	    if(progress_callback)
		{
		    kas_progress_info_t cpi;
		    bzero(&cpi,sizeof(cpi));
		    cpi.pid = s->pid;
		    cpi.state = KAS_INFO_DONE;
		    cpi.message_status = KAS_EINFO_ACCEPT_CLEAN;
		    for(rc=0;rc<message_metadata->nrcpts;rc++)
			{
			    cpi.rcpt = message_metadata->rcptlist[rc];
			    (*progress_callback)(s,&cpi,callback_param);
			}
		}
	    return KASerrOK;
	}
    s->file_pos = lseek(file_fd,SEEK_CUR,0);
    s->datafile_fd = dup(file_fd);
    s->data_size = file_data_size;
    s->metadata = clone_metadata(message_metadata);
    if(!s->metadata)
	return KASerrPARAMS_NULLPTR;
    s->read_func = NULL;
    s->header_sent = 0;
    s->line_complete = 1;
    s->end_sent = 0;
    s->current_status.status = KAS_STATUS_SENDING;
    if((rc=fill_metadata(s))!=KASerrOK)
	{
	    return rc;
	}
    s->write_func = write_metadata;
    s->progress_cb = progress_callback;
    s->progress_cb_param = callback_param;
    s->current_status.status = KAS_STATUS_SENDING;
    gettimeofday(&s->scan_start,NULL);
    PROGRESS_CALLBACK_SIMPLE(s,KAS_INFO_STARTSEND);
    STATUS_CALLBACK(s);
    return KASerrIN_PROGRESS;
}


// callback gateways
 void __single_to_many_status_gw(kas_session_t* _session,
                                     kas_status_t *old_status,
                                     kas_status_t *new_status,
                                     void* param)
{
    kas_multicallback_gateway_t *cbdata = (kas_multicallback_gateway_t*)param;
    if(_session && cbdata && cbdata->multisession && cbdata->status_cb)
	(cbdata->status_cb)(cbdata->multisession,
			    cbdata->index,
			    old_status,
			    new_status,
			    cbdata->status_cb_data);


}

void __single_to_many_progress_gw(kas_session_t	* _session,
				  kas_progress_info_t	*info,
				  void			* param)
{

    kas_multicallback_gateway_t *cbdata = (kas_multicallback_gateway_t*)param;
    if(_session && cbdata && cbdata->multisession && cbdata->progress_cb)
	(cbdata->progress_cb)(cbdata->multisession,
			      cbdata->index,
			      info,
			      cbdata->progress_cb_data);
}

kas_multisession_t *
KasMCreateSession(unsigned int num_sessions,
		  unsigned int buf_size,
		  int read_timeout,
		  int write_timeout,
		  unsigned int max_spam_size_in_bytes,
		  unsigned int scanner_data_limit_in_kbytes,
		  kas_cb_multi_status_t *status_cb,
		  void *status_cb_data)
{
    kas_multisession_t *ret;
    unsigned int i,errcnt;
    if(num_sessions<1)
	return NULL;
    if(!(ret = malloc(sizeof(unsigned int)+(num_sessions+1)*sizeof(__kas_multisession_entry_t))))
	return NULL;
    ret->count = num_sessions;
    ret->pfds = malloc(sizeof(struct pollfd)*(ret->count+0));
    ret->socket_read = malloc(sizeof(int)*(ret->count+0));
    ret->socket_write = malloc(sizeof(int)*(ret->count+0));
    if(!ret->pfds || !ret->socket_write ||!ret->socket_read)
	{
	    free(ret->pfds);
	    free(ret->socket_read);
	    free(ret->socket_write);
	    free(ret);
	    return NULL;
	}
    bzero(&ret->sessions[0],num_sessions*sizeof(__kas_multisession_entry_t));
    for(i=0,errcnt=0;i<num_sessions;i++)
	{
	    if(KasSCreateSession(&ret->sessions[i].kas_session,
				 buf_size, read_timeout,write_timeout,
				 max_spam_size_in_bytes,scanner_data_limit_in_kbytes,
				 __single_to_many_status_gw,
				 &ret->sessions[i].cb_data))
		{
		    ret->sessions[i].cb_data.index = i;
		    ret->sessions[i].cb_data.multisession = ret;
		    ret->sessions[i].cb_data.status_cb = status_cb;
		    ret->sessions[i].cb_data.status_cb_data = status_cb_data;
		}
	    else
		errcnt++;
	}
    if(errcnt)
	{
	    KasMSessionDestroy(ret);
	    return NULL;
	}
    else
	return ret;

}
#define CHECK_INDEX(ses,idx)do{if(!ses) return KASerrPARAMS_NULLPTR; if(idx>=ses->count)return KASerrINDEX_RANGE;}while(0)


int KasMScanFile(kas_multisession_t *s,
		 unsigned int index,
		 kas_metadata_t * message_metadata,
		 char *filename_to_scan,
		 kas_cb_multi_progress_t *progress_callback,
		 void *callback_param)
{
    CHECK_INDEX(s,index);
    s->sessions[index].cb_data.progress_cb = progress_callback;
    s->sessions[index].cb_data.progress_cb_data = callback_param;
    return KasSScanFile(&s->sessions[index].kas_session,
			message_metadata,filename_to_scan,
			__single_to_many_progress_gw,
			&s->sessions[index].cb_data);

}
int KasMScanOpenFile(kas_multisession_t *s,
		     unsigned int index,
		     kas_metadata_t * message_metadata,
		     int file_fd,
		     off_t data_size,
		     kas_cb_multi_progress_t *progress_callback,
		     void *callback_param)
{
    CHECK_INDEX(s,index);
    s->sessions[index].cb_data.progress_cb = progress_callback;
    s->sessions[index].cb_data.progress_cb_data = callback_param;
    return KasSScanOpenFile(&s->sessions[index].kas_session,
			message_metadata,file_fd,data_size,
			__single_to_many_progress_gw,
			&s->sessions[index].cb_data);

}

static int KasMCheckSessions(kas_multisession_t *sesptr)
{
    int i;
    if(!sesptr) return KASerrPARAMS_NULLPTR;
    if(sesptr->count<1)
	return KASerrPARAMS_NOTINITED;
    for(i=0;i<sesptr->count;i++)
	if(KasCheckSession(&sesptr->sessions[i].kas_session)!=KASerrOK)
	    return KASerrPARAMS_NOTINITED;
    return KASerrOK;
}


void KasMSessionDestroy(kas_multisession_t *ret)
{
    int i;
    if(!ret || !ret->count)
	return;
    free(ret->pfds);
    free(ret->socket_read);
    free(ret->socket_write);
    for(i=0;i<ret->count;i++)
	KasSSessionDestroy(&ret->sessions[i].kas_session);
    free(ret);
}

int KasMConnect(kas_multisession_t *s, unsigned int index,const char *daemon_addr, int timeout )
{
    CHECK_INDEX(s,index);
    return KasSConnect(&s->sessions[index].kas_session,daemon_addr,timeout);
}

int KasMDisconnect(kas_multisession_t *s, unsigned int index)
{
    CHECK_INDEX(s,index);
    return KasSDisconnect(&s->sessions[index].kas_session);
}

int KasMPing(kas_multisession_t *s, unsigned int index )
{
    CHECK_INDEX(s,index);
    return KasSPing(&s->sessions[index].kas_session);
}


static int KasMNearestTimeout(kas_multisession_t *sessp, struct timeval *now, int timeout)
{
    int start_to = KasSNearestTimeout(&sessp->sessions[0].kas_session,now,timeout);
    int current_to,i;

    for(i=1;i<sessp->count;i++)
	{
	    current_to = KasSNearestTimeout(&sessp->sessions[i].kas_session,now,timeout);
	    if(current_to>=0 && current_to <start_to)
		start_to = current_to;
	}
    return start_to;
}

int KasMPreparePoll(kas_multisession_t *sessp, struct pollfd *pfds, unsigned int count)
{
    int i,rc;
    if((rc = KasMCheckSessions(sessp))!= KASerrOK)
	return rc;

    for(i=0;i<sessp->count && i<count;i++)
	{
	    if((rc=KasSPreparePoll(&sessp->sessions[i].kas_session,pfds,i))!=KASerrOK)
		return rc;
	}
    return KASerrOK;
}
int KasMPrepareSelect(kas_multisession_t *sessp, fd_set *readfds, fd_set *wrfds)
{
    int i,rc;
    if((rc = KasMCheckSessions(sessp))!= KASerrOK)
	return rc;

    for(i=0;i<sessp->count ;i++)
	{
	    if((rc=KasSPrepareSelect(&sessp->sessions[i].kas_session,readfds,wrfds))!=KASerrOK)
		return rc;
	}
    return KASerrOK;
}

static int  multi_check_timeout2(kas_multisession_t *sesptr,struct timeval *now)
{
    int rc,errcnt,i;

    if((rc = KasMCheckSessions(sesptr))!= KASerrOK)
	return rc;
    for(i=0,errcnt=0;i<sesptr->count;i++)
	if(check_timeout2(&sesptr->sessions[i].kas_session,now)!=KASerrOK)
	    errcnt++;
    return errcnt?__KASerrTIMEOUT:KASerrOK;
}

static int  check_all_idle(kas_multisession_t *sesptr)
{
    int rc,cnt,i;

    if((rc = KasMCheckSessions(sesptr))!= KASerrOK)
	return rc;
    for(i=0,cnt=0;i<sesptr->count;i++)
	if(sesptr->sessions[i].kas_session.current_status.status != KAS_STATUS_IDLE)
	    cnt++;
    return cnt>0?0:1;
}

static int  count_idle(kas_multisession_t *sesptr)
{
    int rc,cnt,i;

    if((rc = KasMCheckSessions(sesptr))!= KASerrOK)
	return rc;
    for(i=0,cnt=0;i<sesptr->count;i++)
	if(sesptr->sessions[i].kas_session.current_status.status == KAS_STATUS_IDLE)
	    cnt++;
    return cnt;
}


static int KasMProcessIOResult(kas_multisession_t *msesptr,int *socket_read, int *socket_write)
{
    int rc,i,errcnt,allok=1;
    kas_status_t kstatus;

    bzero(&kstatus,sizeof(kstatus));
    if((rc = KasMCheckSessions(msesptr))!= KASerrOK)
	return rc;
    for(i=0,errcnt=0;i<msesptr->count;i++)
	{
	    rc = KasSProcessIOResult(&msesptr->sessions[i].kas_session,
				     msesptr->socket_read[i],msesptr->socket_write[i]);
	    if(rc!=KASerrOK && rc!=KASerrIN_PROGRESS)
		errcnt++;
	    if(rc!=KASerrOK)
		allok=0;
	}
    return errcnt?KASerrIO_ERROR:(allok?KASerrOK:KASerrIN_PROGRESS);
}

int KasMProcessPollResult(kas_multisession_t *msesptr, struct pollfd* pollfds, unsigned int maxidx)
{
    int i,rc;
    kas_session_t *sesptr;

    if(!pollfds ||!msesptr)
	return KASerrPARAMS_NULLPTR;
    if((rc = KasMCheckSessions(msesptr))!= KASerrOK)
	return rc;
    bzero(msesptr->socket_read,sizeof(int)*msesptr->count);
    bzero(msesptr->socket_write,sizeof(int)*msesptr->count);
    for(i=0;i<msesptr->count && i < maxidx;i++)
	{
	    sesptr = &msesptr->sessions[i].kas_session;
	    if(pollfds[i].fd != sesptr->socket) continue;
#ifdef POLLRDNORM
	    if(pollfds[i].revents & (POLLIN|POLLRDNORM|POLLRDBAND))
#else
	    if(pollfds[i].revents & (POLLIN))
#endif
		msesptr->socket_read[i] = 1;
#ifdef POLLRDNORM
	    if(pollfds[i].revents & (POLLOUT|POLLWRNORM|POLLWRBAND))
#else
	    if(pollfds[i].revents & (POLLOUT))
#endif
		msesptr->socket_write[i] = 1;
	    if(pollfds[i].revents & (POLLERR|POLLHUP))
		msesptr->socket_write[i] = 1;
	}
    return KasMProcessIOResult(msesptr,msesptr->socket_read,msesptr->socket_write);
}

int KasMProcessSelectResult(kas_multisession_t *msesptr, fd_set *readfd, fd_set *writefd)
{
    int i,rc;
    kas_session_t *sesptr;

    if(!msesptr)
	return KASerrPARAMS_NULLPTR;
    if((rc = KasMCheckSessions(msesptr))!= KASerrOK)
	return rc;
    bzero(msesptr->socket_read,sizeof(int)*msesptr->count);
    bzero(msesptr->socket_write,sizeof(int)*msesptr->count);
    for(i=0;i<msesptr->count;i++)
	{
	    sesptr = &msesptr->sessions[i].kas_session;
	    if(FD_ISSET(sesptr->socket,readfd))
		msesptr->socket_read[i] = 1;
	    if(FD_ISSET(sesptr->socket,writefd))
		msesptr->socket_write[i] = 1;
	}
    return KasMProcessIOResult(msesptr,msesptr->socket_read,msesptr->socket_write);
}

int KasMPeekScanner(kas_multisession_t *sessp,int flags)
{
    int i,rc;
    if((rc = KasMCheckSessions(sessp))!= KASerrOK)
	return -1;
    for(i=0;i<sessp->count;i++)
	if(sessp->sessions[i].kas_session.current_status.status & flags)
	    return i;
    return -1;
}

int KasMCountScanners(kas_multisession_t *sessp,int mask)
{
    int i,rc;
    if((rc = KasMCheckSessions(sessp))!= KASerrOK)
	return -1;
    rc=0;
    for(i=0;i<sessp->count;i++)
	if(sessp->sessions[i].kas_session.current_status.status & mask)
	    rc++;
    return rc;
}

static int KasMEventLoop(kas_multisession_t *sessp, int timeout, int once, int idlecount)
{
    int rc;
    struct timeval now,start;
    int real_timeout,timed;
    int try_retry;
    if((rc = KasMCheckSessions(sessp))!= KASerrOK)
	return rc;

    gettimeofday(&start,NULL);
    real_timeout = KasMNearestTimeout(sessp,&start,timeout);
    while(1)
	{
	    try_retry = 0;
	    bzero(sessp->pfds,sizeof(struct pollfd)*sessp->count);
	    if(KASerrOK!=(rc=KasMPreparePoll(sessp,sessp->pfds,sessp->count)))
		return rc;
	    rc = poll(sessp->pfds,sessp->count,real_timeout);
	    gettimeofday(&now,NULL);
	    if(rc<0)
		{
		    if(errno==EINTR)
			try_retry = 1;
		}
	    else if(rc==0)
		{
		    if(multi_check_timeout2(sessp,&now)!=KASerrOK)
			{
			    return KASerrIN_PROGRESS;
						  // Internal TO elapsed
			}
		}
	    else
		{
		    rc = KasMProcessPollResult(sessp,sessp->pfds,sessp->count);
		    if(rc != KASerrOK && rc != KASerrIN_PROGRESS)
			return rc;
		    if(idlecount>0 && count_idle(sessp)>=idlecount)
			return KASerrOK;
		    if(check_all_idle(sessp))
			return KASerrOK;
		    if(multi_check_timeout2(sessp,&now)!=KASerrOK)
			return KASerrIN_PROGRESS;
						  // Internal TO elapsed
		}
	    if(!once || try_retry)
		{
		    timed=diff_msec(&start,&now);
		    if(timeout>0)
			{
			    if(timed>0)
				{
				    if( timeout<timed)
					return KASerrIN_PROGRESS;
				    else
					timeout-=timed;
				}
			}
		    real_timeout = KasMNearestTimeout(sessp,&now,timeout);
		    if(real_timeout==0)
			return KASerrIN_PROGRESS;
		}
	    else
		return KASerrIN_PROGRESS;
	}
}

int KasMProcessEvents(kas_multisession_t *sessp, int timeout)
{
    return KasMEventLoop(sessp,timeout,1,0);
}

int KasMWaitIdle(kas_multisession_t *sessp, int count, int timeout)
{
    if(count_idle(sessp)>=count)
	return KASerrOK;
    return KasMEventLoop(sessp,timeout,0,count);
}




EXTERN_C_END
