/* -*- C -*-
 * Spamtest connection library
 * File: spamtest_util.c
 *      (C)  Alex Tutubalin <lexa@lexa.ru>
 *	Apr-May 2003
 *
 * Created: Wed Apr 23 22:08:08 2003
 */
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include "spamtest.h"
#include "spamtest_internal.h"

int comp_hdr(const void *a1,const void *a2)
{
    hdraction_t *h1=(hdraction_t *)a1;
    hdraction_t *h2=(hdraction_t *)a2;
    int r;
    if(h1->type != h2->type)
	return (h1->type - h2->type)*1000;
    if(h1->header && h2->header && (r=strcmp(h1->header,h2->header)))
	return r;
    if(h1->value && h2->value && (r=strcmp(h1->value,h2->value)))
	return r;
    return 0;
}

#define STATUS_ALLOC 10
#define BODY_ALLOC1		102400

int msgstat_addbody(message_status_t *msg, msg_heap_t *heap, char *line, int len)
{
    if(!msg || !line)
	return -1;
    if(!msg->bouncemsg)
	if(!(msg->bouncemsg=msgtext_init(heap)))
	    return -1;
    return msgtext_puttext(msg->bouncemsg,line,len);
}


message_status_t *make_new_status(spamtest_status_t *st)
{
    message_status_t *ret;
    if(!st || !st->storeptr) return NULL;
    if(!st->ms_array)
	{
	    if(!(st->ms_array = heap_allocz(st->storeptr,sizeof(message_status_t)*STATUS_ALLOC)))
		return NULL;
	    st->count = 0;
	    st->alloc = STATUS_ALLOC;
	}
    else if(st->count==st->alloc)
	{
	    message_status_t *newa;
	    newa = heap_reallocz(st->storeptr,st->ms_array,
				st->alloc*sizeof(message_status_t),
				(st->alloc+STATUS_ALLOC)*sizeof(message_status_t));
	    if(newa)
		{
		    st->alloc+=STATUS_ALLOC;
		    st->ms_array = newa;
		}
	    else
		return NULL;
	}
    ret = &st->ms_array[st->count];
    st->count++;
    bzero(ret,sizeof(*ret));
    ret->bouncemsg = NULL;
    ret->rcpts = rcptlist_init(st->storeptr);
    return ret;
}




int add_action(message_status_t *msg, msg_heap_t *heap, int type, char *hdr, char *val)
{
    hdraction_t *a = make_new_action(msg,heap);
    if(!a) return -1;
    if(!(a->header = heap_strdup(heap,hdr)))
	return -1;
    if(val)
	if(!(a->value = heap_strdup(heap,val)))
	    return -1;
    a->type = type;
    return 0;
}

hdraction_t *make_new_action(message_status_t *ms,msg_heap_t *heap)
{
    hdraction_t *ret;
    if(!ms || !heap) return NULL;

    if(!ms->act_array)
	{
	    if(!(ms->act_array = heap_allocz(heap,sizeof(hdraction_t)*STATUS_ALLOC)))
		return NULL;
	    ms->action_count = 0;
	    ms->action_alloc = STATUS_ALLOC;
	}
    else if(ms->action_count==ms->action_alloc)
	{
	    hdraction_t *newa;
	    newa = heap_reallocz(heap,ms->act_array,
				ms->action_alloc*sizeof(hdraction_t),
				(ms->action_alloc+STATUS_ALLOC)*sizeof(hdraction_t));
	    if(newa)
		{
		    ms->action_alloc+=STATUS_ALLOC;
		    ms->act_array = newa;
		}
	    else
		return NULL;
	}
    ret = &ms->act_array[ms->action_count];
    ms->action_count++;
    bzero(ret,sizeof(*ret));
    return ret;
}
    

headers_list_t* make_changed_headers(message_status_t * sts, message_t *message)
{
    headers_list_t* newhdr = headers_clone(message);
    int i;
    hdraction_t *hdra;
    if(!newhdr) return NULL;

    for(i=0;i<sts->action_count;i++)
	{
	    hdra = &sts->act_array[i];
	    if(hdra->type == STS_HEADER_DEL && hdra->header)
		{
		    if(-1==headers_del_hdr(newhdr,hdra->header))
			goto ERROR;
		}
	    else if(hdra->type == STS_HEADER_CHG && hdra->header)
		{
		    if(-1==headers_del_hdr(newhdr,hdra->header))
			goto ERROR;
		    if(-1==headers_add_hdr(newhdr,hdra->header,hdra->value?hdra->value:""))
			goto ERROR;
		}
	    else if(hdra->type == STS_HEADER_ADD && hdra->header && hdra->value )
		{
		    if(-1==headers_add_hdr(newhdr,hdra->header,hdra->value))
			goto ERROR;
		}
	    else if(hdra->type == STS_HEADER_NEW && hdra->header && hdra->value )
		{
		    if(-1==headers_new_hdr(newhdr,hdra->header,hdra->value))
			goto ERROR;
		}
	    else if(hdra->type == STS_HEADER_PREPEND && hdra->header && hdra->value )
		{
		    if(-1==headers_prepend_hdr(newhdr,hdra->header,hdra->value))
			goto ERROR;
		}
	}
    return newhdr;
  ERROR:
    return NULL;
}

spamtest_status_t* glue_actions(spamtest_status_t *s)
{
    spamtest_status_t *result;

    int i,j,k;
    message_status_t *ms,*mret=NULL,*msp=NULL,*msfrom=NULL;
    hdraction_t *hdr;
    recipient_list_t *rcpts=NULL;
    int action_index;

    if(!s || !s->storeptr) return NULL;
    
    if(s->count <2)
	return s;
    
    if(!(result = heap_allocz(s->storeptr,sizeof(spamtest_status_t))))
	return NULL;

    // copy status, reason and so 
    result->storeptr = s->storeptr;
    result->status = s->status;
    result->errorcode = s->errorcode;
    result->reason = heap_strdup(s->storeptr,s->reason?s->reason:"");
    if(s->count<1)
	return result;

    // pass 1, find 1st change action
    for(j=0;j<s->count;j++)
	{
	    ms = &s->ms_array[j];
	    if(ms->action == STS_ACTION_CHANGE)
		{
		    msp = ms;
		    if(ms->mailfrom && *ms->mailfrom)
			msfrom = ms;
		    break;
		}
	}

    if(!msp)
	{
	    // We have only bounce messages, create single
	    // bounce from it
	    if(!(mret = make_new_status(result)))
		return NULL;
	    mret->action = STS_ACTION_BOUNCE;
	    for(j=0;j<s->count;j++)
		{
		    ms = &s->ms_array[j];
		    if(ms->action == STS_ACTION_BOUNCE)
			{
			    if(ms->mailfrom && !mret->mailfrom)
				if(!(mret->mailfrom = heap_strdup(result->storeptr,ms->mailfrom)))
				    return NULL;

			    if(ms->bouncemsg && !mret->bouncemsg)
				mret->bouncemsg = ms->bouncemsg; 

			    if(ms->rcpts && mret->rcpts)
				for(i=0;i<rcptlist_count(ms->rcpts);i++)
				    if(-1 ==rcptlist_adduniq(mret->rcpts,rcptlist_getn(ms->rcpts,i)))
					return NULL;
			}
		}
	    return result;
	}

    // pass 2 - alloc and copy
    if(!(rcpts = rcptlist_init(result->storeptr)))
	return NULL;
    
    if(!(mret = make_new_status(result)))
	return NULL;
    mret->action = STS_ACTION_CHANGE;
    if(msfrom)
	mret->mailfrom = heap_strdup(result->storeptr,msfrom->mailfrom);
    else
	mret->mailfrom = heap_strdup(result->storeptr,""); // postmaster 
    if(!mret->mailfrom)
	return NULL;

    // work with 1st action;
    action_index = -1;
    for(k=0; k<s->count;k++)
	if(s->ms_array[k].action == STS_ACTION_CHANGE)
	    {
		action_index = k;
		break;
	    }
    if(action_index<0)
	return result; // was empty set

    //copy all actions from 1st list, double check for deletions
    ms = &s->ms_array[action_index];
    if(ms->rcpts)
	for(i=0;i<rcptlist_count(ms->rcpts);i++)
	    rcptlist_adduniq(mret->rcpts,rcptlist_getn(ms->rcpts,i));
    for(j=0;j<ms->action_count;j++)
	{
	    hdr = &ms->act_array[j];
	    if(!hdr->header)
		continue;

	    if(hdr->type == STS_HEADER_DEL)
		{
		    // is really such deletion exists in all action
		    // lists ?
		    int chgcount = 0;
		    for(k=action_index+1; k<s->count;k++)
			{
			    if(s->ms_array[k].action == STS_ACTION_CHANGE)
				{
				    chgcount++;
				    for(i=0;i<s->ms_array[k].action_count;i++)
					if(s->ms_array[k].act_array[i].type == STS_HEADER_CHG
					   && s->ms_array[k].act_array[i].header
					   && !strcmp(s->ms_array[k].act_array[i].header,hdr->header))
					    {
						chgcount--;
						break;
					    }
				}
			    if(chgcount == 0)
				if(add_action(mret,result->storeptr,
					      hdr->type,hdr->header,hdr->value)<0)
				    return NULL;
			}
		}
	    else 
		{
		    // ADD/NEW/PREPEND
		    if(add_action(mret,result->storeptr,hdr->type,hdr->header,hdr->value)<0)
			return NULL;
		}
	}
    // we've copied 1st actionlist, go to other actions....
    for(k=action_index+1;k<s->count;k++)
	{
	    ms = &s->ms_array[k];
	    if(ms->rcpts)
		for(i=0;i<rcptlist_count(ms->rcpts);i++)
		    rcptlist_adduniq(mret->rcpts,rcptlist_getn(ms->rcpts,i));
	        for(j=0;j<ms->action_count;j++)
		    {
			hdr = &ms->act_array[j];
			if(!hdr->header)
			    continue;
			if(hdr->type == STS_HEADER_DEL)
			    continue;
			else if(hdr->value)
			    {
				// find dupes in mret if any
				int found = 0;
				for(i=0;i<mret->action_count;i++)
				    if(mret->act_array[i].type == hdr->type
				       && !strcmp(mret->act_array[i].header,hdr->header)
				       && (!hdr->value || 
					   !strcmp(mret->act_array[i].value,hdr->value)
					   )
					)
					{
					    found=1;
					    break;
					}
				if(!found)
				    if(add_action(mret,result->storeptr,
						  hdr->type,hdr->header,hdr->value)<0)
					return NULL;
			    }
		    }
	}

    return result;
}

static void dump_action(hdraction_t *hdr)
{
    if(!hdr) return;
    if(hdr->type == STS_HEADER_DEL)
	printf("DELETE: %s\n",hdr->header?hdr->header:"header field missing");
    else if(hdr->type == STS_HEADER_CHG)
	printf("CHANGE: %s|%s\n",hdr->header?hdr->header:"header field missing",
	       hdr->value?hdr->value:"header value missing");
    else if(hdr->type == STS_HEADER_PREPEND)
	printf("PREPEND: %s|%s\n",hdr->header?hdr->header:"header field missing",
	       hdr->value?hdr->value:"header value missing");
    else if(hdr->type == STS_HEADER_ADD)
	printf("ADD: %s|%s\n",hdr->header?hdr->header:"header field missing",
	       hdr->value?hdr->value:"header value missing");
    else if(hdr->type == STS_HEADER_NEW)
	printf("NEW: %s|%s\n",hdr->header?hdr->header:"header field missing",
	       hdr->value?hdr->value:"header value missing");
    else
	printf("Unknown HDR change code %d\n",hdr->type);
}

static void recipient_writer(const char *rcpt,size_t unused)
{
    printf("RCPT TO: %s\n",rcpt);
}

void dump_message_status(spamtest_status_t *smsg)
{
    message_status_t *msg;
    int i,j;
    if(!smsg || !smsg->ms_array || !smsg->count ) return;

    

    for(i=0;i<smsg->count;i++)
	{
	    msg = &smsg->ms_array[i];

	    printf("==Action: %s\n",msg->action==STS_ACTION_CHANGE?"CHANGE":"Bounce");
	    if(msg->mailfrom)
		printf("MAIL FROM: %s\n",msg->mailfrom);
	    rcptlist_iterate(msg->rcpts,recipient_writer);
	    if(msg->action == STS_ACTION_CHANGE)
		{
		    if(msg->action_count>0)
			{
			    printf("Actions:\n");
			    for(j=0;j<msg->action_count;j++)
				dump_action(&msg->act_array[j]);
			}
		    else
			printf("Empty Actions list\n");
		}
	    else if(msg->action == STS_ACTION_BOUNCE)
		{
		    if(msg->bouncemsg)
			{
			    printf("Bounce message:\n");
			    for(j=0;j<msgtext_count_chunks(msg->bouncemsg);j++)
				fwrite(msgtext_get_chunk(msg->bouncemsg,j),1,
				       msgtext_chunk_len(msg->bouncemsg,j),
				       stdout);
			}
		    else
			printf("Empty bounce message\n");
		}
	    else
		printf("Unknown action type %d\n",msg->action);
	}
}
